/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"

#include "str-lib.h"
#include "rafutil.h"
#include "dafstype.h"
#include "entity.h"
#include "dafs.h"
#include "smemory.h"

Dll void APIENTRY
i_SetNoData(iEntityPtr entity)
{
	if (gCallBack)
		MatrCallBack(mWillUpdate,gUserData,entity);

	i_FreeContents(entity);

	if (gCallBack)
		MatrCallBack(mUpdate,gUserData,entity);
}

void
i_FreeContents(iEntityPtr entity)  
{

	switch (entity->data->dt) {
	case iDString:
		if (entity->data->d.at.text)
			smfree(entity->data->d.at.text);
		break;
	case iDUString:
		if (entity->data->d.ut.unitext)
			smfree(entity->data->d.ut.unitext);
		break;
	case iDPosSet:
		if (entity->data->d.pp.len && entity->data->d.pp.posp)
			smfree(entity->data->d.pp.posp);
		break;
	case iDUserData:
		if (entity->data->d.ud.user && entity->data->d.ud.len)
			smfree(entity->data->d.ud.user);
		break;
	default: break;
	}
	entity->data->dt = iDNo;
}

Dll const char * APIENTRY
i_GetText(const iEntityPtr entity,short *conf)
{
	if (entity->data->dt != iDString) return 0;
	if (conf) *conf = entity->data->d.at.conf;
	return entity->data->d.at.text;
}

Dll iDAFSError APIENTRY
i_SetText(iEntityPtr entity, const char *text,short conf)
{
	if (!entity) return NotEnoughData;
	if (gCallBack)
		MatrCallBack(mWillUpdate,gUserData,entity);

	if (text) {
		char *temp;
		int len;
		len = strlen(text)+1;
		temp = (char*)smalloc(len);
		if (!temp)
			return MallocFailError;
		strcpy(temp,text);
		i_FreeContents(entity);
		entity->data->dt = iDString;
		entity->data->d.at.len = len;
		entity->data->d.at.text = temp;
		entity->data->d.at.conf = conf;
	} else {
		i_FreeContents(entity);
	}

	if (gCallBack)
		MatrCallBack(mUpdate,gUserData,entity);

	return DAFSOK;
}

Dll const iUCode * APIENTRY
i_GetUText(const iEntityPtr entity,short *conf)  
{
	if (entity->data->dt != iDUString)
		return 0;
	if (conf) *conf = entity->data->d.ut.conf;
	return entity->data->d.ut.unitext;
}

Dll iDAFSError APIENTRY
i_SetUText(iEntityPtr entity, const iUCode *text,short conf)  
{

	if (!entity) return NotEnoughData;
	if (gCallBack)
		MatrCallBack(mWillUpdate,gUserData,entity);

	if (text) {
		iUCode *temp;
		int2 len;
		len = ucslen(text)+1;
		temp = (iUCode*)smalloc(len * sizeof(iUCode));
		if (!temp)
			return MallocFailError;
		ucscpy(temp,text);
		
		i_FreeContents(entity);
		
		entity->data->dt = iDUString;
		entity->data->d.ut.len = len;
		entity->data->d.ut.unitext = temp;
		entity->data->d.ut.conf = conf;
	} else
		i_FreeContents(entity);

	if (gCallBack)
		MatrCallBack(mUpdate,gUserData,entity);

	return DAFSOK;
}

Dll int2 APIENTRY
i_GetConf(const iEntityPtr entity)  
{
	int2 size;
	int2 *pptr;

	if (!entity)
		return -1;
	switch(entity->data->dt){
	case iD1PosSet:
	case iDConf:
	case iDShort:
		return entity->data->d.ps.conf;
	case iDPosSet:
		if (entity->data->d.pp.len) {
			pptr = i_GetPosLimit(entity,&size);
			if(!pptr) return entity->data->d.pp.posp[0].conf;
			else return entity->data->d.pp.posp[*pptr].conf;
		}
		else
			return 0;
		break;
	case iDUString:
	        return entity->data->d.ut.conf;
	case iDString:
	        return entity->data->d.at.conf;
	default:
		;
	}
	return -1;
}

Dll void APIENTRY
i_SetConf(iEntityPtr entity,int2 conf)  
{
	if (!entity)
		return;
	if (gCallBack)
		MatrCallBack(mWillUpdate,gUserData,entity);
	
	i_FreeContents(entity);
	entity->data->dt = iDConf;
	entity->data->d.ps.conf = conf;
	if (gCallBack)
		MatrCallBack(mUpdate,gUserData,entity);
}

Dll int2 APIENTRY
i_GetGlyph(const iEntityPtr entity, int2 *conf)
{
	int2 size;
	int2 *pptr;

	if (!entity)
		return 0;
	switch(entity->data->dt){
	case iD1PosSet:
		if (conf)
			*conf = entity->data->d.ps.conf;
		return entity->data->d.ps.chr;
	case iDPosSet:
		if (entity->data->d.pp.len) {
			pptr = i_GetPosLimit(entity,&size);
			if(!pptr) {
			    if(conf) 
				    *conf = entity->data->d.pp.posp[0].conf;
			    return entity->data->d.pp.posp[0].chr;
			}
			else {
			    if(conf) 
				    *conf = entity->data->d.pp.posp[*pptr].conf;
			    return entity->data->d.pp.posp[*pptr].chr;
			}
		}
	default:
		return 0;
	}
	return 0;
}

Dll void APIENTRY
i_SetGlyph(iEntityPtr entity, int2 chr, int2 conf)
{	
	int4 index;
	iEntityPtr *owners;

	if (!entity || !entity->data)
		return;
	if (gCallBack)
		MatrCallBack(mWillUpdate,gUserData,entity);

	i_FreeContents(entity);
	entity->data->dt = iD1PosSet;
	entity->data->d.ps.conf = conf;
	entity->data->d.ps.chr = chr;
	owners = entity->data->matrs;
	if (entity->data->use >1)
		for (index=0;index<entity->data->use;index++)
			i_NoPosLimit(entity->data->matrs[index]);
	if (gCallBack)
		MatrCallBack(mUpdate,gUserData,entity);
}

Dll int2 APIENTRY
i_GetShort(const iEntityPtr entity, int2* conf)
{
	if (!entity)
		return 0;
	switch(entity->data->dt){
	case iDShort:
		if (conf)
			*conf = entity->data->d.sd.conf;
		return entity->data->d.sd.num;
	default:
		return 0;
	}
}

Dll void APIENTRY
i_SetShort(iEntityPtr entity, int2 num, int2 conf)
{	int4 index;
	iEntityPtr *owners;

	if (!entity || !entity->data)
		return;
	if (gCallBack)
		MatrCallBack(mWillUpdate,gUserData,entity);

	i_FreeContents(entity);
	entity->data->dt = iDShort;
	entity->data->d.sd.conf = conf;
	entity->data->d.sd.num = num;
	owners = entity->data->matrs;
	if (entity->data->use >1)
		for (index=0;index<entity->data->use;index++)
			i_NoPosLimit(entity->data->matrs[index]);
	if (gCallBack)
		MatrCallBack(mUpdate,gUserData,entity);
}

Dll iPosSetPtr APIENTRY
i_GetPosSet(iEntityPtr entity, int2 *n)
{
	if (n) *n = 0;
	else return 0;
	if (!entity)
		return 0;
	switch(entity->data->dt){
	case iDPosSet:
		*n=entity->data->d.pp.len;
		return entity->data->d.pp.posp;
	case iD1PosSet:
		if (entity->data->d.ps.chr)
			*n=1;
		return &(entity->data->d.ps);
	default: 
		return 0;
	}
}

Dll iDAFSError APIENTRY
i_SetPosSet(iEntityPtr entity, int2 n, const iPosSetPtr  chrs)
{
	int4 i;
	iPosSetPtr temp = 0;
	iPosSet tempp;
	
	if (!entity )
		return NotEnoughData;
	
	if (gCallBack)
		MatrCallBack(mWillUpdate,gUserData,entity);
	if (n > 1) {
		temp = (iPosSetPtr)smalloc(sizeof(iPosSet)*n);
		if (!temp) return MallocFailError;
		for(i=0;i<n;i++)
			temp[i]=chrs[i];
	} else if (n) {
		tempp = *chrs;
	}
	
	i_FreeContents(entity);
	
	if (n > 1) {
		entity->data->dt = iDPosSet;
		entity->data->d.pp.len = n;
		entity->data->d.pp.posp=temp;		
	} else if (n) {
		entity->data->d.ps = tempp;
		entity->data->dt = iD1PosSet;
	} else {
		entity->data->dt = iDPosSet;
		entity->data->d.pp.posp = 0;
		entity->data->d.pp.len = 0;
	}
	if (gCallBack)
		MatrCallBack(mUpdate,gUserData,entity);

	return DAFSOK;
}

Dll iDTag APIENTRY
i_GetEDataType(iEntityPtr entity)
{
	if (!entity)
		return 0;
	if (entity->data)
		return entity->data->dt;
	else
		return 0;
}
