/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */
#include "rafport.h"
#include "rafutil.h"

#include "smemory.h"
#include "entity.h"
#include "dafs.h"

/*
// this table is small -- 8 bytes
*/
iTypeTable i_GlobalType;         

static int4 
isDataEnSub(iEntityPtr target)
{
    return (
	(i_GetEDataType(target) != iDNo) && 
	(i_GetEDataType(target) != iDConf) && 
	(i_GetEDataType(target) != iDUserData) &&
	((i_GetEDataType(target) != iDShort) || 
	 target->data->type == iSpaceT || 
	 target->data->type == iNewLineT)
	);
}


static int4 
_iAncestorIsData(iEntityPtr en)
{
    iEntityPtr parent = en ? i_GetParentEntity(en) : 0;
    int4 value;
	
    value =  (parent ? 
	      isDataEnSub(parent) || _iAncestorIsData(parent) : 
	      FALSE);
	
    return value;  
}

static int4 
_iIsDataEn(iEntityPtr target)
{
    return isDataEnSub(target) && !_iAncestorIsData(target);
}

int4 
IsHit(iEntityPtr target,iTypeIndex type, iSearchType method, void *name)
{
    iEntityPtr par;
    if ((target->data->type != type) && type) return 0;
    switch (method) {
    case ALL:	
	return 1;
    case POSSET: 
	return i_GetGlyph(target,0)==(long)name;
    case PROPERTY: 
	return i_FindProp(target, (char*)name) != 0;
    case NOCHILD:
	return !i_GetChildEntity(target);
    case DATAEN:
	return _iIsDataEn(target);
	break;
    case NOTOR: 
	if(par=i_GetParentEntity(target)){
	    if(i_GetLogicalType(par)==iOr)
		return 0;
	}
	return 1;
    }
    return 0;
}

static iEntityPtr 
SearchForward(iEntityPtr entity,iTypeIndex type,
	      iSearchType method, void *name)
{
    iEntityPtr cur,find;

#if 0	
    if ((entity->data->use > 1) && (entity != entity->data->matrs[0]))
	return 0;
#endif
    if (IsHit(entity,type,method,name))	
	return entity;
    for (cur=i_GetChildEntity(entity);cur; cur=i_GetNextEntity(cur))
	if (find=SearchForward(cur,type,method,name))
	    return find;
    return 0;
}

static iEntityPtr 
SearchBack(iEntityPtr entity,iTypeIndex type,
	   iSearchType method, void *name)
{
    iEntityPtr cur,find;
	
#if 0
    if ((entity->data->use > 1) && (entity != entity->data->matrs[0]))
	return 0;
#endif
    for (cur=i_GetLastEntity(entity);cur; cur=i_GetPrevEntity(cur))
	if (find=SearchBack(cur,type,method,name))
	    return find;
    if (IsHit(entity,type,method,name))	
	return entity;
    return 0;
}

static iEntityPtr 
findtype(iEntityPtr src, iTypeIndex type)
{
    iEntityPtr cur,find;
	
    if(!src)
	return 0;
    if(i_GetEntityType(src)==type)
	return src;
    for(cur=i_GetChildEntity(src);cur;cur=i_GetNextEntity(cur))
	if(find=findtype(cur,type))
	    return find;
    return 0;
}

Dll iEntityPtr APIENTRY
i_FindEntityForward(iEntityPtr home, iTypeIndex type,
		    iSearchType method, void *name)
{
    iEntityPtr target;
	
    if (home && home->data->child){ 
	home = i_GetChildEntity(home);
	if (target = SearchForward(home,type,method,name))
	    return target;
    }
    for (;;) {
	if (home->next)
	    home = i_GetNextEntity(home);
	else {
	    while (home && !home->next)
		home = i_GetParentEntity(home);
	    if (!home) return 0;
	    home = i_GetNextEntity(home);
	}
	if (target = SearchForward(home,type,method,name))
	    return target;
    }
    return 0;
}

Dll iEntityPtr APIENTRY
i_FindEntityBack(iEntityPtr home, iTypeIndex type,
		 iSearchType method, void *name)
{
    iEntityPtr target;
	
    for (;;) {
	while (!home->prev) {
	    home = i_GetParentEntity(home);
	    if (!home) return 0;
	    if (IsHit(home,type,method,name))
		return home;
	}
	home = i_GetPrevEntity(home);
	if (target = SearchBack(home,type,method,name))
	    return target;
    }
    return 0;
}

void MatrCallBack(iMess message, void *userData, iEntityPtr entity)
{
	int4 i;
	
	if (entity->data->use>1) {
		for(i=0; i<entity->data->use; i++)
			gCallBack(message, userData, entity->data->matrs[i]);
	} else
		gCallBack(message, userData, entity);
}		

static void 
AddAtEnd(iEntityPtr parent,iEntityPtr np) 
{
	iEntityPtr e;
	
	if (parent)
		np->prdata = parent->data;
	else
		np->prdata = 0;
	np->next = 0;
	np->prev = 0;
	if (!parent) return;
	if (!parent->data->child) {
		parent->data->child = np;
	} else {
		e = parent->data->last;
		e->next = np;
		np->prev = e;
	}
	parent->data->last = np;
}
static void AddBefore(iEntityPtr sibling,iEntityPtr np) 
{
	iDataPtr data;
	
	data = np->prdata = sibling->prdata;
	np->next = sibling;
	np->prev = sibling->prev;
	if (np->prev)
		np->prev->next = np;
	sibling->prev = np;
	if (!np->prev && data)
		data->child = np;
}

static void 
AddAfter(iEntityPtr sibling,iEntityPtr np) 
{
	iDataPtr data;
	
	data = np->prdata = sibling->prdata;
	np->next = sibling->next;
	np->prev = sibling;
	if (np->next)
		np->next->prev = np;
	sibling->next = np;
	if (!np->next && data)
		data->last = np;
}

static void 
AddAtFront(iEntityPtr parent,iEntityPtr np)
{
	iEntityPtr e;
	
	if (parent&&parent->data)
		np->prdata = parent->data;
	else
		return;
	np->next = 0;
	np->prev = 0;
	if (!parent->data->child) {
		parent->data->last = np;
	} else {
		e = parent->data->child;
		e->prev = np;
		np->next = e;
	}
	parent->data->child = np;
}

static void 
Unhook(iEntityPtr np)
{
	iEntityPtr p,n;
	
	n = np->next;
	p = np->prev;
	
	if (p) p->next = n;
	else {
		if (np->prdata)
			np->prdata->child = n;
	}
	if (n) n->prev = p;
	else {
		if (np->prdata)
			np->prdata->last = p;
	}
	np->next = 0;
	np->prev = 0;
	np->prdata = 0;
}

static iEntityPtr i_NewEntityOnly(void)
{
	iEntityPtr np;
	
	np = (iEntityPtr) smalloc(sizeof(iEntity));
	if (!np)
		return 0;
	np->data = 0;
	np->whichpos=0;
	np->num = 0;
	
	return np;
}

static iDAFSError
InitData(iDataPtr dp, iEntityPtr head) 
{
	iBox box={0,0,0,0};
	
	dp->matrs = (iEntityPtr*)smalloc(sizeof(iEntityPtr));
	if(!dp->matrs)
		return MallocFailError;
	dp->matrs[0] = head;
	dp->area.a_type=iABox;
	dp->area.a.box=box;
	dp->ct = iAnd;
	dp->child = 0;
	dp->last = 0;
	dp->id = 0;
	dp->use = 1;
	
	dp->userPtr = 0;
	dp->props = 0;
	dp->dt = iDNo;
	dp->d.at.len = 0;
	dp->d.at.text = 0;

	return DAFSOK;
}

Dll iEntityPtr APIENTRY
i_NewEntityBefore(const iTypeIndex type,iEntityPtr sibling)
{
	iEntityPtr np;
	iDataPtr dp, data;
	
	np = i_NewEntityOnly();
	if (!np)
		return 0;
	dp = (iDataPtr) smalloc(sizeof(iData));
	if (!dp) {
		smfree(np);
		return 0;
	}
	if(InitData(dp,np)){
		smfree(dp);
		smfree(np);
		return 0;
	}
	np->data = dp;
	AddBefore(sibling, np);	
	data = sibling->prdata;
	if (data) {
		dp->im = data->im;
		if (dp->im) dp->im->use++;
	} else {
		dp->im = 0;
	}
	dp->type = type;
	
	if (gCallBack) gCallBack(mCreate,gUserData,np);
	return np;
}

Dll iEntityPtr APIENTRY
i_NewEntityAfter(const iTypeIndex type,iEntityPtr sibling)
{
	iEntityPtr np;
	iDataPtr dp, data;
	
	np = i_NewEntityOnly();
	if (!np)
		return 0;
	dp = (iDataPtr) smalloc(sizeof(iData));
	if (!dp) {
		smfree(np);
		return 0;
	}
	if(InitData(dp,np)){
		smfree(dp);
		smfree(np);
		return 0;
	}
	np->data = dp;
	AddAfter(sibling, np);
	data = sibling->prdata;
	if (data) {
		dp->im = data->im;
		if (dp->im) dp->im->use++;
	} else {
		dp->im = 0;
	}
	dp->type = type;
	
	if (gCallBack) gCallBack(mCreate,gUserData,np);
	return np;
}

Dll iEntityPtr APIENTRY
i_NewEntity(const iTypeIndex type,iEntityPtr parent)
{
	iEntityPtr np;
	iDataPtr dp;
	
	np = i_NewEntityOnly();
	if (!np)
		return 0;
	dp = (iDataPtr) smalloc(sizeof(iData));
	if (!dp) {
		smfree(np);
		return 0;
	}
	if(InitData(dp, np)){
		smfree(dp);
		smfree(np);
		return 0;
	}
	np->data = dp;
	AddAtEnd(parent, np);
	if (parent) {
		dp->im = parent->data->im;
		if (dp->im) 
			dp->im->use++;
	} else {
		dp->im = 0;
	}
	dp->type = type;
	
	if (gCallBack) gCallBack(mCreate,gUserData,np);
	return np;
}


Dll int4 APIENTRY
i_CountShares(iEntityPtr en)
{
	if (en&&en->prdata)
		return en->data->use;
	return 0;
}

Dll int4 APIENTRY
i_CountChildren(iEntityPtr en)
{
	int4 i;
	iEntityPtr cur;
	
	i = 0;
	for (cur = i_GetChildEntity(en);cur;cur = i_GetNextEntity(cur)) i++;
	return i;
}

Dll int4 APIENTRY
i_CountPrev(iEntityPtr cur)
{
	int4 i;
	
	i = -1;
	for (;cur;cur = i_GetPrevEntity(cur)) i++;
	return i;
}

Dll int4 APIENTRY
i_CountNext(iEntityPtr cur)
{
	int4 i;
	
	i = -1;
	for (;cur;cur = i_GetNextEntity(cur)) i++;
	return i;
}

Dll iDAFSError APIENTRY
i_AddPosLimit(iEntityPtr entity, int2 pos)
{
	int i;
	int2 *temp;

	for (i=0;i<entity->num;i++) {
		if (entity->whichpos[i] == pos)
			return DAFSOK;  /* all ready in array */
	}
	if (gCallBack)
		MatrCallBack(mWillUpdate,gUserData,entity);

	if (entity->num) 
		temp = (int2*)smrealloc(entity->whichpos, sizeof(int2)*(++entity->num));
	else{ 
		temp = (int2*)smalloc(sizeof(int2));
		entity->num=1;
	}
	if (!temp) {
		entity->num=0;
		return MallocFailError;
	}
	entity->whichpos = temp;
	entity->whichpos[entity->num-1] = pos;

	if (gCallBack)
		MatrCallBack(mUpdate,gUserData,entity);
	return DAFSOK;
}

Dll iDAFSError APIENTRY
i_SetPosLimit(iEntityPtr entity, int2 *pos,int2 size)
{
	int i;
	iDAFSError e;
	
	for(i=0;i<size;i++)
		if(e=i_AddPosLimit(entity,pos[i]))
			return e;
	return DAFSOK;
}

Dll iDAFSError APIENTRY 
i_RemPosLimit(iEntityPtr entity, int2 index) 
{
	iDAFSError err = DAFSOK;

	if (gCallBack)
		MatrCallBack(mWillUpdate,gUserData,entity);

	if (index>=entity->num)
		return OutOfBoundsError;

	while(++index<entity->num)
		entity->whichpos[index-1] = entity->whichpos[index];
	if (entity->num==1) {
		smfree(entity->whichpos);
		entity->whichpos = 0;
		entity->num = 0;
	} else entity->num--;

	if (gCallBack)
		MatrCallBack(mUpdate,gUserData,entity);
	return err;
}

Dll int2 * APIENTRY
i_GetPosLimit(iEntityPtr entity, int2 *size)
{	
	if (entity && entity->num){
		if (size) *size = entity->num;
		return entity->whichpos;
	}
	if (size) *size=0;
	return 0;
}

Dll iDAFSError APIENTRY
i_TrimPosLimit(iEntityPtr entity)
{	
	int2 i,j, found=0;
	
	if (!entity) return -1;
	if (!entity->num||(entity->data->use!=1)) return DAFSOK;
	switch (entity->data->dt) {
	case iDPosSet:
		for (i=0;i<entity->data->d.pp.len;i++){
			for (j=0;j<entity->num;j++)
				if (i == (entity->whichpos[j] - found))
					break;
			if (j==entity->num) {
				found++;
				entity->data->d.pp.len--;
				for (j=i;j<entity->data->d.pp.len;j++)
					entity->data->d.pp.posp[j] = entity->data->d.pp.posp[j+1];
				i--;
			}
		}
		if (entity->num != entity->data->d.pp.len)
			return -1;
		break;
		
	case iD1PosSet:
		
		if (entity->num != 1) return -1;
		if (entity->whichpos[0] != 0) return -1;
		break;
		
	default:
		
		return -1;
		break;	
	}
	i_NoPosLimit(entity);
	return DAFSOK;
}	

Dll void APIENTRY
i_NoPosLimit(iEntityPtr entity)
{
	if (entity->num) {
		smfree(entity->whichpos);
		entity->num=0;
	}
}
	
Dll iPosSetPtr APIENTRY
i_GetTrimPosSet(iEntityPtr cur,iPosSetPtr temp,int2 *vp, unsigned short thresh)
{
	int2 valid, lval;
	int2 i,j,nv;
	int2 *wp;
	iPosSetPtr ps;
	
	ps = i_GetPosSet(cur,&valid);
	if(!valid){
		*vp=0;
		return temp;
	}
	wp = i_GetPosLimit(cur,&lval);
	nv = 0;
	if (!wp) {
		*vp=0;
		for (i=0;i<valid;i++) {
			if (!(ps[i].conf<thresh))
				temp[(*vp)++] = ps[i];
		}
		return temp;
	} 
	else {
		for (i=0;i<valid;i++) {
			for (j=0;j<lval;j++){ 
				if (wp[j] == i)
					break;
			}
			if (j!=lval)
				if (!(ps[i].conf<thresh))
					temp[nv++] = ps[wp[j]];
		}
	}
	*vp = nv;
	return temp;
}

Dll short APIENTRY
i_GetLimitGlyph(iEntityPtr cur)
{
	int2 valid, lval;
	int  i, j;
	int2 *wp;
	iPosSetPtr ps;
	
	ps = i_GetPosSet(cur,&valid);
	if(!valid){
		return 0;
	}
	wp = i_GetPosLimit(cur,&lval);
	if (!wp) {
		return ps[0].chr;
	}
	for (i=0;i<valid;i++) {
		for (j=0;j<lval;j++){ 
			if (wp[j] == i)
				break;
		}
		if (j!=lval)
			return ps[i].chr;
	}
	return 0;
}

static iDAFSError 
iSetTopHeader(iDataPtr data, iEntityPtr head)
{
	int4 use, i;
	iEntityPtr *matrs;
	
	if (!data||!head)
		return -1;
	if (data->use > 1){
		matrs = data->matrs;
		if (matrs[0]!= head){
			use = data->use;
			for (i=1; i<use; i++)
				if (matrs[i] == head){
					matrs[i]=matrs[0];
					matrs[0]=head;
				}
		}
	}
	return DAFSOK;
}

Dll iEntityPtr APIENTRY
i_GetNextEntity(const iEntityPtr entity)
{
	if (entity)
		/*if (entity->next&&entity->next->data->use>1){
		  if (iSetTopHeader(entity->next->data,entity->next))
		  return 0; 
		  return entity->next;
		  }
		  else*/
		return entity->next;
	else
		return 0;
}

Dll iEntityPtr APIENTRY
i_GetPrevEntity(const iEntityPtr entity)
{
	if (entity)
		/*if (entity->prev&&entity->prev->data->use>1) {
		  if (iSetTopHeader(entity->prev->data,entity->prev))
		  return 0; 
		  return entity->prev;
		  }
		  else*/
		return entity->prev;
	else
		return 0;
}

Dll iEntityPtr APIENTRY
i_GetChildEntity(const iEntityPtr entity)
{
	if (!entity)
		return 0;
	if (entity->data&&entity->data->child){
		if (iSetTopHeader(entity->data, entity))
			return 0;
		return entity->data->child;
	}
	else
		return 0;
}

Dll iEntityPtr APIENTRY
i_GetLastEntity(const iEntityPtr entity)
{
	if (!entity)
		return 0;
	if (entity->data&&entity->data->child){
		if (iSetTopHeader(entity->data, entity))
			return 0;
		return entity->data->last;
	}
	else
		return 0;
}

Dll iEntityPtr APIENTRY
i_GetParentEntity(const iEntityPtr entity)
{
	if (entity&&entity->prdata)
		return entity->prdata->matrs[0];
	
	return 0;
}

static int4 lookup(iEntityPtr home,iEntityPtr target)
{
	int4 i;
	
	home = i_GetParentEntity(home);
	if (!home) return 0;
	if (home == target) return 1;
	if (home->data->use == 1) return lookup(home,target);
	for(i=0; i<home->data->use; i++){
		if(lookup(home->data->matrs[i],target))
			return 1;
	}
	return 0;
}

Dll iEntityPtr APIENTRY
i_GetParentOtherPath(iEntityPtr home,iEntityPtr target)
{
	int4 i;
	
	home = i_GetParentEntity(home);
	if (!home) return 0;
	if (home->data->use == 1) return home;
	for(i=0; i<home->data->use; i++){
		if(!lookup(home->data->matrs[i],target))
			return home->data->matrs[i];
	}
	return 0;
}

Dll iEntityPtr APIENTRY
i_GetParentPath(iEntityPtr home,iEntityPtr target)
{
	int4 i;
	
	home = i_GetParentEntity(home);
	if (!home) return 0;
	if (home == target || home->data->use == 1) return home;
	for(i=0; i<home->data->use; i++){
		if(lookup(home->data->matrs[i],target))
			return home->data->matrs[i];
	}
	return 0;
}

static int4 
CountMe(iEntityPtr en,iDataPtr dp)
{
	int4 count;
	iEntityPtr cp;
	
	count = en->data == dp;
	for (cp = i_GetChildEntity(en);cp;cp = i_GetNextEntity(cp))
		count += CountMe(cp,dp);
	return count;
}

Dll void APIENTRY
i_SetLogicalType(iEntityPtr entity, iLogicalType type)
{
	if (!entity || !entity->data) return;
	entity->data->ct = type;
}

Dll iLogicalType APIENTRY
i_GetLogicalType(iEntityPtr entity)
{
	if (entity && entity->data) return entity->data->ct;
	return -1;
}

Dll void APIENTRY
i_DisposeEntity(iEntityPtr entity)
{
	int4 i;
	
	Unhook(entity);
	
	i_NoPosLimit(entity);

	if(entity->data){
		
		if (!(--entity->data->use)) {
			DisposeEntityData(entity);
		} else {
			for (i=0;i<entity->data->use+1;i++)
				if (entity == entity->data->matrs[i])
					break;
			for (;i<entity->data->use;i++)
				entity->data->matrs[i] = entity->data->matrs[i+1];
			if (entity->data->use == 0) {
				smfree(entity->data->matrs);
				entity->data->matrs = 0;
			} /* else there is no need to smrealloc matrs to be smaller */
		}
	}
	smfree(entity);
}

Dll iDAFSError APIENTRY
i_ShareEntity(iEntityPtr entity, iEntityPtr parent, iEntityPtr * epp)
{
	iEntityPtr new,*temp;
	iDataPtr dp;
	
	if(!parent||!entity||!entity->data) return NotEnoughData;
	
	new=i_NewEntityOnly();
	if(!new) return MallocFailError;
	AddAtEnd(parent, new);
	
	dp=entity->data;
	dp->use++;
	temp = (iEntityPtr*)smrealloc(dp->matrs,sizeof(iEntityPtr) * dp->use);
	if(!temp)
		return MallocFailError;
	dp->matrs = temp;
	dp->matrs[dp->use-1] = new;
	
	new->data=dp;
	
	if (gCallBack) gCallBack(mCreate,gUserData,new);
	if(epp) *epp=new;
	return DAFSOK;
}

Dll void APIENTRY
i_SetImage(iEntityPtr entity, iImagePtr image)
{
	if (!entity||(entity->data->im==image))
		return;
	if (gCallBack)
		MatrCallBack(mWillUpdate,gUserData,entity);
	if (entity->data->im && entity->data->use == 1)
		i_DisposeImage(entity->data->im);
	if (image) image->use+=entity->data->use;
	entity->data->im = image;
	if (gCallBack)
		MatrCallBack(mUpdate,gUserData,entity);
}

Dll iImagePtr APIENTRY
i_GetImage(const iEntityPtr entity)
{
	if (!entity)
		return 0;
	return entity->data->im;
}

Dll iTypeIndex APIENTRY
i_GetEntityType(const iEntityPtr entity)
{
	if (!entity)
		return 0;
	return entity->data->type;
}

Dll void APIENTRY
i_SetEntityType(iEntityPtr entity,iTypeIndex type)
{
	if (!entity)
		return;
	if (gCallBack)
		MatrCallBack(mWillUpdate,gUserData,entity);
	entity->data->type = type;
	if (gCallBack)
		MatrCallBack(mUpdate,gUserData,entity);
}

Dll iOrthogonPtr APIENTRY
i_NewOrthogon(int2 n)
{	
	iOrthogonPtr poly;
	
	poly = (iOrthogonPtr)smalloc(sizeof(iOrthogon));
	if (!poly) return 0;
	
	poly->n=n;
	poly->point=0;
	if (n) {
		poly->point = (iPoint*)smalloc(sizeof(iPoint)*n);
		if (!poly->point) {
			smfree(poly);
			return 0;
		}
	}
	return poly;
}

Dll iDAFSError APIENTRY
i_ForceAreaType(const iEntityPtr entity, iATag tag)
{	
	iDAFSError err=DAFSOK;
	iBox box;
	iOrthogonPtr poly=0;
	iBoxListPtr BoxList;

	if (entity&&entity->data){
		if (entity->data->area.a_type!=tag){
			switch(entity->data->area.a_type){
			
			case iABox:	
				if (tag==iAOrthogonList){
					if (entity->data->area.a.box.w) {
						if(err = i_Box2Orthogon(entity->data->area.a.box, &poly))
							goto error;
					}
					entity->data->area.a.brdr=i_NewOrthogonList();
					if (!entity->data->area.a.brdr) { err=MallocFailError; goto error;}

					entity->data->area.a_type=iAOrthogonList;
					if (poly) {
						if (err=i_AppendOrthogon(entity, poly))
							goto error;
						poly=0;
					}
				}
				else{  /* convert to BoxList */
					box=entity->data->area.a.box;
					entity->data->area.a_type=iABoxList;
					if(err=i_Box2BoxList(box, &entity->data->area.a.box_list))
						goto error;
				}
				break;
			case iABoxList:
				if (tag==iABox){
					if (err=i_SetBox(entity, i_GetEnclosingBox(entity)))
						goto error;
				}
				else{ /* convert to border */
					BoxList=entity->data->area.a.box_list;
					entity->data->area.a_type=iAOrthogonList;
					if (err = i_BoxList2OrthogonList(BoxList,&(entity->data->area.a.brdr)))
						goto error;
					i_DisposeBoxList(BoxList);
				}
				break;
			case iAOrthogonList:
				if (tag==iABox){
					if (err=i_SetBox(entity, i_GetEnclosingBox(entity)))
						goto error;
				}
				else{  /* convert to BoxList */
					if(err = i_OrthogonList2BoxList(entity->data->area.a.brdr, &BoxList))
						goto error;
					entity->data->area.a_type = iABoxList;
					i_DisposeOrthogonList(entity->data->area.a.brdr);
					entity->data->area.a.box_list = BoxList;
				}
			}
		}
	}
	else{
		err=BadEntityPassed;
	}
error:
	if (poly) i_DisposeOrthogon(poly);
	return err;
}

Dll iDAFSError APIENTRY
i_GetArea(const iEntityPtr entity, void **data, iATag *tag)
{
	iDAFSError err=DAFSOK;
	if (entity&&entity->data){
		*tag=entity->data->area.a_type;
		switch(*tag){
		case iABox:		
			*data=(void*)&entity->data->area.a.box;
			break;
		case iAOrthogonList:	
			*data=(void*)entity->data->area.a.brdr;
			break;
		case iABoxList:	
			*data=(void*)entity->data->area.a.box_list;
			break;
		}
	}
	else
		err=NotEnoughData;
	return err;
}

Dll iOrthogonPtr APIENTRY
i_CopyOrthogon(iOrthogonPtr shape)
{	int4 i;
	iOrthogonPtr poly=0;

	if (shape->n==2){
		poly = i_NewOrthogon(4);
		if (!poly)
			return 0;
		poly->point[0].x = poly->point[3].x = shape->point[0].x;
		poly->point[0].y = poly->point[1].y = shape->point[0].y;
		poly->point[1].x = poly->point[2].x = shape->point[1].x+shape->point[0].x;
		poly->point[2].y = poly->point[3].y = shape->point[1].y+shape->point[0].y;
	} else {
		poly = i_NewOrthogon(shape->n);
		if (!poly)	
			return 0;
		for (i=0; i<shape->n; i++){
			poly->point[i].x = shape->point[i].x;
			poly->point[i].y = shape->point[i].y;
		}
	}
	return poly;
}

Dll void APIENTRY
i_TranslateOrthogon(iOrthogonPtr poly, short x, short y)
{	int4 n;

	if (poly){
		n=poly->n;
		if (n==2)
			n--;
		while(n--){
			poly->point[n].x += x;
			poly->point[n].y += y;
		}
	}
}
		
Dll iDAFSError APIENTRY
i_SetOrthogonList(iEntityPtr entity, iOrthogonListPtr border)
{
	if (!border)
		halt("NULL orthgon list passed to i_SetOrthogonList\n");

	if (entity && entity->data){		
		iOrthogonListPtr copy = i_CopyOrthogonList(border);
		if(copy){
			if (gCallBack)
				MatrCallBack(mWillMove,gUserData,entity);
			i_DisposeArea(entity);
			entity->data->area.a_type = iAOrthogonList;
			entity->data->area.a.brdr = copy;
			if (gCallBack)
				MatrCallBack(mMove,gUserData,entity);
			return DAFSOK;
		}
		else
			return MallocFailError;
	}
	else
		return NotEnoughData;
}

Dll iDAFSError APIENTRY
i_SetBoxList(iEntityPtr entity, iBoxListPtr box_list)
{
	if (!box_list)
		halt("NULL box list passed to i_SetBoxList\n");

	if (entity && entity->data){
		iBoxListPtr copy = i_CopyBoxList(box_list);
		if(copy){			
			if (gCallBack)
				MatrCallBack(mWillMove,gUserData,entity);
			i_DisposeArea(entity);
			entity->data->area.a_type = iABoxList;		
			entity->data->area.a.box_list = copy;
			if (gCallBack)
				MatrCallBack(mMove,gUserData,entity);
			return DAFSOK;
		}
		else
			return MallocFailError;
	}
	else	   
		return NotEnoughData;
}

iDAFSError
SetBoundsType(iEntityPtr entity, iATag tag)
{
	iDAFSError err=DAFSOK;
	
	if(entity->data->area.a_type!=tag){
		switch(entity->data->area.a_type){
		case iAOrthogonList:	
			i_DisposeOrthogonList(entity->data->area.a.brdr);
			break;
		case iABoxList:	
			i_DisposeBoxList(entity->data->area.a.box_list);
			break;
		case iABox:
		default:			
			break;
		}
		
		switch(tag){
		
		case iAOrthogonList:	
			entity->data->area.a.brdr=i_NewOrthogonList();
			if (!entity->data->area.a.brdr)
				err=MallocFailError;
			break;
		case iABoxList:	
			entity->data->area.a.box_list=i_NewBoxList();
			if (!entity->data->area.a.box_list)
				err=MallocFailError;
			break;
		case iABox:
		default:			
			break;
		}
		entity->data->area.a_type=tag;
	}
	return err;
}	

iDAFSError
i_AppendOrthogon(iEntityPtr entity, iOrthogonPtr ortho)
{
	iOrthogonListPtr border;
	iDAFSError err=DAFSOK;
	iATag tag;

	if (entity&&entity->data&&ortho){
		if (err=i_ForceAreaType(entity, iAOrthogonList))
			return err;
		if (err=i_GetArea(entity, (void **)&border, &tag))
			return err;
		if (err=i_AddOrthogon2OrthogonList(border,ortho))
			return err;
	} else
		halt("i_AppendOrthogon: not enough data\n");

	return err;
}

Dll iDAFSError APIENTRY
i_AddOrthogon2OrthogonList(iOrthogonListPtr border, iOrthogonPtr poly)
{
	iOrthogonArrayPtr temp;

	if (border->num_poly++)
		temp = (iOrthogonArrayPtr)smrealloc(border->poly,sizeof(iOrthogonArrayPtr)*border->num_poly);
	else
		temp = (iOrthogonPtr*)smalloc(sizeof(iOrthogonPtr)*border->num_poly);
	if (!temp)
		return MallocFailError;
	border->poly = temp;
	border->poly[border->num_poly-1] = poly;
	return DAFSOK;
}

Dll void APIENTRY
i_DisposeOrthogon(iOrthogonPtr poly)
{
	if (poly){
		if (poly->point)
			smfree(poly->point);
		poly->n=0;
		smfree(poly);
	}
}

Dll iOrthogonListPtr APIENTRY
i_CopyOrthogonList(iOrthogonListPtr original)
{	iOrthogonListPtr copy=0;
	int4 i;
	if (original){
		copy = i_NewOrthogonList();
		if (copy){
			copy->num_poly=original->num_poly;
			copy->poly = (iOrthogonArrayPtr)smalloc(sizeof(iOrthogonPtr)*original->num_poly);
			for (i=0; i<copy->num_poly; i++){
				copy->poly[i]=i_CopyOrthogon(original->poly[i]);
			}
		}
	}
	return copy;
}

Dll iOrthogonListPtr APIENTRY
i_NewOrthogonList(void)
{
	iOrthogonListPtr border;
	
	border=(iOrthogonListPtr)smalloc(sizeof(iOrthogonList));
	if (border){
		border->num_poly=0;
		border->poly=0;
	}
	return border;
}	

Dll void APIENTRY
i_DisposeOrthogonList(iOrthogonListPtr border)
{	int4 i;
	if (border&&border->num_poly){
		for(i=0; i<border->num_poly; i++)
			i_DisposeOrthogon(border->poly[i]);
		smfree(border->poly);
		border->num_poly=0;
	}
	if (border)
		smfree(border);
	border=0;
}

void
i_DisposeArea(iEntityPtr entity)
{
	switch(entity->data->area.a_type){
	
	case iAOrthogonList:
			i_DisposeOrthogonList(entity->data->area.a.brdr);
			entity->data->area.a_type=iABox;
			break;
	case iABoxList:
			i_DisposeBoxList(entity->data->area.a.box_list);
			break;
	case iABox:
	default:
			break;
	}
}

Dll iBox APIENTRY
i_GetEnclosingBox(const iEntityPtr entity)
{
	iBox bx;
	BOX extents;

	if (entity)
		switch(entity->data->area.a_type){
			case iABox:		
				bx=entity->data->area.a.box;
				break;
			case iAOrthogonList:	
				i_OrthogonList2Box(entity->data->area.a.brdr, &bx);
				break;
			case iABoxList:	
				extents=entity->data->area.a.box_list->extents;
				bx.x=extents.x1;
				bx.y=extents.y1;
				bx.w=ABS(extents.x1-extents.x2);
				bx.h=ABS(extents.y1-extents.y2);
				break;
		}
	return bx;
}

Dll iDAFSError APIENTRY
i_SetBox(iEntityPtr entity, iBox bounds)
{
	iDAFSError err = DAFSOK;

	if (entity && entity->data){
		if (gCallBack)
			MatrCallBack(mWillMove,gUserData,entity);
		i_DisposeArea(entity);
		entity->data->area.a_type=iABox;
		entity->data->area.a.box=bounds;
		if (gCallBack)
			MatrCallBack(mMove,gUserData,entity);
	} else
		err=BadEntityPassed;
	return err;
}

void
i_SwapUPoly(void* border) {
	iOrthogonPtr new;
	if (border) {
		new = (iOrthogonPtr)border;
		i_SwapULong(&new->n);
		i_SwapXPoints(new->point,new->n);
	}
}

Dll void APIENTRY
i_SwapUBox(void* box) {
	iBox *new;
	
	if (box){
		new = (iBox*)box;
		i_SwapBox(new); 
	}
}

Dll void APIENTRY
i_SwapULBox(void* box) {
	iLBox *new;
	if (box){
		new = (iLBox*)box;
		i_SwapLBox(new);
	}
}

Dll void APIENTRY
i_SwapULong(void* void_ptr) {
	long *var;
	if (void_ptr){
		var = (long*)void_ptr;
		*var = i_SwapL(*var);
	}
}

Dll void APIENTRY
i_SwapUShort(void* void_ptr) {
	short *var;
	if (void_ptr){
		var = (short*)void_ptr;
		*var = i_SwapS(*var);
	}
}

Dll void APIENTRY
i_MoveToFront(iEntityPtr entity, iEntityPtr new_parent)
{
	iEntityPtr old_parent;
	
	if (!entity)
		return;
	old_parent = i_GetParentEntity(entity);
	if (gCallBack){
		if (old_parent)
			MatrCallBack(mWillUpdate,gUserData,old_parent);
		if (new_parent && old_parent!=new_parent)
			MatrCallBack(mWillUpdate,gUserData,new_parent);
	}
	Unhook(entity);
	if (new_parent)
		AddAtFront(new_parent,entity);
	if (gCallBack){
		if (old_parent)
			MatrCallBack(mUpdate,gUserData,old_parent);
		if (new_parent && i_GetParentEntity(entity)!=new_parent)
			MatrCallBack(mUpdate,gUserData,new_parent);
	}
}

Dll void APIENTRY
i_MoveToBack(iEntityPtr entity, iEntityPtr new_parent)
{
	iEntityPtr old_parent;
	
	old_parent = i_GetParentEntity(entity);
	if (gCallBack){
		if (old_parent)
			MatrCallBack(mWillUpdate,gUserData,old_parent);
		if (new_parent && old_parent!=new_parent)
			MatrCallBack(mWillUpdate,gUserData,new_parent);
	}
	Unhook(entity);
	if (new_parent)
		AddAtEnd(new_parent,entity);
	if (gCallBack){
		if (old_parent)
			MatrCallBack(mUpdate,gUserData,old_parent);
		if (new_parent && i_GetParentEntity(entity)!=new_parent)
			MatrCallBack(mUpdate,gUserData,new_parent);
	}
}

Dll void APIENTRY
i_MoveAfter(iEntityPtr entity, iEntityPtr PrevSibling)
{	
	iEntityPtr old_parent;
	
	old_parent = i_GetParentEntity(entity);
	if (entity == PrevSibling)
		return;
	
	if (PrevSibling && PrevSibling->next == entity) 
	    return;
	
	if (gCallBack){
		if (PrevSibling && PrevSibling->prdata)
			MatrCallBack(mWillUpdate,gUserData,i_GetParentEntity(PrevSibling));
		if (entity->prdata)
			MatrCallBack(mWillUpdate,gUserData,i_GetParentEntity(entity));
	}
	Unhook(entity);
	if (PrevSibling && PrevSibling->prdata) {
		if (entity->next = PrevSibling->next)
			entity->next->prev = entity;
		else PrevSibling->prdata->last = entity;
		entity->prev = PrevSibling;
		PrevSibling->next = entity;
		entity->prdata = PrevSibling->prdata;
	}
	if (gCallBack){
		if (PrevSibling && PrevSibling->prdata)
			MatrCallBack(mUpdate,gUserData,i_GetParentEntity(PrevSibling));
		if (old_parent)
			MatrCallBack(mUpdate,gUserData,old_parent);
	}
}

Dll void APIENTRY
i_MoveBefore(iEntityPtr entity, iEntityPtr NextSibling)
{	
	iEntityPtr old_parent;
	
	old_parent = i_GetParentEntity(entity);
	
	if (NextSibling && NextSibling->prev==entity) 
	    return;
	
	if (gCallBack){
		if (NextSibling && NextSibling->prdata)
			MatrCallBack(mWillUpdate,gUserData,i_GetParentEntity(NextSibling));
		if (entity->prdata)
			MatrCallBack(mWillUpdate,gUserData,i_GetParentEntity(entity));
	}
	Unhook(entity);
	if (NextSibling && NextSibling->prdata) {
		if (entity->prev = NextSibling->prev)
			entity->prev->next = entity;
		else NextSibling->prdata->child = entity;
		entity->next = NextSibling;
		NextSibling->prev = entity;
		entity->prdata = NextSibling->prdata;
	}
	if (gCallBack){
		if (NextSibling && NextSibling->prdata)
			MatrCallBack(mUpdate,gUserData,i_GetParentEntity(NextSibling));
		if (old_parent)
			MatrCallBack(mUpdate,gUserData,old_parent);
	}
}

Dll iDAFSError APIENTRY
i_OrEntity(iEntityPtr *head_entity)
{
	iEntityPtr ne, en;
	iCallBackProc save;
	
	en = *head_entity;
#if 0
	if (i_GetLogicalType(en) != iAnd) 
		return AlreadyOrEntity;
#endif
	if (gCallBack)
		gCallBack(mWillUpdate,gUserData,en);
	if (gCallBack) save = gCallBack;
	else save = 0;
	gCallBack = 0;
	ne = i_NewEntity(en->data->type,i_GetParentEntity(en));
	if (!ne) return MallocFailError;
	i_MoveAfter(ne,en);
	i_SetLogicalType(ne, iOr );
	i_MoveToBack(en,ne);
	
	if (save) {
		gCallBack = save;
		gCallBack(mUpdate,gUserData,en);
	}
	*head_entity = ne;
	return DAFSOK;
}

Dll void APIENTRY
i_GrowOrthogon(iOrthogonPtr poly, iOrthogonPtr npoly, short size)
{	
	int4 i;
	short oldx, oldy, nextx, nexty, x, y;

	if (!poly || poly->n<4){
		npoly->n=0;
		return;
	}
	npoly->n = poly->n;
	oldx = poly->point[0].x;
	oldy = poly->point[0].y;

	for (i=1; i<poly->n-1; i++){
		nextx = poly->point[i+1].x;
		nexty = poly->point[i+1].y;
		x =poly->point[i].x;
		y =poly->point[i].y;
		if (y<=nexty && y>=oldy)
			npoly->point[i].x = x+size-1;
		else
			npoly->point[i].x = x-size;
		if (x>=nextx && x<=oldx)
			npoly->point[i].y = y+size-1; 
		else
			npoly->point[i].y = y-size;
		oldx = poly->point[i].x;
		oldy = poly->point[i].y;
	}
	nextx = poly->point[1].x;
	nexty = poly->point[1].y;
	oldx = poly->point[poly->n-1].x;
	oldy = poly->point[poly->n-1].y;
	x =poly->point[0].x;
	y =poly->point[0].y;
	if (y<=nexty && y>=oldy)
		npoly->point[0].x = x+size-1; 
	else
		npoly->point[0].x = x-size;
	if (x>=nextx && x<=oldx)
		npoly->point[0].y = y+size-1; 
	else
		npoly->point[0].y = y-size;
	nextx = poly->point[0].x;
	nexty = poly->point[0].y;
	oldx = poly->point[poly->n-2].x;
	oldy = poly->point[poly->n-2].y;
	x =poly->point[poly->n-1].x;
	y =poly->point[poly->n-1].y;
	if (y<=nexty && y>=oldy)
		npoly->point[poly->n-1].x = x+size-1; 
	else
		npoly->point[poly->n-1].x = x-size;
	if (x>=nextx && x<=oldx)
		npoly->point[poly->n-1].y = y +size-1; 
	else
		npoly->point[poly->n-1].y = y-size;
}


iEntityPtr 
i_GetShares(iEntityPtr entity, iEntityPtr prev_borrow)
{
	int i;

	if(prev_borrow == 0)
		return entity->data->matrs[0];
	for(i = 0; i < entity->data->use; i++)
		if(entity->data->matrs[i] == prev_borrow)
			break;
	if(i == entity->data->use)
		halt("bad data supplied to i_GetShares.\n");
	if(i == entity->data->use - 1)
		return 0;
	return entity->data->matrs[i+1];
}

/*
// basically, the only error that can occur is that
// we run out of memory
*/
Dll iDAFSError APIENTRY
i_CopyOneEntity(iEntityPtr orig, iEntityPtr parent_of_copy, 
			 iEntityPtr *new_entity_ptr)
{
  iEntityPtr new_entity = 0;
  iDAFSError err = DAFSOK;
  int2 val;

  if(orig == 0)
	halt("i_CopyEntity: invalid input data.\n");

  new_entity = i_NewEntity(i_GetEntityType(orig), parent_of_copy);

  if(new_entity == 0)
	err = MallocFailError;
  else {
	short conf;
	char *text;
	iUCode *utext;
	short short_val;
	iPosSetPtr pos_set = 0;
	int2 n;
	int2 glyph_val;
	void *data;
	iATag tag;
	iBox box;
	iBoxListPtr box_list;
	iOrthogonListPtr ortho_list;

	i_SetLogicalType(new_entity, i_GetLogicalType(orig));
	/* bogus */
	i_GetArea(orig, &data, &tag);
	switch(tag) {
	case iABox:
	  box = i_GetEnclosingBox(orig);
	  i_SetBox(new_entity, box);
	  break;
	case iAOrthogonList:
	  ortho_list = (iOrthogonListPtr) data;
	  err = i_SetOrthogonList(new_entity, ortho_list);
	  break;
	case iABoxList:
	  box_list = (iBoxListPtr) data;
	  err = i_SetBoxList(new_entity, data);
	  break;
	default:
	  halt("i_CopyEntity: unknown area type.\n");
	}

	i_SetImage(new_entity, i_GetImage(orig));

	/* bogus -- what is userPtr for these days -- anything */
	/* we can't do anything except copy ? */

	/* copy props */
	if(err == DAFSOK)
	  err = i_CopyAllProps(orig, new_entity);

	if(err == DAFSOK){
	  switch(i_GetEDataType(orig)) {
	  case iDNo:
		break;
	  case iDPosSet:
		pos_set = i_GetPosSet(orig, &n);
		err = i_SetPosSet(new_entity, n, pos_set);
		break;
	  case iDString:
		text = (char*)i_GetText(orig, &conf);
		err = i_SetText(new_entity, text, conf);
		break;
	  case iD1PosSet:
		glyph_val = i_GetGlyph(orig,&conf);
		i_SetGlyph(new_entity, glyph_val, conf);
		break;
	  case iDUserData:
		/* what about this ? */
		break;
	  case iDConf: 
		conf = i_GetConf(orig);
		i_SetConf(new_entity, conf);
		break;
	  case iDShort:
		short_val = i_GetShort(orig, &conf);
		i_SetShort(orig, short_val, conf);
		break;
	  case iDUString:
		utext = (iUCode*)i_GetUText(orig, &conf);
		err = i_SetUText(new_entity, utext, conf);
		break;
	  default:
		halt("i_CopyEntity: unrecognized data type.\n");
	  }
	}
	if(val=i_GetShort(orig,&conf))
		i_SetShort(new_entity,val,conf);
	/* copy data */
  }

  if(new_entity && err != DAFSOK) {
	i_DisposeEntity(new_entity);
	new_entity = 0;
  }

  if (new_entity_ptr) *new_entity_ptr = new_entity;

  return err;
}


Dll iDAFSError APIENTRY
i_CopyEntity(iEntityPtr orig, iEntityPtr parent_of_copy, 
			 iEntityPtr *new_entity_ptr)
{
	iEntityPtr child;
	iEntityPtr new_entity;
	iDAFSError err=DAFSOK;
	
	err=i_CopyOneEntity(orig,parent_of_copy,&new_entity);
	
	for(child = i_GetChildEntity(orig); err == DAFSOK && child;
		child = i_GetNextEntity(child)) {
	  err = i_CopyEntity(child, new_entity, 0);
	}

	if (err && new_entity)
		i_DisposeEntity(new_entity);
	else if (new_entity_ptr)
		*new_entity_ptr=new_entity;

	return err;
}

