/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"
#include "rafutil.h"
#include "ucs.h"
#include "c-lib.h"
#include "smemory.h"
#include "str-lib.h"

#include "dafstype.h"
#include "entity.h"
#include "dafs.h"
#include "utf.h"

static iDAFSError 
fixalloc(iUCode **text,int4 *pos,int4 nl,int4 *len)
{
	iUCode *nt;
	if ((nl + *pos) >= *len) {
		(*len) += nl + 100;
		nt = (iUCode*)smrealloc(*text,sizeof(iUCode) * (*len));
		if (!nt) return MallocFailError;
		*text = nt;
	}
	return DAFSOK;
}

static iDAFSError 
GetAllUText(iEntityPtr en,iUCode **text,int4 *pos,int4 *len,iUCode *nonrec)
{
	iEntityPtr cur;
	iUCode nc;
	char *s;
	iUCode *us;
	iDAFSError e;
	iUCode *pt;
	iUCode c,nr;
	int4 t;
	nr=(nonrec) ? *nonrec : '^';
	if(!i_GetUStringProp(en,"Alias",&us)){
		t = ucslen(us);
		if (e = fixalloc(text,pos,t,len))
			return e;
		pt = (*text) + *pos;
		while (*us) *pt++ = *us++;
		(*pos) += t;
		*pt = 0;
		return DAFSOK;
	}
	t = i_GetEntityType(en);
	if (t == iSpaceT || t == iNewLineT) {
		nc = (t == iSpaceT? ' ': '\n');
		t = i_GetShort(en,0);
		if (e = fixalloc(text,pos,t,len))
			return e;
		pt = (*text) + *pos;
		(*pos)+=t;
		while (t--) *pt++ = nc;
		*pt = 0;
		return DAFSOK;
	}
	cur = i_GetChildEntity(en);
	if (!cur) {
		t = i_GetEDataType(en);
		if (t == iDNo) return DAFSOK;
		if (t == iDString) {
			s=(char*)i_GetText(en,0);
			t=strlen(s);
			if (e = fixalloc(text,pos,t,len))
				return e;
			pt = (*text) + *pos;
			while (*s) *pt++ = *s++;
			(*pos) += t;
			*pt = 0;
			return DAFSOK;
		}
		if (t == iDUString) {
			pt=(iUCode*)i_GetUText(en,0);
			t=ucslen(pt);
			if (e = fixalloc(text,pos,t,len))
				return e;
			ucscat((*text) + *pos,pt);
			(*pos) += t;
			return DAFSOK;
		}
		if (e = fixalloc(text,pos,1,len))
			return e;
		(*text)[(*pos)++] = (c=i_GetGlyph(en,0)) ? c : nr;
		(*text)[*pos] = 0;
		return DAFSOK;
	}
	t = i_GetLogicalType(en) == iAnd;
	for (; cur;cur = i_GetNextEntity(cur)) {
		if (e = GetAllUText(cur,text,pos,len,nonrec))
			return e;
		if (!t) break;
	}
	return DAFSOK;
}

Dll iUCode * APIENTRY
i_AllocUText(iEntityPtr en)
{
	int4 len = 100,pos = 0;
	iUCode *text = (iUCode*)smalloc(len * sizeof(iUCode));
	iDAFSError e;
	
	if (!text) return 0;
	*text = 0;
	e = GetAllUText(en,&text,&pos,&len,0);
	if (e) {
		smfree(text);
		text = 0;
	}
	return text;
}

Dll void APIENTRY
i_FreeUText(iUCode *text)
{
	smfree(text);
}

#define UA_ALLOC 25

/*
// bogus this is odd -- typedefs are defined elsewhere,
// but the structures are defined here
// why bother
*/

struct iUAEntry {
	iUCode index;
	char *string;
};

struct iUATable {
	int4 num;
	iUAEntry *entries;
};


typedef int (*LocalCompare)(const void *, const void *);

static int /* must be int for dos and unix */
tableCmp(iUAEntryPtr a, iUAEntryPtr b)
{
	return a->index - b->index;
}

Dll iDAFSError APIENTRY
i_ReadUATable(iUATablePtr *ua_table, char *filename)
{
	RafFile fp;
	int4 i = 0, 
	     alloc_size = 0;
	char line[256], *p, *string;
	iUATablePtr table;
	iDAFSError e=DAFSOK;

	/* allocate space for the table */
	table = (iUATablePtr) smalloc(sizeof(iUATable));
	if(!table) return MallocFailError;
	*ua_table = table;
	table->entries = NULL;
	table->num = 0;

	if(!(fp = riOpen(filename, "rt")))
		return ReadError;

	/* process the filter table */
	while(!riEof(fp))
		if(riGets(line, sizeof line, fp) && strlen(line) > 2){
			char *r, *s;

			/* add space to the table */
			if(i == alloc_size){
				iUAEntryPtr temp;

				alloc_size += UA_ALLOC;
				if(i)
					temp = smrealloc(table->entries, sizeof(iUAEntry) * alloc_size);
				else
					temp = smalloc(sizeof(iUAEntry) * alloc_size);
				if (!temp) {
					e=MallocFailError;
					goto error;
				} else
					table->entries = temp;
			}

			table->entries[i].string = NULL;

			/* find the end of the number */
			for(r = line; *r != ' ' && *r != '"' && *r != '\t'; r++);
			*r = '\0';
			s = ++r;

			/* take either decimal or hexadecimal */
			switch(line[1]){
			  case 'x':
				  table->entries[i].index = (iUCode)strtol(line, 0, 16);
				  break;
			  default:
				  table->entries[i].index = atoi(line);

			}

			/* eat up the white space */
			while (*s == ' ' || *s == '\t' || *s == '"')
				s++;

			/* copy the string */
			string = (char *) smalloc(strlen(s)+1);
			if(!string){
				e = MallocFailError;
				goto error;
			}			
			strcpy(string, s);
			table->entries[i].string = string;

			/* eliminate last quote if it exists */
			if(p = (char*) strrchr((char*)table->entries[i].string,'"')) *p = (char)'\0';

			i++;
		}
	table->num = i;

	/* use quick sort to put table in order */
	qsort(table->entries, 
		  (size_t) table->num, 
		  (size_t) sizeof (iUAEntry), 
		  (LocalCompare)tableCmp);

error:
	riClose(fp);
	if (e) {
		if(table->entries){
			i = 0;

			while(i < table->num && table->entries[i].string)
				smfree(table->entries[i].string);
			
			smfree(table->entries);
		}
		smfree(table);
		*ua_table=0;	
	}

	return e;
}

Dll void APIENTRY
i_DisposeUATable(iUATablePtr table)
{
	if(table->entries){
		int i;
		
		for(i = 0; i < table->num; i++)
			if(table->entries[i].string)
				smfree(table->entries[i].string);

		smfree(table->entries);
	}
		
	smfree(table);
} 

Dll char * APIENTRY
i_AllocASCIIText(iEntityPtr en, iUATablePtr table)
{
	int4 len = 100,pos = 0,i;
	iUCode *text = (iUCode*)smalloc(len * sizeof(iUCode));
	char *ascii_text,*p;
	iDAFSError e;
	
	if(!en)
		return 0;
	if (!text) return 0;
	*text = 0;
	e = GetAllUText(en,&text,&pos,&len,0);
	if (e) {
		smfree(text);
		return 0;
	}
	ascii_text=smalloc((pos + 1) *sizeof(char));
	if (!ascii_text) {
		smfree(text);
		return 0;
	}
	len = pos;
	/* convert uni to ascii */
	if(table){
		int4 i, j = 0, size = pos;
		iUAEntry key;
		iUAEntryPtr entry; 

		for(i = 0; i < pos; i++){
			key.index = text[i];
			if(entry = (iUAEntryPtr) bsearch(&key, 
											 table->entries, 
											 (size_t) table->num, 
											 (size_t) sizeof(iUAEntry), 
											 (LocalCompare) tableCmp)){
				size += strlen(entry->string);
				if ((j+size) >= len) {
					len += size * 4;
					if(!(p = smrealloc(ascii_text, len+1))) {
						smfree(text);
						smfree(ascii_text);
						return 0;
					}
					ascii_text = p;
				}
				strcpy(ascii_text+j, entry->string);
				j = strlen(ascii_text);
			} else ascii_text[j++] = (char) text[i];
		}
		ascii_text[j] = 0;
	}
	else {
		for (i=0; i<pos; i++)
			ascii_text[i] = (char) text[i];
		ascii_text[i] = 0;
	}

	smfree(text);
	return ascii_text;
}

Dll void APIENTRY
i_FreeASCIIText(char *text)
{
	smfree(text);
}
