/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */
#include "rafport.h" 
#include "entity.h"
#include "dafs.h"
#include "smemory.h"
#include "region.h"

iDAFSError
i_AppendBox(iEntityPtr entity, iBox box)
{
	iDAFSError err=DAFSOK;
	iOrthogonPtr poly;
	iData *data;
	iBox old;
	iBoxListPtr reg1, reg2, result;
	data = entity->data;
	switch (entity->data->area.a_type){
	
	case iABox:
		if (data->area.a.box.w&&box.w){
			old=data->area.a.box;
			data->area.a_type=iABoxList;
			reg1=i_NewBoxList();
			reg2=i_NewBoxList();
			result=i_NewBoxList();
			if (!reg1||!reg2||!result)
				return MallocFailError;
			i_UnionBoxWithBoxList(&old, reg1, reg2);
			i_UnionBoxWithBoxList(&box, reg2, result);
			data->area.a.box_list=result;
			i_DisposeBoxList(reg1);
			i_DisposeBoxList(reg2);
								
		}
		else
			if (box.w)
				data->area.a.box=box;
		break;
	case iABoxList:
		reg1=data->area.a.box_list;
		reg2=i_NewBoxList();
		if (!reg1||!reg2)
			return MallocFailError;
		i_UnionBoxWithBoxList(&box, reg1, reg2);
		i_DisposeBoxList(reg1);
		data->area.a.box_list=reg2;
		break;
	case iAOrthogonList:			
		if(err = i_Box2Orthogon(box, &poly))
			return err;
		i_AppendOrthogon(entity, poly);
		i_DisposeOrthogon(poly);
		break;
	}
	return err;
}

static void removepoly(iOrthogonListPtr border, int4 i)
{
	i_DisposeOrthogon(border->poly[i]);
	if(--border->num_poly)
		for (;i<border->num_poly; i++)
			border->poly[i]=border->poly[i+1];
	else
		smfree(border->poly);
}

Dll iDAFSError APIENTRY
i_SubtractOrthogonList(iOrthogonListPtr shapes, iOrthogonListPtr bites)
{
	iDAFSError e;
	iBoxListPtr new, recreg, region, total, reg;
	iOrthogonListPtr temp;
	int4 i, j, k, orig_polys;
	
	if (shapes&& shapes->num_poly&&bites && bites->num_poly){
		recreg = i_NewBoxList();
		total = i_NewBoxList();		
		for (j=0; j<bites->num_poly; j++){
			if(!(e = i_Orthogon2BoxList(bites->poly[j], &reg)))
				return e;
			i_UnionBoxList(reg, recreg, total);
			i_DisposeBoxList(recreg);
			i_DisposeBoxList(reg);
			recreg=total;
		}
		orig_polys=shapes->num_poly;
		for(i=0; i<orig_polys; i++){
			if (shapes->poly[i]->n>2){
				if(!(new = i_NewBoxList()))
					return MallocFailError;
				if(e = i_Orthogon2BoxList(shapes->poly[i], &region)){
					i_DisposeBoxList(new);
					return e;
				}
				i_SubtractBoxList(region, recreg, new);
				if (e = i_BoxList2OrthogonList(new, &temp))
					return e;
				if (temp->num_poly){
					i_DisposeOrthogon(shapes->poly[i]);
					shapes->poly[i]=temp->poly[0];
					for(k=1; k<temp->num_poly; k++)
						i_AddOrthogon2OrthogonList(shapes, temp->poly[k]);
				}
				else{
					removepoly(shapes, i--);
					orig_polys--;
				}
				smfree(temp->poly);
				temp->num_poly=0;
				i_DisposeBoxList(region);
				i_DisposeBoxList(new);
			}
		}
		smfree(temp);
	}
	i_DisposeBoxList(recreg);
	return DAFSOK;
}

Dll iDAFSError APIENTRY
i_IntersectOrthogon(const iOrthogonPtr poly1, const iOrthogonPtr poly2,
				iOrthogonListPtr border)
{
	iDAFSError err=DAFSOK;
	iBoxListPtr one, two, result;
	iOrthogonPtr copy1, copy2;
	
	copy1 = i_CopyOrthogon(poly1);
	copy2 = i_CopyOrthogon(poly2);
	result = i_NewBoxList();
	if(err = i_Orthogon2BoxList(copy1, &one))
		return err;
	if(err = i_Orthogon2BoxList(copy2, &two))
		return err;
	i_IntersectBoxList(one, two, result);
	if (result->numRects)
		err = i_BoxList2OrthogonList(result, &border);
	else
		border->num_poly = 0;
	i_DisposeOrthogon(copy1);
	i_DisposeOrthogon(copy2);
	i_DisposeBoxList(one);
	i_DisposeBoxList(two);
	if (!err)
		i_DisposeBoxList(result);
	return err;
	
}

Dll iDAFSError APIENTRY
i_IntersectOrthogonList(const iOrthogonListPtr one, const iOrthogonListPtr two,
				  iOrthogonListPtr ret)
{
	iOrthogonList *border;
	iDAFSError err=DAFSOK;
	iBoxListPtr reg1, reg2, result;
	
	if(err = i_OrthogonList2BoxList(one, &reg1))
		return err;
	if(err = i_OrthogonList2BoxList(two, &reg2))
		return err;
	
	result = i_NewBoxList();
	if(!result)
		return MallocFailError;

	if(err = i_IntersectBoxList(reg1, reg2, result))
		return err;
	if(err = i_BoxList2OrthogonList(result, &border))
		return err;
	i_DisposeBoxList(result);
	
	return err;
}

Dll iDAFSError APIENTRY
i_UnionOrthogon(const iOrthogonPtr poly1, const iOrthogonPtr poly2, iOrthogonListPtr
			border)
{

	iBoxListPtr one, two, result;
	iOrthogonPtr copy1, copy2;
	iDAFSError err = DAFSOK;
	
	copy1 = i_CopyOrthogon(poly1);
	copy2 = i_CopyOrthogon(poly2);
	if (!poly1->n || !poly2->n){
		if (poly1->n){
			border->num_poly=1;
			border->poly=(iOrthogonArrayPtr)smalloc(sizeof(iOrthogonPtr));
			if (!border->poly)
				return MallocFailError;
			border->poly[0]=copy1;
			i_DisposeOrthogon(copy2);
		}
		if (poly2->n){
			border->num_poly=1;
			border->poly=(iOrthogonArrayPtr)smalloc(sizeof(iOrthogonPtr));
			if (!border->poly)
				return MallocFailError;
			border->poly[0]=copy2;
			i_DisposeOrthogon(copy1);
		}
	}
	else {
		if(!(result = i_NewBoxList()))
			err = MallocFailError;
		if(!err) err = i_Orthogon2BoxList(copy1, &one);
		if(!err) err = i_Orthogon2BoxList(copy2, &two);
		if(!err) i_UnionBoxList(one, two, result);
		if(!err) err = i_BoxList2OrthogonList(result, &border);
		if(err)
			border->num_poly=0;
		i_DisposeBoxList(one);
		i_DisposeBoxList(two);
		i_DisposeBoxList(result);
		i_DisposeOrthogon(copy1);
		i_DisposeOrthogon(copy2);
	}

	return err;
}	

/*iDAFSError i_IntersectOrthogonList(const iEntityPtr entity, const iOrthogonPtr poly)
{
	int4 i,j,original_polys;
	iOrthogonList *border, *new_border;
	void *data;
	iATag tag;
	iDAFSError err=DAFSOK;
	iBox box;
	if (err = i_GetArea(entity, &data, &tag))
		return err;
	if (tag==iABox){
		box=*(iBox*)data;
		border->poly[0]=i_Box2Orthogon(box);
		border->num_poly=1;
	}
	else 
		border = (iOrthogonList*)data;

	original_polys=border->num_poly;
	for (i=0; i<original_polys; i++){
		new_border = i_IntersectOrthogon(border->poly[i], poly);
		if (!new_border->num_poly){
			removepoly(border, i--);
			original_polys--;
		}
		else{
			i_DisposeOrthogon(border->poly[i]);
			border->poly[i]=new_border->poly[0];
			for (j=1; j<new_border->num_poly;j++)
				if (err=i_AppendOrthogon(entity, new_border->poly[j]))
					return err;

			smfree(new_border->poly);
			smfree(new_border);
		}
	}
	if (tag==iABox){
		smfree(border->poly);
		smfree(border);
	}
	return err;
}*/

Dll iDAFSError APIENTRY
i_UnionOrthogonList(const iOrthogonListPtr border1, 
					const iOrthogonListPtr border2,
					iOrthogonListPtr result)
{	
	iBoxListPtr one, two, three;
	iDAFSError err = DAFSOK;

	if(err = i_OrthogonList2BoxList(border1, &one))
		return err;
	if(err = i_OrthogonList2BoxList(border2, &two)){
		i_DisposeBoxList(one);
		return err;
	}

	if(!(three = i_NewBoxList()))
		return MallocFailError;

	if(err = i_UnionBoxList(one, two, three))
		return err;
	if(err = i_BoxList2OrthogonList(three, &result))
		result->num_poly = 0;
	i_DisposeBoxList(one);
	i_DisposeBoxList(two);
	i_DisposeBoxList(three);
	 
	return err;
}

Dll iDAFSError APIENTRY
i_OrthogonList2BoxList(iOrthogonListPtr shapes, iBoxListPtr *boxes)
{	
	iDAFSError e;
	iBoxListPtr recreg, total, reg;
	int4 j;

	*boxes = 0;
	if (shapes && shapes->num_poly){

		if(!(recreg = i_NewBoxList()))
			return MallocFailError;
		
		for (j = 0; j < shapes->num_poly; j++){
			if(!(total = i_NewBoxList()))
				return MallocFailError;
			if(e = i_Orthogon2BoxList(shapes->poly[j], &reg))
				return e;
			i_UnionBoxList(reg, recreg, total);
			i_DisposeBoxList(recreg);
			i_DisposeBoxList(reg);
			recreg = total;
		}
	}

	*boxes = total;
	return DAFSOK;
}				
	
