
/*
 * Copyright (c) 1995 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

/*
// really only used in image mode
// this allows you to limit which overlay information
// you will see
// basically what entities
//
// bogus -- I didn't thoroughly review this
// I suspect it could be cleaner, more table driven
// have less overlap in functions
// but since we don't create a million of these
// I will defer reviewing it
*/

#include "rafport.h"
#include "illum.h"

#include <X11/Xutil.h>

#include <Xm/RowColumn.h>
#include <Xm/ToggleBG.h>
#include <Xm/BulletinB.h>
#include <Xm/PanedW.h>

#include "misc/XArg.h"
#include "misc/misc.h"
#include "svpmap.h"
#include "config.h"
#include "dllist.h"
#include "smemory.h"

#include "EntityAux.h" // for conf class
#include "misc/bb.h" // for bbNullP
#include "ChooseDialog.h"

// bogus this depends on information that isnt
// kept here and could easily get messed up
// flag for other type, must be power of 2 > (1<<iGlyphT)
const int type_other=(1<<15);

struct Choice {
	const char *name;
	Widget widget;
	char   original_position;
	char   current_position;
	char   default_position;
};

typedef struct Choice Choice;

/*
// bogus should dynamically update to show
// what entity types are currently available
*/

Choice type_choices[] = {
	{iDocument, NULL, 1, 1, 1},
	{iParagraph, NULL, 1, 1, 1},
	{iLine, NULL, 1, 1, 1},
	{iWord, NULL, 1, 1, 1},
	{iGlyph, NULL, 1, 1, 1},
	{"other", NULL, 1, 1, 1},
};

Choice glyph_choices[] = {
	{"low rec", NULL, 1, 1, 1},
	{"non rec", NULL, 1, 1, 1},
	{"multi rec", NULL, 1, 1, 1},
	{"O0", NULL, 1, 1, 1},
	{"other", NULL, 1, 1, 1},
};

#define FLAGGED     "flagged"
#define NOT_FLAGGED "not flagged"

Choice flag_choices[] = {
	{FLAGGED, NULL, 1, 1, 1},
	{NOT_FLAGGED, NULL, 1, 1, 1},
};



static void
apply_pushed(Widget w, void *user_data, XmAnyCallbackStruct *cbs)
{
	ChooseDialog *chooser = (ChooseDialog *) user_data;
	chooser->applyPushed();
}

static void
revert_pushed(Widget w, void *user_data, XmAnyCallbackStruct *cbs)
{
	//ChooseDialog *chooser = (ChooseDialog *) user_data;
	
	int i;
	
	for(i = 0; i < XtNumber(type_choices); i++){
		type_choices[i].current_position =  type_choices[i].original_position;
		XmToggleButtonGadgetSetState(type_choices[i].widget, 
									 type_choices[i].current_position,
									 1);
	}
	
	for(i = 0; i < XtNumber(glyph_choices); i++){
		glyph_choices[i].current_position =  glyph_choices[i].original_position;
		XmToggleButtonGadgetSetState(glyph_choices[i].widget, 
									 glyph_choices[i].current_position,
									 1);
	}
	
	for(i = 0; i < XtNumber(flag_choices); i++){
		flag_choices[i].current_position =  flag_choices[i].original_position;
		XmToggleButtonGadgetSetState(flag_choices[i].widget, 
									 flag_choices[i].current_position,
									 1);
	}
}

static void
default_pushed(Widget w, void *user_data, XmAnyCallbackStruct *cbs)
{
	//ChooseDialog *chooser = (ChooseDialog *) user_data;
	
	int i;
	
	for(i = 0; i < XtNumber(type_choices); i++){
		type_choices[i].current_position =  type_choices[i].default_position;
		XmToggleButtonGadgetSetState(type_choices[i].widget, 
									 type_choices[i].current_position,
									 1);
	}
	
	for(i = 0; i < XtNumber(glyph_choices); i++){
		glyph_choices[i].current_position =  glyph_choices[i].default_position;
		XmToggleButtonGadgetSetState(glyph_choices[i].widget, 
									 glyph_choices[i].current_position,
									 1);
	}
	
	for(i = 0; i < XtNumber(flag_choices); i++){
		flag_choices[i].current_position =  flag_choices[i].default_position;
		XmToggleButtonGadgetSetState(flag_choices[i].widget, 
									 flag_choices[i].current_position,
									 1);
	}
	apply_pushed(w,user_data,cbs);
}

static void
close_pushed(Widget w, void *user_data, XmAnyCallbackStruct *cbs)
{
	ChooseDialog *chooser = (ChooseDialog *) user_data;
	chooser->hide();
}

static void
ok_pushed(Widget w, void *user_data, XmAnyCallbackStruct *cbs)
{
	ChooseDialog *chooser = (ChooseDialog *) user_data;
	
	chooser->applyPushed();
	chooser->hide();
}

static ActionAreaItem action_items[] = {
	{ "Ok",     ok_pushed,     NULL          },
	{ "Apply",  apply_pushed,  NULL          },
	{ "Revert", revert_pushed, NULL          },
	{ "Default",default_pushed, NULL          },
	{ "Close",  close_pushed,  NULL          },
};


static void
typeToggleMe(Widget w, int ind, XmToggleButtonCallbackStruct *state)
{
	type_choices[ind].current_position = state->set;
}

static void
glyphToggleMe(Widget w, int ind, XmToggleButtonCallbackStruct *state)
{
	glyph_choices[ind].current_position = state->set;
}

static void
flagToggleMe(Widget w, int ind, XmToggleButtonCallbackStruct *state)
{
	flag_choices[ind].current_position = state->set;
}


void
ChooseDialog::applyPushed()
{
	int i;
	DLList list;
	int buttons_changed, types_changed=FALSE;
	
	int old_options = this->option_mask;
	int old_flagged = this->flagged;
	int old_not_flagged = this->not_flagged;
	
	
	dllistCtor(&list);
	for(i = 0; i < XtNumber(type_choices); i++){
		types_changed = types_changed || type_choices[i].current_position != type_choices[i].original_position;
		type_choices[i].original_position =  type_choices[i].current_position;
		if(type_choices[i].current_position)
			dllistAppend(&list, (char *) type_choices[i].name);
	}
	this->setTypes(&list);
	
	dllistClear(&list);
	for(i = 0; i < XtNumber(glyph_choices); i++){
		glyph_choices[i].original_position =  glyph_choices[i].current_position;
		if(glyph_choices[i].current_position)
			dllistAppend(&list, (char *) glyph_choices[i].name);
	}
	this->setOptions(&list);
	
	
	dllistClear(&list);
	for(i = 0; i < XtNumber(flag_choices); i++){
		flag_choices[i].original_position =  flag_choices[i].current_position;
		if(flag_choices[i].current_position)
			dllistAppend(&list, (char *) flag_choices[i].name);
	}
	this->setFlags(&list);
	
	dllistDtor(&list);
	
	buttons_changed = 
		(types_changed ||
		 old_options != this->option_mask ||
		 old_flagged != this->flagged || 
		 old_not_flagged != this->not_flagged);
	
	(*this->choose_function) (this->user_data, buttons_changed);

}

void
ChooseDialog::create()
{
	int i;
	Widget pane;
	Widget action_a;
	XArg args;
	Widget parent = this->app_shell;
		
	args.reset();
	args.setArg(XmNtitle,  "Display in Image mode");
		
	this->dialog = 
		XmCreateBulletinBoardDialog(parent,
									"choose",
									args.getArgs(),
									args.getSize());
	args.setArg(XmNsashHeight, 1);
	args.setArg(XmNsashWidth,  1);
	pane = XmCreatePanedWindow(this->dialog,
							   "pane",
							   args.getArgs(),
							   args.getSize());
		
	Widget toggle_box;
	toggle_box = XtVaCreateWidget("togglebox",
								  xmRowColumnWidgetClass, pane,
								  XmNpacking,        XmPACK_COLUMN,
								  XmNnumColumns,     1,
								  NULL);
		
	/* simply loop thru the strings creating a widget for each one */
	for (i = 0; i < XtNumber(type_choices); i++) {
		Widget b;
		type_choices[i].original_position = 
		  getTypeVisible(type_choices[i].name);
		b = XtVaCreateManagedWidget(type_choices[i].name,
									xmToggleButtonGadgetClass, 
									toggle_box, 
									XmNset, type_choices[i].original_position,
									NULL);
		type_choices[i].widget = b;
		XtAddCallback(b, XmNvalueChangedCallback, 
					  (XtCallbackProc) typeToggleMe, (void *) i);
	}
	XtManageChild(toggle_box);
		
	Widget glyph_toggle_box;
	glyph_toggle_box = XtVaCreateWidget("glyph_togglebox",
										xmRowColumnWidgetClass, pane,
										XmNpacking,        XmPACK_COLUMN,
										XmNnumColumns,     1,
										NULL);
		
	/* simply loop thru the strings creating a widget for each one */
	for (i = 0; i < XtNumber(glyph_choices); i++) {
		Widget b;
		b = XtVaCreateManagedWidget(glyph_choices[i].name,
									xmToggleButtonGadgetClass, 
									glyph_toggle_box, 
									XmNset, glyph_choices[i].original_position,
									NULL);
		glyph_choices[i].widget = b;
		XtAddCallback(b, XmNvalueChangedCallback, 
					  (XtCallbackProc) glyphToggleMe, (void *) i);
	}
	XtManageChild(glyph_toggle_box);
		
	Widget flag_toggle_box;
	flag_toggle_box = XtVaCreateWidget("flag_togglebox",
									   xmRowColumnWidgetClass, pane,
									   XmNpacking,        XmPACK_COLUMN,
									   XmNnumColumns,     1,
									   NULL);
		
	/* simply loop thru the strings creating a widget for each one */
	for (i = 0; i < XtNumber(flag_choices); i++) {
		Widget b;
		b = XtVaCreateManagedWidget(flag_choices[i].name,
									xmToggleButtonGadgetClass, 
									flag_toggle_box, 
									XmNset, flag_choices[i].original_position,
									NULL);
		flag_choices[i].widget = b;
		XtAddCallback(b, XmNvalueChangedCallback, 
					  (XtCallbackProc) flagToggleMe, (void *) i);
	}
	XtManageChild(flag_toggle_box);
		
	action_items[0].data = this;
	action_items[1].data = this;
	action_items[2].data = this;
	action_items[3].data = this;
	action_items[4].data = this;
		
	action_a = CreateActionArea(pane, action_items, XtNumber(action_items));
	XtManageChild(pane);
}

ChooseDialog::ChooseDialog(MainWindow *input_mw, Widget input_app_shell, 
				 ChooseFunction input_choose_function, 
				 void *input_user_data)
	: Dialog(input_mw, input_app_shell), 
	  choose_function(input_choose_function),
	  user_data(input_user_data)
{
	this->map = (stringVoidPtrMap *) smalloc(sizeof(stringVoidPtrMap));
	svmapInit(this->map, 0, 100);
	DLList l;
	int i;
	
	dllistCtor(&l);
	for(i = 0; i < XtNumber(type_choices); i++){
		dllistAppend(&l, (char *) type_choices[i].name);
	}
	this->setTypes(&l);
	
	dllistClear(&l);
	for(i = 0; i < XtNumber(glyph_choices); i++){
		dllistAppend(&l, (char *) glyph_choices[i].name);
	}
	this->setOptions(&l);
	
	dllistClear(&l);
	for(i = 0; i < XtNumber(flag_choices); i++){
		dllistAppend(&l, (char *) flag_choices[i].name);
	}
	this->setFlags(&l);
	
	dllistDtor(&l);
}

ChooseDialog::~ChooseDialog()
{
	svmapCleanup(this->map);
	smfree(this->map);
}

int
ChooseDialog::qualify(iEntityPtr entity,int *parent_flagged)
{
	iImagePtr image;
	iBox box;
	int e_type;
	int ok;
	long flagged;
	
	image = i_GetImage(entity);
	if (!image) return 0;
	box = i_GetEnclosingBox(entity);
	if (bbNullP(box)) 
		return 0;
	
	if (! (this->flagged || this->not_flagged)) return 0;
	if ( (this->flagged && !this->not_flagged) ||
		 (!this->flagged && this->not_flagged)) {
		flagged=(!i_GetLongProp(entity, "Flag", &flagged)) || *parent_flagged;
		*parent_flagged=flagged;
		
		if(flagged && this->flagged == FALSE)
			return 0;
		if(!flagged && this->not_flagged == FALSE)
			return 0;
	}
	
	e_type = i_GetEntityType(entity);
	if ((e_type > iGlyphT || e_type == 0) && !isVisibleSet(entity))
		ok = getTypeVisible("other");
	else
		ok = getVisible(entity);

	if(ok && e_type == iGlyphT) {
		return eaConfClassify(entity) & option_mask;
	} else
		return ok;
}

void
ChooseDialog::_choose(iEntityPtr entity, DLList *l_return,int *parent_flag)
{
	int flag=*parent_flag;
	
	if (!(entity = pickpath(entity))) 
		return;
	
	iEntityPtr child;
	
	if(this->qualify(entity,&flag))
		dllistAppend(l_return, entity);
	
	for(child = i_GetChildEntity(entity); child; child = i_GetNextEntity(child)) {
		this->_choose(child, l_return,&flag);
	}
}

void
ChooseDialog::choose(iEntityPtr entity, DLList *l_return)
{
	int temp=0;
	this->_choose(entity,l_return,&temp);
}

void
ChooseDialog::setTypes(DLList *l)
{
	for(int i=0; i_TypeToName(i); i++)
		setTypeVisible(i_TypeToName(i), FALSE);
	setTypeVisible("other", FALSE);
	for(Ref r = dllistFirst(l); r; dllistNext(l, &r))
		setTypeVisible((char *)dllistContents(l,r), TRUE);
}

void
ChooseDialog::setOptions(DLList *l)
{
	Ref r;
	char *s;
	
	this->option_mask = 0;

	for(r = dllistFirst(l); r; dllistNext(l, &r)) {
		aString as;
		s = (char *) dllistContents(l, r);

		as = s;
		if(as == "low rec")
			this->option_mask = this->option_mask | agLOWREC;
		else if (as == "non rec")
			this->option_mask = this->option_mask | agNONREC;
		else if (as == "multi rec")
			this->option_mask = this->option_mask | agMULTIREC;
		else if (as == "O0")
			this->option_mask = this->option_mask | agOZERO;
		else if (as == "other")
			this->option_mask = this->option_mask | agOTHER;
	}
}

void
ChooseDialog::setFlags(DLList *l)
{
	Ref r;
	char *s;
	
	this->flagged = 0;
	this->not_flagged = 0;
	
	for(r = dllistFirst(l); r; dllistNext(l, &r)) {
		aString as;
		s = (char *) dllistContents(l, r);

		as = s;
		if(as == FLAGGED)
			this->flagged = 1;
		else if (as == NOT_FLAGGED)
			this->not_flagged = 1;
	}
}
