
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */


#ifndef _Document_h_
#define _Document_h_

#include <X11/StringDefs.h>
#include <X11/Intrinsic.h> 
#include <Xm/Xm.h>

#include <sys/types.h>
#include <sys/stat.h>

#include "dllist.h"
#include "FileName.h"

#include "illum.h"

enum SaveType {
	saveDAFSB,
	saveASCII,
	saveUnicode,
};

class Document {
public:
	MainWindow *mw;
	FileName current_directory;
	FileName file_name;
	int dirty;
	SaveType save_type,would_be_savetype;
	ModeType current_mode;
	time_t original_mod_time;

	// this resets when you open a new document
	// it is saved when you exit ooc mode or flag mode
	// it is referenced when you enter ooc mode or flag mode
	aString last_ooc_page;
	int     last_ooc_index;
	aString last_flag_page;
	
	// dyn
	iEntityPtr root_entity;
	Mode *modes[NUM_MODE_TYPES];
	
	Document(MainWindow *mw);
	~Document();
	void purgeModes();
	void purgeInactiveModes();
	time_t calculateModTime();
	
	SelectionManager *getSM();
	
	void setDirty();
	void clearDirty();
	int dirtyP();
	
	Mode *getCurrentMode();
	void closeFile();
	int openFile(const FileName& fn, ModeType mode);
	int  saveFile(); 
	int  newFile(); 
	int  initDoc(iEntityPtr entity, ModeType mode); 
	void saveAsFile();
	void setSaveType(SaveType type);
	
	void selectPage(const aString&, int sub_page = 0);
	
	void goOOCMode();
	void goImageMode();
	void goFlagMode();
	void goHierMode();
	void goTreeMode();
	void goRTEMode();
	void changeMode(ModeType input_mode);
	
	void configure();
	
#ifdef RAF_INTERNAL
	void setMailType(const aString& mail_type);
	aString getMailType();
#endif
	
	void addNewEntity(iEntityPtr, int do_callback);
	
	void createTextMode();
};

#ifdef RAF_INTERNAL
#define MAIL_TYPE_PROP "MailType"

#define UNKNOWN_TYPE "UNKNOWN"
#define HAND_TYPE "Handwriting"
#define FIXED_TYPE "Fixed Pitch"
#define VARIABLE_TYPE "Variable Pitch"

enum MailType {
	mtUNKNOWN,
	mtHAND,
	mtFIXED,
	mtVARIABLE,
};
#endif

#endif
