
/*
 * Copyright (c) 1994, 1995 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

/* 
// currently a sui generis object that follows the selection
// but maybe we can have multiple ones
*/

#include "rafport.h"
#include "rafutil.h"

#include "illum.h"

#include <Xm/Label.h>
#include <Xm/PanedW.h>
#include <Xm/BulletinB.h>
#include <Xm/SelectioB.h>
#include <Xm/PushB.h>
#include <Xm/PushBG.h>
#include <Xm/TextF.h>
#include <Xm/LabelG.h>
#include <Xm/Form.h>
#include <Xm/Frame.h>
#include <Xm/DrawingA.h>
#include <Xm/List.h>
#include <Xm/RowColumn.h>
#include <Xm/MessageB.h>
#include "smemory.h"


#include "misc/Unicode.h"
#include "misc/XArg.h"
#include "misc/misc.h"
#include "ucs.h"

#include "misc/bb.h"

#include "EntityAux.h"
#include "MainWindow.h"
#include "SelectionManager.h"
#include "Application.h"
#include "Document.h"
#include "ViewCanvas.h"
#include "ListDialog.h"
#include "TypePrefDialog.h"
#include "EntityDialog.h"
#include "EntContentDialog.h"
#include "PropContentDialog.h"
#include "BoxEditDialog.h"

#include "utf.h"
#include "uString.h"
#include "misc/Unicode.h"

// x y w h

#define EVIMW 150
#define EVIMH 60

#define SELECTED_TITLE "Selection Dialog"
#define CURSOR_TITLE   "Selection Dialog"

// doesnt do enough to be worth enabling
#if 0
#define EDIT_PROPS
#endif

extern void i_SetPBBBounds(iEntityPtr entity, iBox pbb);
extern iBox i_GetPBBBounds(iEntityPtr entity);

static void editBox(Widget w, void *data, XmAnyCallbackStruct *cbs);
static void editPBBBox(Widget w, void *data, XmAnyCallbackStruct *cbs);

static void setBoxVal(Widget w, int val);
static void clearBoxVal(Widget w);



void
EntityDialog::updateTypes(void)
{
}


EntityDialog::EntityDialog(MainWindow * input_mw, Widget input_app_shell)
	: Dialog(input_mw, input_app_shell),
	  rowcol(0), pbblab(0),
	  type_field(0), typetext(0), prop_field(0),
	  data_holder(0), data_text_field(0), data_list_field(0), data_label(0),
	  entity(0), entity_popup(0),dirtyview(0),
	  next_entity(0), next_index(0), rcframe(0),
	  posset_dialog(0), prop_dialog(0)
{
	int i;
	
	for (i=0;i<NUM_BOUNDS_FIELDS;i++) {
		bounds[i] = 0;
		pbounds[i] = 0;
	}
}


EntityDialog::~EntityDialog()
{
	if (this->dialog)
		XtDestroyWidget(this->dialog);
}

static void close_pushed(Widget w, void *user_data, XmAnyCallbackStruct * cbs)
{
	Widget dialog = (Widget) user_data;
	XtUnmanageChild(dialog);
}

static ActionAreaItem action_items[] = {
	{"Close", close_pushed, NULL},
};

void
EntityDialog::drawImageNow(Region rgn)
{
	Display *display = XtDisplay(this->draw);
	char *bits = 0;
	int2 dw;
	iBox tbox,box;
	int iw,ih,offx,offy;
	int index=0, charlength, index_offset=0;
	Dimension fw,fh;
	iImagePtr image;
	XArg args;
	GC mygc;

	args.reset();
	args.setArg(XmNwidth, &fw);
	args.setArg(XmNheight, &fh);
	args.getValues(this->draw);
	
	mygc = this->mw->view_canvas->grayGC();
	if (rgn)
		XSetRegion(display,mygc,rgn);
	XFillRectangle(display, XtWindow(this->draw),mygc,0,0,fw,fh);
	// what is the difference between setRegion and setClipMask?
	if (rgn)
		XSetClipMask(display,mygc,None);

	iEntityPtr en = 0;

	if(this->entity) {
		en = this->entity;
		index=this->index;
		tbox = i_GetEnclosingBox(this->entity);
	} else if (this->next_entity) {
		en = this->next_entity;
		index=this->next_index;
		tbox = i_GetEnclosingBox(this->next_entity);
		//tbox.w = 1;
	}

	if(!en)
		return;

	image = i_GetImage(en);

	if(!image)
		return;

	iw = i_GetImageWidth(image);
	ih = i_GetImageHeight(image);

	if (i_GetEDataType(en)==iDString){
		charlength=strlen(i_GetText(en,0));
		if(charlength && tbox.w>fw){
			index_offset = (index-charlength/2)*(int)tbox.w/charlength;
		}
	}
	else
		if (i_GetEDataType(en)==iDUString){
			charlength=ucslen(i_GetUText(en,0));
			if (charlength && tbox.w>fw){
				index_offset = (index-charlength/2)*(int)tbox.w/charlength;
			}
		}
	tbox.x+=index_offset;		
	box  = tbox;
	
	// since we display always in mag 1, don't have to scale
	tbox.x -= (fw - tbox.w)/2;
	tbox.y -= (fh - tbox.h)/2;
	if (tbox.x < 0) {
		offx = -tbox.x;
		tbox.x = 0;
	} else offx = 0;
	if (tbox.y < 0) {
		offy = -tbox.y;
		tbox.y = 0;
	} else offy = 0;
	tbox.w = fw-offx;
	tbox.h = fh-offy;
	if (tbox.x + tbox.w > iw)
		tbox.w = iw - tbox.x;
	if (tbox.y + tbox.h > ih)
		tbox.h = ih - tbox.y;
	box.x -= tbox.x;
	box.y -= tbox.y;
	int normal_image_mag_exp = 1 << 16;
	if (i_Image2BitMap(image, tbox, normal_image_mag_exp, &dw, &bits))
		return;
	// create the image to be the size of the bitmap
	XImage *x_image = XCreateImage(display, 0, 1, XYBitmap, 
								   0, // tricky -- ignore initial bits
								   bits,
								   tbox.w, tbox.h, 
								   8, // quanta of scan line
								   dw);
	x_image->byte_order = MSBFirst;
	x_image->bitmap_bit_order = MSBFirst;
	_XInitImageFuncPtrs(x_image);

	// create the pixmap to same size

	// copy the entire image to the pixmap
	mygc = this->mw->view_canvas->copyGC();
	if (rgn)
		XSetRegion(display,mygc,rgn);
	// we do this again because we changed the gc
	// do I have to worry about the gc needing to change back?

	XPutImage(display,XtWindow(this->draw), mygc, x_image, 
			  0,0, offx,offy, tbox.w, tbox.h);

	// we want to free the image ourselves, with smfree cause it was smalloced
	if (rgn)
		XSetClipMask(display,mygc,None);
	x_image->data=0;
	XDestroyImage(x_image);
	smfree(bits);

	// can these hardcoded sizes be good

	if(this->entity) {
		// we have an entity selected instead of a cursor
		int maxpoints=50, i,j;
		iOrthogonListPtr border;
		iOrthogonPtr poly, newpoly;
		XRectangle rect;
		iATag tag;
		void *data;
		
		newpoly=i_NewOrthogon(maxpoints);
		if (!i_GetArea(this->entity, &data, &tag)){
			switch(tag){
			
			case iABoxList:  /* change to iAOrthogonList and fall through  */
				    i_ForceAreaType(this->entity, iAOrthogonList);
					if (i_GetArea(this->entity, &data, &tag))
						break;
			case iAOrthogonList:
					border=(iOrthogonList*)data;		
					if (border && border->num_poly)
						for (i=0; i<border->num_poly; i++){
							poly=border->poly[i];
							if(poly->n){
								if (poly->n >= maxpoints){
									maxpoints = poly->n+1;
									newpoly->point = (iPoint *)
										smrealloc(newpoly->point,
												  sizeof(iPoint)*(poly->n+1));
								}
								for (j=0; j<poly->n; j++) {
									newpoly->point[j].x = 
										(short)((poly->point[j].x))+offx-tbox.x;
									newpoly->point[j].y = 
										(short)((poly->point[j].y))+offy-tbox.y;
								}
								newpoly->n = poly->n;

								newpoly->point[newpoly->n].x = newpoly->point[0].x;
								newpoly->point[newpoly->n++].y = newpoly->point[0].y;
								XDrawLines(display, XtWindow(this->draw), 
											this->mw->view_canvas->blueGC(), 
											(XPoint*)newpoly->point, newpoly->n, 
											CoordModeOrigin);

							}
						}
				break;

			case iABox:
					box=*(iBox*)data;
					/* change the dimensions of the bounding box (rect) 
					   to be outside of the entity's bounding box */
					rect.x = box.x + offx - tbox.x - 1;
					rect.y = box.y + offy - tbox.y - 1;
					rect.width =  box.w + 1;
					rect.height = box.h + 1;
					XDrawRectangle(display, XtWindow(this->draw), 
									this->mw->view_canvas->blueGC(),
									rect.x, rect.y,
									rect.width, rect.height);
					break;
			}
		}
		i_DisposeOrthogon(newpoly);
	} else {
		// we have a cursor selected
		// so we only want to draw the cursor
		GC gc = this->mw->view_canvas->redGC();
		XDrawRectangle(display,XtWindow(this->draw),
					   gc,
					   box.x+offx,box.y+offy,1,box.h);
		XDrawRectangle(display,XtWindow(this->draw),
					   gc,
					   box.x+offx,box.y+offy+box.h,box.w,1);
	}

}

static Region PCBregion = 0;

static void PopupCB(Widget w,EntityDialog *tf,XmDrawingAreaCallbackStruct *cbs)
{
	XGraphicsExposeEvent *ev;
	XRectangle rect;
	Display *display;

	display = XtDisplay(w);
	if (cbs->reason == XmCR_EXPOSE) {
		ev = (XGraphicsExposeEvent *)cbs->event;
		rect.x = ev->x;rect.y = ev->y;
		rect.width = ev->width;
		rect.height = ev->height;
		if (!PCBregion)
			PCBregion = XCreateRegion();
		XUnionRectWithRegion(&rect,PCBregion,PCBregion);
		if (!ev->count) {
			tf->drawImageNow(PCBregion);
			XDestroyRegion(PCBregion);
			PCBregion = 0;
		}
	} else if (cbs->reason == XmCR_RESIZE && XtIsRealized(w)) { /* must be a resize */
		XClearArea(display,XtWindow(w),0,0,0,0,1);
	}
}

static void addCellLabel(Widget w,Widget *np,char *str,int align)
{
	Widget n;
	XmString xm_string;
	XArg args;

	args.reset();
	xm_string = XmStringCreateLocalized(str);
	args.setArg(XmNlabelString, xm_string);
	args.setArg(XmNalignment, align);
	args.setArg(XmNrecomputeSize, FALSE);
	n = XmCreateLabel(w, str,args.getArgs(), args.getSize());
	XmStringFree(xm_string);
	XtManageChild(n);
	if (np) *np = n;
}

static char *colNames[NUM_BOUNDS_FIELDS] = {"X","Y","W","H"};


static XtTranslations empty_translations = 0;

static void
setSensitive(EntityDialog *ev, int sens)
{
	XtSetSensitive(ev->type_field, sens);
	XtSetSensitive(ev->proplab,    sens);
	XtSetSensitive(ev->data_label, sens);
	XtSetSensitive(ev->rowcol,     sens);
}


static void
makeTypeList(XmString **list_ptr, int *n_ptr)
{
	XmString *list;
	char * name;
	int type_index;
	int list_index;

	for (int n = 0; i_TypeToName(n); n++) {
	}
	
	list = (XmString *) smalloc(sizeof(XmString) * (n + 2));
	if (!list) {
		// bogus
		halt("argh, out of memory\n");
	}

	// skip the null named type
	for (list_index = 0, type_index = 0; type_index < n; type_index++) {
		name = (char *) i_TypeToName(type_index);
		if(strcmp(name, "") != 0)
			list[list_index++] = XmStringCreateLocalized(name);
	}

	*list_ptr = list;
	*n_ptr = list_index;
}

static void
chooseIt(Widget w,
		 EntityDialog *ev,
		 XmAnyCallbackStruct *cbs)
{
	if(ev->entity == 0)
		return;


	// you aren't allowed to change type in the text mode
	// because that would affect packing
	if(ev->mw->document->current_mode == vtRTE)
		return;

	ListDialog *select_list;
	XmString title=XmStringCreateLocalized("Choose Type");
	select_list=new ListDialog(ev->mw, ev->app_shell, 
							   ev->mw->app->app_context, title,
							   0, 0);
	
	int list_size,i;
	XmStringTable list;
	makeTypeList(&list,&list_size);
	
	aString foo;
	foo = select_list->showModal(list, list_size);
	
	for(i = 0; i < list_size; i++)
		XmStringFree(list[i]);
	smfree(list);
	
	XmStringFree(title);
	delete select_list;
	// set the type using the returned string

	if(foo != ""){
		iTypeIndex t;
		i_NameToType(foo.string(),&t); /* who needs to check errors anyway? */
		i_SetEntityType(ev->entity, t);

		ev->mw->document->setDirty();
		if (t==list_size){
			ev->mw->type_pref_view->update();
		}
	}
}


static void
editData(Widget w, void *user_data, XmAnyCallbackStruct * cbs)
{
	EntityDialog * ed = (EntityDialog *) user_data;

	if(ed->entity == 0)
		return;

	
	// you aren't allowed to change type in the text mode
	// because we are not set up to respond to events
	if(ed->mw->document->current_mode == vtRTE)
		return;

	if(ed->posset_dialog == 0)
		ed->posset_dialog = new EntContentDialog(ed->mw, ed->app_shell, 
												 ed->mw->app->app_context );
	ed->posset_dialog->showModal(ed->entity);
}

#ifdef EDIT_PROPS
static void
editProp(Widget w, void *user_data, XmAnyCallbackStruct * cbs)
{
	EntityDialog * ed = (EntityDialog *) user_data;
	iPropPtr prop;


	if(ed->entity == 0)
		return;

	if(ed->prop_dialog == 0)
		ed->prop_dialog = new PropContentDialog(ed->mw, ed->app_shell, 
												ed->mw->app->app_context );
	int *position_list = 0;
	int position_count;
	int some_selected;
	some_selected = XmListGetSelectedPos(ed->prop_field,
										 &position_list,
										 &position_count);

	if(some_selected){
		int i;
		for(i = 1, prop = i_FirstProp(ed->entity); 
			i < position_list[0]; 
			i++, prop = i_NextProp(ed->entity, prop)) 
			;

		const char *prop_name = i_GetPropName(prop);
		// will stop when it finds the right prop
		ed->prop_dialog->showModal(ed->entity, prop, prop_name);
	}
	if(position_list)
		free(position_list);
}
#endif

void 
EntityDialog::makeView()
{
	XArg args;
	int i;
	XmString xm_string;
	Widget lposf,lpropf,rowcolf,holder;
	Widget prop_and_pos_holder;
	Widget top_holder;

	top_holder = this->dialog;

	args.reset();
	args.setArg(XmNleftAttachment,XmATTACH_FORM);
	args.setArg(XmNrightAttachment,XmATTACH_FORM);
	args.setArg(XmNbottomAttachment,XmATTACH_FORM);

	holder = XmCreateForm(top_holder,"holder",
			      args.getArgs(), args.getSize());
	XtManageChild(holder);

	args.reset();
	args.setArg(XmNtopAttachment,XmATTACH_FORM);
	args.setArg(XmNleftAttachment,XmATTACH_FORM);
	args.setArg(XmNrightAttachment,XmATTACH_FORM);
	args.setArg(XmNbottomAttachment,XmATTACH_WIDGET);
	args.setArg(XmNbottomWidget,holder);
	args.setArg(XmNwidth, EVIMW);
	args.setArg(XmNheight, EVIMH);
	int4 lPixmap = None;
	args.setArg(XmNbackgroundPixmap, lPixmap);

	this->draw = XmCreateDrawingArea(top_holder,"popup",args.getArgs(),args.getSize());
	XtAddCallback(this->draw,XmNexposeCallback,PopupCB,this);
	XtAddCallback(this->draw,XmNresizeCallback,PopupCB,this);
	XtManageChild(this->draw);

	args.reset();
	args.setArg(XmNshadowType, XmSHADOW_ETCHED_IN);
	args.setArg(XmNleftAttachment,XmATTACH_FORM);
	args.setArg(XmNbottomAttachment,XmATTACH_FORM);
	rowcolf = XmCreateFrame(holder,"frame",args.getArgs(),args.getSize());
	XtManageChild(rowcolf);
	this->rcframe=rowcolf;
	
	args.reset();
	args.setArg(XmNpacking,XmPACK_COLUMN);
	args.setArg(XmNnumColumns,5); 	
	args.setArg(XmNorientation,XmVERTICAL);
	args.setArg(XmNisAligned,FALSE);
	this->rowcol = XmCreateRowColumn(rowcolf,"Boxes",
				args.getArgs(),args.getSize());

	addCellLabel(this->rowcol,0,"    ",XmALIGNMENT_CENTER);
	args.reset();
	xm_string = XmStringCreateLocalized("Box:");
	args.setArg(XmNalignment, XmALIGNMENT_END);
	args.setArg(XmNlabelString, xm_string);
	args.setArg(XmNleftAttachment,XmATTACH_FORM);
	args.setArg(XmNrightAttachment,XmATTACH_FORM);
	args.setArg(XmNrecomputeSize, FALSE);

	Widget box_edit = 
		XmCreatePushButton(this->rowcol, "name",args.getArgs(), args.getSize());

	XtAddCallback(box_edit, 
				  XmNactivateCallback,
				  editBox, 
				  this);
	XtManageChild(box_edit);
	
	args.reset();
	xm_string = XmStringCreateLocalized("PBB:");
	args.setArg(XmNalignment, XmALIGNMENT_END);
	args.setArg(XmNlabelString, xm_string);
	args.setArg(XmNleftAttachment,XmATTACH_FORM);
	args.setArg(XmNrightAttachment,XmATTACH_FORM);
	args.setArg(XmNrecomputeSize, FALSE);
	Widget PBBbox_edit = 
		XmCreatePushButton(this->rowcol, "name",args.getArgs(), args.getSize());

	XtAddCallback(PBBbox_edit, 
				  XmNactivateCallback,
				  editPBBBox, 
				  this);
	XtManageChild(PBBbox_edit);
	this->pbblab = PBBbox_edit;
#if 0
	addCellLabel(this->rowcol,0,"Box:",XmALIGNMENT_END);
	addCellLabel(this->rowcol,&this->pbblab,"PBB:",XmALIGNMENT_END);
#endif
	for (i=0;i<NUM_BOUNDS_FIELDS;i++) {
		addCellLabel(this->rowcol,0,colNames[i],XmALIGNMENT_CENTER);
		addCellLabel(this->rowcol,&this->bounds[i],"",XmALIGNMENT_CENTER);
		addCellLabel(this->rowcol,&this->pbounds[i],"",XmALIGNMENT_CENTER);
	}
	XtManageChild(this->rowcol);

 	Widget w;
  	args.reset();
 	this->type_field = XmCreateForm(holder,"type",
 										args.getArgs(), args.getSize());
 	args.reset();
 	args.setArg(XmNtopAttachment,XmATTACH_FORM);
 	args.setArg(XmNleftAttachment,XmATTACH_FORM);
 	args.setArg(XmNbottomAttachment,XmATTACH_FORM);
 	xm_string = XmStringCreateLocalized("Type:");
 	args.setArg(XmNlabelString, xm_string);
	args.setArg(XmNrecomputeSize, FALSE);
 	w = XmCreateLabel(this->type_field, "text_field",
 										args.getArgs(), args.getSize());
 	XmStringFree(xm_string);
 	XtManageChild(w);
 
 	args.reset();
 	args.setArg(XmNtopAttachment,XmATTACH_FORM);
 	args.setArg(XmNleftAttachment,XmATTACH_WIDGET);
 	args.setArg(XmNleftWidget,w);
 	args.setArg(XmNrightAttachment,XmATTACH_FORM);
 	args.setArg(XmNbottomAttachment,XmATTACH_FORM);
  	xm_string = XmStringCreateLocalized("             ");
  	args.setArg(XmNlabelString, xm_string);
	args.setArg(XmNrecomputeSize, FALSE);
	this->typetext = XmCreatePushButton(this->type_field, "text_field",
										args.getArgs(), args.getSize());

	void *v = this;
	XtCallbackProc cb = (XtCallbackProc) chooseIt;
	XtAddCallback(this->typetext, 
				  XmNactivateCallback,
				  cb, 
				  v);

 	XmStringFree(xm_string);
 	XtManageChild(this->typetext);
  	i = 4;

	args.reset();
	args.setArg(XmNleftAttachment,XmATTACH_FORM);
	args.setArg(XmNtopAttachment,XmATTACH_FORM);
	args.setArg(XmNbottomAttachment,XmATTACH_WIDGET);
	args.setArg(XmNbottomWidget,rowcolf);
	args.setArg(XmNbottomOffset,i);
	args.setArg(XmNrightAttachment,XmATTACH_OPPOSITE_WIDGET);
	args.setArg(XmNrightWidget,rowcolf);
	args.setValues(this->type_field);
	XtManageChild(this->type_field);


	// Property and Position holder
	args.reset();
	args.setArg(XmNtopAttachment,XmATTACH_FORM);
	args.setArg(XmNbottomAttachment,XmATTACH_FORM);
	args.setArg(XmNrightAttachment, XmATTACH_FORM);
	args.setArg(XmNleftAttachment,XmATTACH_WIDGET);
	args.setArg(XmNleftWidget,rowcolf);
	prop_and_pos_holder = XmCreateForm(holder,
									   "prop_and_pos", 
									   args.getArgs(),
									   args.getSize());

	// PropertyStuff
	args.reset();
	args.setArg(XmNtopAttachment,XmATTACH_FORM);
	args.setArg(XmNbottomAttachment,XmATTACH_FORM);
	args.setArg(XmNleftAttachment,XmATTACH_FORM);
	args.setArg(XmNrightAttachment, XmATTACH_POSITION);
	args.setArg(XmNrightPosition, 33);
	lpropf = XmCreateForm(prop_and_pos_holder,
						  "properties",
						  args.getArgs(), args.getSize());

	args.reset();
	xm_string = XmStringCreateLocalized("Properties");
	args.setArg(XmNalignment, XmALIGNMENT_BEGINNING);
	args.setArg(XmNlabelString, xm_string);
	args.setArg(XmNtopAttachment,XmATTACH_FORM);
	args.setArg(XmNleftAttachment,XmATTACH_FORM);
	args.setArg(XmNrightAttachment,XmATTACH_FORM);
	args.setArg(XmNrecomputeSize, FALSE);
	this->proplab = XmCreateLabel(lpropf, "name",args.getArgs(), args.getSize());
	XmStringFree(xm_string);
	XtManageChild(this->proplab);

	args.reset();
	args.setArg(XmNleftAttachment,XmATTACH_FORM);
	args.setArg(XmNrightAttachment,XmATTACH_FORM);
	args.setArg(XmNbottomAttachment,XmATTACH_FORM);
	args.setArg(XmNtopAttachment,XmATTACH_WIDGET);
	args.setArg(XmNtopWidget,this->proplab);
	args.setArg(XmNselectionPolicy, XmSINGLE_SELECT);

	args.setArg(XmNresizable,FALSE);
	args.setArg(XmNlistSizePolicy,XmRESIZE_IF_POSSIBLE);


#ifndef EDIT_PROPS
	// make it so you can't select anything
	// by snarfing the translation table
	if(empty_translations == 0)
		empty_translations = XtParseTranslationTable("");
	args.setArg(XmNtranslations, empty_translations);
#endif

	this->prop_field = XmCreateScrolledList(lpropf,"Properties",
			args.getArgs(),args.getSize());
#ifdef EDIT_PROPS
	XtAddCallback(this->prop_field, 
				  XmNsingleSelectionCallback, 
				  editProp, 
				  this);
#endif
	XtManageChild(this->prop_field);
	XtManageChild(lpropf);
	XtManageChild(prop_and_pos_holder);


	args.reset();
	args.setArg(XmNtopAttachment,XmATTACH_FORM);
	args.setArg(XmNbottomAttachment,XmATTACH_FORM);
	args.setArg(XmNrightAttachment,XmATTACH_FORM);
	args.setArg(XmNleftAttachment,XmATTACH_WIDGET);
	args.setArg(XmNleftWidget,lpropf);
	lposf = XmCreateForm(prop_and_pos_holder,
						 "lposf",
						 args.getArgs(), args.getSize());

	args.reset();
	xm_string = XmStringCreateLocalized("Choices ...");
	args.setArg(XmNalignment, XmALIGNMENT_CENTER);
	args.setArg(XmNlabelString, xm_string);
	args.setArg(XmNtopAttachment,XmATTACH_FORM);
	args.setArg(XmNleftAttachment,XmATTACH_FORM);
	args.setArg(XmNrightAttachment,XmATTACH_FORM);
	args.setArg(XmNrecomputeSize, FALSE);
	this->data_label = 
		XmCreatePushButton(lposf, "name",args.getArgs(), args.getSize());

	XtAddCallback(this->data_label, 
				  XmNactivateCallback,
				  editData, 
				  this);

	
	XmStringFree(xm_string);
	XtManageChild(this->data_label);

	args.reset();
	args.setArg(XmNleftAttachment,XmATTACH_FORM);
	args.setArg(XmNrightAttachment,XmATTACH_FORM);
	args.setArg(XmNbottomAttachment,XmATTACH_FORM);
	args.setArg(XmNtopAttachment,XmATTACH_WIDGET);
	args.setArg(XmNtopWidget,this->data_label);

	this->data_holder = XmCreateForm(lposf,"data_holder",
									 args.getArgs(), args.getSize());

	args.reset();
	args.setArg(XmNleftAttachment,XmATTACH_FORM);
	args.setArg(XmNrightAttachment,XmATTACH_FORM);
	args.setArg(XmNbottomAttachment,XmATTACH_FORM);
	args.setArg(XmNtopAttachment,XmATTACH_FORM);
	args.setArg(XmNselectionPolicy, XmSINGLE_SELECT);
	args.setArg(XmNlistSizePolicy,XmRESIZE_IF_POSSIBLE);
	args.setArg(XmNresizable,FALSE);

	if(empty_translations == 0)
		empty_translations = XtParseTranslationTable("");
	args.setArg(XmNtranslations, empty_translations);

	this->data_list_field = 
		XmCreateScrolledList(this->data_holder,"data_list_field",
							 args.getArgs(),args.getSize());
	
	//XtManageChild(this->data_text_field);
	XtManageChild(this->data_list_field);
	XtManageChild(this->data_holder);
	
	XtManageChild(lposf);

	setSensitive(this, FALSE);
	if (XmIsMotifWMRunning(this->app_shell) == FALSE) {
		Widget action_a;
		action_items[0].data = this->dialog;
		// bogus I havn't tried to figure this out.  
		// At least it needs to be tested
		// this was for use under twm
		// maybe it create a close button and so on
		action_a = CreateActionArea(this->dialog, action_items, 
					    XtNumber(action_items));
	}
	// no action area because close from window menu is enough
}


void
EntityDialog::create()
{
	XArg args;
	Widget parent = this->app_shell;
	args.setArg(XmNtitle, SELECTED_TITLE);
	int4 lPixmap = None;
	args.setArg(XmNbackgroundPixmap,lPixmap);
	this->current_title = SELECTED_TITLE;
	
	this->dialog =
		XmCreateFormDialog(parent,"view",args.getArgs(),args.getSize());
	this->makeView();
}

void
EntityDialog::setTitle()
{
	aString title;
	if(this->entity){
		title = SELECTED_TITLE;
	} else if (this->next_entity) {
		title = CURSOR_TITLE;
	}
	
	if(title != ""){
		if(title != this->current_title) {
			XArg args;
			args.setArg(XmNtitle, (char *) title.string());
			args.setValues(this->dialog);
			this->current_title = title;
		}
	}
}

static void prprop(iEntityPtr en,char *name,int type,char *buf)
{	
	long i;
	long size;
	void *p;
	char *s;
	iUCode *us;
	int4 b;

	switch (type) {
	case iPLong:
		i_GetLongProp(en,name,&i);
		sprintf(buf,"%s %d",name,i);
		break;
	case iPBool:
		i_GetBoolProp(en,name,&b);
		if (b) sprintf(buf,"%s TRUE",name);
		else sprintf(buf,"%s FALSE",name);
		break;
	case iPString:
		i_GetStringProp(en,name,&s);
		sprintf(buf,"%s %s",name,s);
		break;
	case iPUString:
		i_GetUStringProp(en,name,&us);
		sprintf(buf,"%s ",name);
		ucs2utf(buf+strlen(buf), us);
		break;
	case iPUserData:
	case iPUserIData:
		i_GetUserDataProp(en,name,&p,&size,0);
		sprintf(buf,"%s %d 0x%x",name,size,p);
		break;
	case iPUserPtr:
		i_GetUserPtrProp(en,name,&p,0);
		sprintf(buf,"%s 0x%x",name,p);
		break;
	default:
		sprintf(buf,name);
		break;
	}
}

static iUCode *glyphstr(short p)
{
        static iUCode buf[3];
        switch (p) {
        case 0:
			buf[0] = 'N';
			buf[1] = 'R';
			buf[2] = 0;
        case 1:
			buf[0] = 'N';
			buf[1] = 'Z';
			buf[2] = 0;
        case ' ':
			buf[0] = 'S';
			buf[1] = 'P';
			buf[2] = 0;
        default:
			buf[0] = p;
			buf[1] = 0;
        }

		return buf;
}
static int
withoutBBs(iEntityPtr entity)
{
	iEntityPtr child = 0;
	int without_bbs;
	
	without_bbs = i_GetEnclosingBox(entity).w == 0;
	for(child = i_GetChildEntity(entity); 
		child && without_bbs; ) {
		without_bbs = withoutBBs(child);
		child = i_GetNextEntity(child);
	}

	return without_bbs;
}

static int
childrenWithoutBBs(iEntityPtr entity)
{
	iEntityPtr child;
	int children = FALSE;
	int without_bbs = TRUE;

	for(child = i_GetChildEntity(entity); child && without_bbs; child = i_GetNextEntity(child)){
		children = TRUE;
		without_bbs = withoutBBs(child);
	}

	return children && without_bbs;
}

void
EntityDialog::doView()
{
	iPosSetPtr ps;
	int i, j, k;
	XArg args;
	XmString xm_string, xm_u_string;
	int2 valid;
	char *text;
	char pos_buf[256];
	char *temp;
	iUCode *u_temp, *u_text;
	iEntityPtr en;
	if (!this->dialog || !XtIsManaged(this->dialog)) return;

	// this does a lot of drawing
	// decides what to draw itself??
	if (this->draw && XtWindow(this->draw))
		this->drawImageNow(0);
		
		// dont draw again for a while
	this->dirtyview = 0;
	XmListDeleteAllItems(this->prop_field);
	XmListDeleteAllItems(this->data_list_field);

	// if there is an entity, lots of data stuffing
		// otherwise set everything insensitive
	if ((en = this->entity) || (en = this->next_entity)) {
		iBox bounds = i_GetPBBBounds(en);

		// put ellipses after name 
		char *buffer, *name = (char *) i_TypeToName(i_GetEntityType(en));
		if(name){
			buffer = (char *) smalloc(sizeof(char)*(strlen(name)+ 5));
			strcpy(buffer, name);
			strcat(buffer, " ...");
		}
		else {
			buffer = (char *) smalloc(sizeof(char)*4);
			strcpy(buffer, "...");
		}
		
		args.reset();
		xm_string = XmStringCreateLocalized((char *) buffer);
		smfree(buffer);
		args.setArg(XmNlabelString, xm_string);
		args.setValues(this->typetext);
		XmStringFree(xm_string);

		args.reset();
		if (bbNullP(bounds) == FALSE) {
			setBoxVal(this->pbounds[0],bounds.x);
			setBoxVal(this->pbounds[1],bounds.y);
			setBoxVal(this->pbounds[2],bounds.w);
			setBoxVal(this->pbounds[3],bounds.h);
			xm_string = XmStringCreateLocalized("PBB:");
		} else {
			clearBoxVal(this->pbounds[0]);
			clearBoxVal(this->pbounds[1]);
			clearBoxVal(this->pbounds[2]);
			clearBoxVal(this->pbounds[3]);
			xm_string = XmStringCreateLocalized("");
		}
		args.setArg(XmNlabelString, xm_string);
		args.setValues(this->pbblab);
		XmStringFree(xm_string);

		bounds = i_GetEnclosingBox(en);
		setBoxVal(this->bounds[0],bounds.x);
		setBoxVal(this->bounds[1],bounds.y);
		setBoxVal(this->bounds[2],bounds.w);
		setBoxVal(this->bounds[3],bounds.h);

		for (iPropPtr p = i_FirstProp(en); p; p = i_NextProp(en,p)) {
			const char *prop_name = i_GetPropName(p);
			
#ifndef RAF_INTERNAL
			if (!strcmp(prop_name,ILLUM_PROP) 
				|| !strcmp(prop_name, EDITING_PROP)) continue;
#endif
			char prop_buf[256];
			prprop(en,(char *) prop_name,i_GetPropType(p),prop_buf);
			xm_string = XmStringCreateLocalized(prop_buf);
			XmListAddItemUnselected(this->prop_field,xm_string,0);
			XmStringFree(xm_string);
		}


		switch (i_GetEDataType(en)) {
		case iD1PosSet:
		case iDPosSet:
			ps = i_GetPosSet(en,&valid);

			for (i=0;i<valid;i++) {
				xm_u_string = 0;
				XmString sum_string = 0;
				XmFontList list = 0;
				iUCode *gs = glyphstr(ps[i].chr);

				xm_u_string = makeCompoundString(XtDisplay(this->dialog),
												 gs, ucslen(gs), &list);
				sprintf(pos_buf," %d", ps[i].conf);
				xm_string = XmStringCreateLocalized(pos_buf);
				sum_string = XmStringConcat(xm_u_string, xm_string);
				XmListAddItemUnselected(this->data_list_field,sum_string,0);
				if(list) {
					XArg args;
					args.reset();
					args.setArg(XmNfontList, list);
					args.setValues(this->data_list_field);
				}
				XmStringFree(xm_u_string);
				XmStringFree(sum_string);
				XmStringFree(xm_string);
			}
			xm_string = XmStringCreateLocalized("Choices ...");

			break;
		case iDShort:
			sprintf(pos_buf,"%d",i_GetShort(en,0));
			xm_string = XmStringCreateLocalized(pos_buf);
			XmListAddItemUnselected(this->data_list_field,xm_string,0);
			XmStringFree(xm_string);
			xm_string = XmStringCreateLocalized("Long ...");
			break;
		case iDString:
			text = (char*)i_GetText(en,0);
			temp = (char*)smalloc(strlen(text)+1);
			i=j=k=0;
			while(text[i]){
				while(text[i] && text[i]!='\n'){
					temp[j++]=text[i++];
				}
				temp[j]=0;
				xm_string = XmStringCreateLocalized(temp);
				XmListAddItemUnselected(this->data_list_field,xm_string,++k);
				XmStringFree(xm_string);			
				j=0;
				if (text[i])
					i++;
			}
			smfree(temp);	
			xm_string = XmStringCreateLocalized("String ...");
			break;
		case iDUserData:
			xm_string = XmStringCreateLocalized("User Data ...");

			break;
		case iDConf:
			sprintf(pos_buf,"%d",i_GetConf(en));
			xm_string = XmStringCreateLocalized(pos_buf);
			XmListAddItemUnselected(this->data_list_field,xm_string,0);
			XmStringFree(xm_string);
			xm_string = XmStringCreateLocalized("Confidence ...");

			break;
		case iDUString:
			{
			XmFontList list = 0;
			u_text = (iUCode*)i_GetUText(en,0);			
			if (u_text) {
				// run the string through context
				uString u_context = uniContext(u_text);

				// ignoring const actually ok because it will rearrange the
				// string but not change the length
				uniDirection(u_context.string());
				u_temp = (iUCode *) smalloc(sizeof(iUCode) * (u_context.length() + 1));

				i=j=k=0;
				while(u_context[i]){
					while(u_context[i] && u_context[i] != (iUCode) '\n')
						u_temp[j++] = u_context[i++];

					u_temp[j] = 0;
					xm_u_string = makeCompoundString(XtDisplay(this->dialog),
													 u_temp, ucslen(u_temp), &list);
					XmListAddItemUnselected(this->data_list_field,xm_u_string,++k);
					if(list) {
						XArg args;
						args.reset();
						args.setArg(XmNfontList, list);
						args.setValues(this->data_list_field);
					}
					XmStringFree(xm_u_string);			
					j=0;
					if (u_text[i])
						i++;
				}
				smfree(u_temp);
			}
			xm_string = XmStringCreateLocalized("Uni String ...");
			}
			break;
		case iDNo:
			if(childrenWithoutBBs(en)){
				XmFontList list = 0;
				u_text = (iUCode*)i_AllocUText(en);
				if (u_text) {
					u_temp = (iUCode*)smalloc((ucslen(u_text)+1)*sizeof(iUCode));
					i=j=k=0;
					while(u_text[i]){
						while(u_text[i] && u_text[i]!=(short)'\n'){
							u_temp[j++]=u_text[i++];
						}
						u_temp[j]=0;
						xm_u_string = makeCompoundString(XtDisplay(this->dialog),
														 u_temp, ucslen(u_temp), &list);
						// this is because the text can be nothing but a newline
						// and so the u_string can be empty
						// this is a kludge and I really just need a unicode
						// text widget
						if(xm_u_string) {
							XmListAddItemUnselected(this->data_list_field,xm_u_string,++k);
							if(list) {
								XArg args;
								args.reset();
								args.setArg(XmNfontList, list);
								args.setValues(this->data_list_field);
							}
							XmStringFree(xm_u_string);			
						}
						j=0;
						if (u_text[i])
							i++;
					}
					smfree(u_temp);	
					// free this because it is alloc uni
					i_FreeUText(u_text);

				}
				xm_string = XmStringCreateLocalized("Children's Text ...");
				break;
			}
		default:
			xm_string = XmStringCreateLocalized("No Content ...");
			break;
		}

		args.reset();
		args.setArg(XmNlabelString, xm_string);
		args.setValues(this->data_label);
		XmStringFree(xm_string);
		
	} else {
		// no entity so blank things and set insensitive
		xm_string = XmStringCreateLocalized("");
		args.reset();
		args.setArg(XmNlabelString, xm_string);
		for (i = 0;i<NUM_BOUNDS_FIELDS;i++) {
			args.setValues(this->bounds[i]);
			args.setValues(this->pbounds[i]);
		}
		args.setValues(this->pbblab);

		
		args.setValues(this->typetext);
		XmStringFree(xm_string);

	}

	setSensitive(this, (this->entity || this->next_entity));
}

void
EntityDialog::updateEntity(iEntityPtr entity)
{
	if (this->dirtyview ||
		(this->entity && (!entity || (entity == this->entity) || (entity = this->next_entity))))
		doView();
}

void
EntityDialog::clearEntity(iEntityPtr en)
{
	if (en && ((en == this->entity) || (en == this->next_entity)))
		this->setEntity(0,0);
}

void
EntityDialog::setEntity(iEntityPtr input_entity, int index)
{
	if ((this->entity == input_entity)&&
		(this->index== index) && 
		(this->next_entity == 0)) return;

	this->dirtyview = 1;
	this->entity = input_entity;
	this->index = index;
	this->next_entity = 0;
	this->next_index = 0;

	this->doView();
}

void
EntityDialog::setCursor(iEntityPtr input_next_entity, int input_next_index)
{
	if (this->next_entity == input_next_entity && 
		this->index == input_next_index &&
		(this->entity == 0) ) return;

	if (this->next_entity != input_next_entity)
		this->dirtyview = 1;
	this->next_entity = input_next_entity;
	this->next_index = input_next_index;
	this->entity = 0;

	this->doView();
}


/*
// These are used for the label widgets
*/
static void setBoxVal(Widget w,int x)
{
  char buf[10];
  XArg args;
  XmString xm_string;
	
  sprintf(buf,"%d",x);
  args.reset();
  xm_string = XmStringCreateLocalized(buf);
  args.setArg(XmNlabelString, xm_string);
  args.setValues(w);
  XmStringFree(xm_string);
}

static void clearBoxVal(Widget w)
{
	XArg args;
	XmString xm_string;
	
	args.reset();
	xm_string = XmStringCreateLocalized("");
	args.setArg(XmNlabelString, xm_string);
	args.setValues(w);
	XmStringFree(xm_string);
}


static void
editBox(Widget w, void *data, XmAnyCallbackStruct *cbs)
{
  EntityDialog *entity_dialog = (EntityDialog *) data;

  // not allowed to do this because it would affect
  // layout
  if(entity_dialog->mw->document->current_mode == vtRTE)
	return;

  BoxEditDialog *d;
  d = new BoxEditDialog("Edit Bounds", entity_dialog);
  d->showModal(entity_dialog->entity);
  delete d;
}

static void
editPBBBox(Widget w, void *data, XmAnyCallbackStruct *cbs)
{
	EntityDialog *entity_dialog = (EntityDialog *) data;

	// maybe this restriction is bogus
	if(entity_dialog->mw->document->current_mode == vtRTE)
	  return;


	PBBEditDialog *d;
	d = new PBBEditDialog("Edit PBB", entity_dialog);
	d->showModal(entity_dialog->entity);
	delete d;
}

