

/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"
#include "illum.h"
#include "rafutil.h"
#include <Xm/FileSB.h>
#include <Xm/TextF.h>
#include <Xm/PushBG.h>
#include "misc/XArg.h"
#include "misc/prompt.h"
#include "FileName.h"
#include "MainWindow.h"
#include "Application.h"
#include "FileDialog.h"
#include "Document.h" // for the definition of SaveType



FileDialog::FileDialog(MainWindow *input_mw, 
					   Widget input_app_shell, 
					   XtAppContext input_app_context,
					   aString input_title,
					   aString input_dir)
	: ModalDialog(input_mw, input_app_shell, input_app_context),
	  title(input_title),
	  dir(input_dir)
{
}

FileDialog::~FileDialog()
{
}

void
FileDialog::setDir(aString input_dir)
{
	this->dir = input_dir;

	if(this->dialog){
		XArg args;
		XmString xms_dir;

		args.reset();
		xms_dir = XmStringCreateLocalized((char *) dir.string());
		args.setArg(XmNdirectory, xms_dir);
		
		args.setValues(this->dialog);

		XmStringFree(xms_dir);
	}
}


static char *
suckText(XmString string)
{
	XmStringContext context;
	char *text = 0;
	XmStringCharSet tag;
	XmStringDirection direction;
	Boolean separator;
	int value;

	value = XmStringInitContext(&context, string);
	if(value == FALSE)
		halt("out of memory.\n");
	value = XmStringGetNextSegment(context, &text, &tag, &direction, &separator);

	return text;
}


void 
fileFailCB(Widget widget,
		   FileDialog *dialog,
		   XmFileSelectionBoxCallbackStruct *cbs)
{
	XArg args;
	XmString dir;
	char *suck_text;

	args.reset();
	args.setArg(XmNdirectory, &dir);
	args.getValues(dialog->dialog);

	args.reset();

	suck_text = suckText(dir);
	dialog->mw->setDir(suck_text);
	XtFree(suck_text);
	XmStringFree(dir);

	dialog->hide();
	dialog->done = TRUE;
}


void 
FileDialog::create()
{
	XArg args;
	XmString xms_dir;
	
	args.reset();

	XmString title=XmStringCreateLocalized((char *) this->title.string());
	args.setArg(XmNdialogTitle, title);

	xms_dir = XmStringCreateLocalized((char *) dir.string());
	args.setArg(XmNdirectory, xms_dir);
	if(this->app_context)
		args.setArg(XmNdialogStyle, XmDIALOG_FULL_APPLICATION_MODAL);
	args.setArg(XmNuserData, this);
	args.setArg(XmNresizePolicy, XmRESIZE_NONE);


	this->dialog = XmCreateFileSelectionDialog(this->app_shell,
											   (char *) this->title.string(),
											   args.getArgs(), 
											   args.getSize());

	XmStringFree(title);					      
	XmStringFree(xms_dir);

	XtAddCallback(this->dialog, XmNcancelCallback, fileFailCB, this);
}



OpenDialog::OpenDialog(MainWindow *input_mw, Widget input_app_shell,
					   aString input_dir)
	: FileDialog(input_mw, input_app_shell, 0, "Load Document",
				 input_dir)
{
}

OpenDialog::~OpenDialog()
{
}

static void 
openAcceptCB (Widget widget,
			  OpenDialog *dialog,
			  XmFileSelectionBoxCallbackStruct *cbs)
{
    char *filename;
    FileName fn;
	
    if (XtIsSubclass(widget, xmTextFieldWidgetClass)) {
		filename = XmTextFieldGetString(widget);
    } else {
		/* file was selected from FileSelectionDialog */
		XmStringGetLtoR(cbs->value, XmSTRING_DEFAULT_CHARSET, &filename);
    }
    
    if (!filename || !*filename) { /* nothing typed? */
		if (filename)
			XtFree(filename);
		return;
    }
	
    int success;
    fn = filename;
   	if (fn.directoryP()) {
		success = 0;
		
		dialog->setDir(fn.string());
		
		// the current_directory changes but the file doesnt
		dialog->mw->setDir(fn);

		// means don't disappear
		return;
	} else 
		success = dialog->mw->document->openFile(fn, dialog->mw->would_be_mode);


    XtFree(filename);

	if(success == 0){
		// do warning in openFile, where it can be more specific
		return;
	}
	
	dialog->hide();
}

void
OpenDialog::create()
{
	this->FileDialog::create();
	XtAddCallback(this->dialog, XmNokCallback, openAcceptCB,   this);
}


SaveAsDialog::SaveAsDialog(MainWindow *input_mw, Widget input_app_shell,
						   XtAppContext input_app_context,
						   aString input_dir)
	: FileDialog(input_mw, input_app_shell, input_app_context,
				 "Save Document",
				 input_dir),
	  savetype_menu(0),
	  save_type(0)
{
}

SaveAsDialog::~SaveAsDialog()
{
}


static void setsaveDAFSBCB (Widget w,SaveType *save,void *call)
{
	*save=saveDAFSB;
}

static void setsaveASCIICB (Widget w,SaveType *save,void *call)
{
	*save=saveASCII;
}

static void setsaveUnicode (Widget w,SaveType *save,void *call)
{
	*save=saveUnicode;
}

#include "misc/misc.h"

static MenuItem savetypes[] = {
	{ 0, "DAFSB", &xmPushButtonGadgetClass, 'D', 
	  NULL, NULL, (XtCallbackProc) setsaveDAFSBCB, 0, NULL },
	{ 0, "ASCII", &xmPushButtonGadgetClass, 'A', 
	  NULL, NULL, (XtCallbackProc) setsaveASCIICB, 0, NULL },
	{ 0, "Unicode", &xmPushButtonGadgetClass, 'U', 
	  NULL, NULL, (XtCallbackProc) setsaveUnicode, 0, NULL },
	NULL_MENU,
};


static void saveAsAcceptCB (Widget widget,
							SaveAsDialog *dialog,
							XmFileSelectionBoxCallbackStruct *cbs)
{
	char *filename;
	aString fn;
	
	if (XtIsSubclass(widget, xmTextFieldWidgetClass)) {
		filename = XmTextFieldGetString(widget);
	} else {
		/* file was selected from FileSelectionDialog */
		XmStringGetLtoR(cbs->value, XmSTRING_DEFAULT_CHARSET, &filename);
	}
    
	if (!filename || !*filename) { /* nothing typed? */
		if (filename)
			XtFree(filename);
		return;
	}
	
	fn = filename;
	FileName foo;
	FileName short_name;
	foo = fn;
	short_name = foo.filename();
	if (foo.directoryP()) {
		dialog->setDir(fn.string());
		dialog->mw->setDir(fn);
		// the current_directory changes but the file doesnt

		// means don't disappear
		return;
	} else if(foo.existsP()){
		aString prompt;
		
		prompt = short_name + " exists. Overwrite?";
		
		if(uiAskUser2(prompt.string(),
				   "Yes", "No", 1) == 2) {
		    return;
		}
	}

	XtFree(filename);
	if(dialog->mw->saveAs(fn)) {
		dialog->hide();

		dialog->done = TRUE;
	}
}


void
SaveAsDialog::create()
{
	XArg args;

	this->FileDialog::create();
	XtAddCallback(this->dialog, XmNokCallback, saveAsAcceptCB, this);
		
	savetypes[0].callback_data=&this->mw->document->would_be_savetype;
	savetypes[1].callback_data=&this->mw->document->would_be_savetype;
	savetypes[2].callback_data=&this->mw->document->would_be_savetype;
	this->savetype_menu=
		BuildMenu(this->dialog, XmMENU_OPTION, "Save As: ", 'S', savetypes);

	args.reset();
	args.setArg(XmNmenuHistory, savetypes[mw->document->save_type].result);
	args.setValues(this->savetype_menu);
	
	XtManageChild(this->savetype_menu);
}

void
SaveAsDialog::setSaveType(int input_save_type)
{
	this->save_type = input_save_type;

	if(this->savetype_menu){
		XArg args;
		args.reset();
		args.setArg(XmNmenuHistory, savetypes[save_type].result);
		args.setValues(this->savetype_menu);
	}
}




