
/*
 * Copyright (c) 1994, 1995 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

// *******************************************************
// This pragma forces inclusion of debugging code
// from the nemed header file or header or same name
#ifdef __GNUC__
#pragma implementation
#endif

#include <stdio.h>
#include <ctype.h>

#include <X11/StringDefs.h>
#include <X11/Intrinsic.h> 
#include <X11/Xutil.h>
#include <Xm/Xm.h>
#include <Xm/DrawingA.h>
#include <Xm/PushB.h>
#include <Xm/ScrolledW.h>
#include <Xm/ScrolledWP.h>

#include "illum.h"

#include "misc/XArg.h"

#include "View.h"
#include "MainWindow.h"
#include "Application.h"
#include "ViewCanvas.h"
#include "Document.h"
#include "EntityAux.h"
#include "SelectionManager.h"
#include "bb.h"
#include "scroll.h"

#include "Frame.h"
#include "text/EntityFace.h"

Frame::Frame(const aString& input_name, View *input_view)
	: name(input_name), view(input_view),
	  managed(0),
	  width(0), height(0),
	  im_deleted(FALSE),
	  should_delete(FALSE),
	  redraw_procid(0)
{
	this->redraw_region.w = 0;
	dllistCtor(&this->face_list);
}

Frame::~Frame()
{
	this->flushRedrawRegion();
	if(this->managed)
		this->unmanage();
	
	Ref r;
	EntityFace *member;
	DLList *list = &this->face_list;
	
	for(r = dllistFirst(list); r; dllistNext(list, &r)){
		member = (EntityFace *) dllistContents(list, r);
		member->deleteCheap();
	}
	
	dllistDtor(&this->face_list);
}

void
Frame::purge()
{
	if(this->managed)
		this->unmanage();
	
	Ref r;
	EntityFace *member;
	DLList *list = &this->face_list;
	
	for(r = dllistFirst(list); r; dllistNext(list, &r)){
		member = (EntityFace *) dllistContents(list, r);
		eaSetFace(member->entity, this->view->getMode(), 0);
		member->deleteCheap();
	}
	
	dllistClear(&this->face_list);
	delete this;
}

int
Frame::getMagExp()
{
	return this->view->main_window->mag_exp;
}

/*
// bogus little function required because the canvas doesn't seem 
// to stick to its original size
*/
void
Frame::syncSize()
{
	if(this->getCanvas()) {
		int h;
		int w;
		XArg args;
		Dimension clip_w;
		Dimension clip_h;
		
		args.reset();
		args.setArg(XmNwidth, &clip_w);
		args.setArg(XmNheight, &clip_h);
		args.getValues(XtParent(this->getCanvas()->canvas));
		
		if (this->should_delete) {
			// well I'd rather be unmanaged
			w=11; // why 11? well why not!
			h=11;
		} else {
			w = this->width;
			h = this->height;
			w = w < 10 ? clip_w - 1 : w;
			h = h < 10 ? clip_h - 1 : h;
		}
#if 0
		// hmm here?
		if(this->should_delete)
			this->getCanvas()->unmanage();
#endif
		
		this->getCanvas()->setSize(w, h);
	}
}


int
Frame::managedP()
{
	return this->managed;
	return this->getCanvas()->managedP();
}

void
Frame::manage()
{
	if (this->managed)
		return;
	
	this->managed = TRUE;

	// bogus -- why do this
	this->_setMag(this->view->main_window->mag_exp, 
				  this->view->main_window->mag_exp);

	if(this->getCanvas()){
		this->getCanvas()->setClient(this);
		// bogus I took this out
		//this->configureCanvas(this, 10, 10);
		this->getCanvas()->manage();
		this->syncSize();
	}
}

void
Frame::unmanage()
{
	if (!this->managed)
		return;
	
	this->managed = FALSE;
	if(this->getCanvas()){
		this->getCanvas()->setClient(0);
		// bogus I took this out
		//this->configureCanvas(0, 10, 10);
		this->getCanvas()->unmanage();
		this->syncSize();
	}
}

void
Frame::pressButton(XEvent *event)
{
    iEntityPtr entity;
	
    switch(event->xbutton.button) {
	case Button1:
		entity = this->findEnclosingEntity(event->xbutton.x, event->xbutton.y);
		if(entity != 0) {
			if(event->xbutton.state & ShiftMask){
				this->toggleItem(entity);
			} else {
				this->setItem(entity);
			}
		} else {
			this->clear();
		}
		break;
	case Button2:
		break;
    }
}


void
Frame::releaseButton(XEvent *event)
{
}

void
Frame::moveButton(int state, int x, int y)
{
#if 0
	printf("frames move button.\n");
#endif
}

void
Frame::reassignValues(const DLList *list, iUCode key_value)
{
	Ref r;
	iEntityPtr entity;
	View *v;
	
	int changed=FALSE;
	
	v = this->view;
	for(r = dllistFirst(list); r; dllistNext(list, &r)){
		entity = (iEntityPtr) dllistContents(list, r);
		if(iGlyphT == i_GetEntityType(entity)) {
			i_SetGlyph(entity, key_value,CERTCONF);
			i_DisposeProp(entity, FLAG_PROP);
			changed=TRUE;
		}
	}
	// we may have been deleted from underneath
	if (changed) 
		v->main_window->document->setDirty();
}

void
Frame::deleteSelection()
{
}

void
Frame::track()
{
	iEntityPtr sole;
	iBox bounds;
	int visible, w, h, ent_w, ent_h;
	
	View *view = this->view;
	MainWindow *mw=view->main_window;


	getWidgetSize(mw->view_canvas->canvas, &w, &h);

	sole = this->getSM()->getSoleItem();
	if (sole) {
		bounds = view->getBounds(sole);
		if(bbNullP(bounds) == FALSE) {
			visible = visiblePositionP(mw->scroll_window, 
									   bounds.x, 
									   bounds.y);
			visible &= visiblePositionP(mw->scroll_window, 
										bounds.x+bounds.w-1, 
										bounds.y+bounds.h-1);
			if(!visible){
#if 0
				// I dont know why
				ent_w = (bounds.w << (mag_exp + 4)) >> 4;
				ent_h = (bounds.h << (mag_exp + 4))>>4;
#else
				ent_w = bounds.w;
				ent_h = bounds.h;
#endif
				if (ent_w<w && ent_h<h)
					this->centerOnSelected(sole);
			}
		}
	}
}


int
Frame::pressKey(iUCode key_value)
{
	if (key_value == '\b') {
		this->view->main_window->deleteSelection(FALSE);
	} else if(key_value == 20) {
		killContentsForSelection();
   	} else {
		if(key_value < ' ' || this->getSM()->getState() == scNOT) {
			ringTheBell();
			return 0;
		}
		
		// bogus why does this make sense here?
		this->track();
		this->reassignValues(this->getItems(), key_value);
	}
	
	return 1;
}

ViewCanvas *
Frame::getCanvas()
{
	return this->view->main_window->view_canvas;
}

void
Frame::configureCanvas(Frame *x, int w, int h)
{
	ViewCanvas *vc = this->view->main_window->view_canvas;
	vc->setSize(w, h);
}

void
Frame::removeFace(EntityFace *face)
{
	Ref r;
	DLList *l = &this->face_list;
	
	r = dllistSeek(l, face);
	//  bogus -- an error condition if I don't find it?
	if(r)
		dllistRemove(l, r);
	eaSetFace(face->entity, this->view->getMode(), 0);
}

void
Frame::addFace(EntityFace *face)
{
	DLList *l = &this->face_list;
	
	dllistAppend(l, face);
}

SelectionManager *
Frame::getSM()
{
	return this->view->getSM();
}

void
Frame::setItem(iEntityPtr entity)
{
	this->getSM()->setItem(entity);
}

void
Frame::toggleItem(iEntityPtr entity)
{
	this->getSM()->toggleItem(entity);
}

void
Frame::clear()
{
	this->getSM()->clear();
}

iEntityPtr
Frame::getSoleItem()
{
	return this->getSM()->getSoleItem();
}

const DLList *
Frame::getItems()
{
	return this->getSM()->getItems();
}



char 
executeRedrawRegion(Frame *frame)
{
	if(frame->resizeNecessaryP()){
		// potentially dangerous loop here
		frame->conditionallyInitialiseCanvas();
		// this says -- call this workproc again because
		// we didnt do our work -- we resized
		// the canvas so that things would fit
		// and we want to come back here when the xwindows
		// stuff is done
		return 0;
	} else {
		if(frame->managedP()){
			// only draw in the visible area
			Widget sw = frame->view->scrolled_window;
			iBox truncated_region, window_region = visibleRectangle(sw);
		
			i_IntersectBox(window_region, frame->redraw_region, &truncated_region);
			if (truncated_region.w > 0 && truncated_region.h > 0)
				frame->drawRegionNow(truncated_region);
		}
	}
	
	frame->redraw_region.w = 0;
	return 1; /* says not to call this workproc again */
}

void
Frame::addRedrawRegion(iBox *b)
{
	if (this->redraw_region.w) {
		i_UnionBox(*b, this->redraw_region, &this->redraw_region);
	} else {
		this->redraw_region = *b;
		this->redraw_procid = 
			XtAppAddWorkProc(this->view->main_window->app->app_context,
							 executeRedrawRegion, this);
	}
}

void 
Frame::flushRedrawRegion()
{
	if (this->redraw_region.w == 0) 
		return;
	XtRemoveWorkProc(this->redraw_procid);
	this->redraw_region.w = 0;
}


void
Frame::drawRegion(iBox box)
{
	this->addRedrawRegion(&box);
}

aString
Frame::getName()
{
	return this->name;
}
