
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANHE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"
#include "rafutil.h"
#include "illum.h"
#include "smemory.h"

#include "MainWindow.h"
#include "Mode.h"
#include "ViewCanvas.h"
#include "EntityAux.h"
#include "HierPage.h"
#include "HE.h"
#include "misc/Unicode.h"
#include "utf.h"

#define WINMAR	4
#define TRISIZE 5
#define SELWID 18
#define TRIMAR (WINMAR + 2 + (TRISIZE * 2))
#define XYLEN (6 * 4)
#define CONTLEN 30

HE::HE(Page *input_page,iEntityPtr input_entity,int ypos,int showval,int lev)
	: EntityFace(input_page, input_entity),
	y(ypos),
	show(showval),
	ind(lev)
{
}

HE::~HE()
{
}

void
HE::redraw()
{
	HierPage *hpage = (HierPage *) this->page;
	ViewCanvas *vc = this->page->getCanvas();
	Widget widget = vc->canvas;

	if (hpage->stale) 
		return;
int leftover = vc->pix_voffset%hpage->ch;
int y = this->y - (vc->pix_voffset - leftover);
	// this sucks John -- order a redisplay, don't
	// use XClearArea
	XClearArea(XtDisplay(widget),XtWindow(widget),
			   0, y,
			   this->page->width,
			   this->page->height - y,1);
	hpage->stale = 1;
}


void 
HE::deleteCheap()
{
	delete this;
}

void HE::deleteAndUpdate()
{
}

void HE::create()
{
}

void 
HE::willUpdate()
{
}

#define ARROWLEN 12
#define ARROWPNT 5

static void drawSel(Display *display, Pixmap pixmap,GC gc,int x,int y,int gray)
{
	XPoint tri[5];
	int len;
	
	x += 2;
	if (gray) {
		tri[0].x = x;				tri[0].y = y - 1 - ARROWPNT;
		tri[1].x = x + ARROWLEN;	tri[1].y = y - 1 - ARROWPNT;
		len = 2;
	} else {
		tri[0].x = x;						tri[0].y = y - 1 - ARROWPNT;
		tri[1].x = x + ARROWLEN;			tri[1].y = y - 1 - ARROWPNT;
		tri[2].x = x + ARROWLEN - ARROWPNT;	tri[2].y = y - 1 - (ARROWPNT * 2);
		tri[3].x = x + ARROWLEN;			tri[3].y = y - 1 - ARROWPNT;
		tri[4].x = x + ARROWLEN - ARROWPNT;	tri[4].y = y - 1;
		len = 5;
	}
	XDrawLines(display,pixmap,gc,tri,len,CoordModeOrigin);
}

void HE::drawAnEntity(Display *display,Pixmap pixmap,GC gc,
		int cw,int ch,int asc,int maxlen)
{
	char *str,cont[CONTLEN + 1],drawbuf[256];
	iUCode ubuf[35];
	iBox b = i_GetEnclosingBox(entity);
	int len;
	int2 valid;
	iPosSetPtr ps;
	XPoint tri[4];
	HierPage *page = (HierPage *) this->page;
	int yp = y + asc - page->getCanvas()->pix_voffset;
	
	if (i_GetChildEntity(entity)) {
		if (show) {
			tri[0].x = WINMAR;					tri[0].y = yp - 1 - TRISIZE;
			tri[1].x = WINMAR + (TRISIZE * 2);	tri[1].y = yp - 1 - TRISIZE;
			tri[2].x = WINMAR + TRISIZE;		tri[2].y = yp - 1;
			tri[3].x = WINMAR;					tri[3].y = yp - 1 - TRISIZE;
		} else {
			tri[0].x = WINMAR + TRISIZE;		tri[0].y = yp - 1 - (TRISIZE * 2);
			tri[1].x = WINMAR + TRISIZE;		tri[1].y = yp - 1;
			tri[2].x = WINMAR + (TRISIZE * 2);	tri[2].y = yp - 1 - TRISIZE;
			tri[3].x = WINMAR + TRISIZE;		tri[3].y = yp - 1 - (TRISIZE * 2);
		}
		XDrawLines(display,pixmap,gc,tri,4,CoordModeOrigin);
	}
	
	if (getEntitySelection(entity) == scITEM)
		drawSel(display,pixmap,gc,TRIMAR,yp,0);
	else if (!show && page->containsSelection(entity))
		drawSel(display,pixmap,gc,TRIMAR,yp,1);
	str = (char*)i_TypeToName(i_GetEntityType(entity));
	aString as = str;
	uString us;
	us.getUnicode(as);

	// just to get the font right
	drawUnicode(display,pixmap,gc,
				TRIMAR + SELWID + (ind * cw),yp,
				us.string(), strlen(str));

	sprintf(drawbuf,"%4d,%4d,%4d,%4d %s ",b.x,b.y,b.w,b.h,
				(i_GetLogicalType(entity) == iAnd?"And":" Or"));
	XDrawString(display,pixmap,gc,TRIMAR + SELWID + (maxlen * cw),yp,drawbuf,strlen(drawbuf));
	len = 0;
	switch (i_GetEDataType(entity)) {
	case iDPosSet:
	case iD1PosSet:
		utf2ucs(ubuf,"(PosSet)   ");
		len += ucslen(ubuf);
		ubuf[len++] = '{';
		ps = i_GetPosSet(entity,&valid);
		if (valid > 30) valid = 27;
		while (valid--) {
			ubuf[len++] = ps->chr;
			ps++;
		}
		ps = i_GetPosSet(entity,&valid);
		if (valid > 30) {
			ubuf[len++] = '.';
			ubuf[len++] = '.';
			ubuf[len++] = '.';
		}
		ubuf[len++] = '}';
		ubuf[len] = 0;
		break;
	case iDString:
		utf2ucs(ubuf,"(String)   ");
		len += ucslen(ubuf);
		strncpy(cont,i_GetText(entity,0),CONTLEN);
		utf2ucs(&ubuf[len],cont);
		break;
	case iDUString:
		utf2ucs(ubuf,"(UString)  ");
		len += ucslen(ubuf);
		ucsncpy(&ubuf[len],i_GetUText(entity,0),CONTLEN);
		break;
	case iDShort:
		utf2ucs(ubuf,"(Long)     ");
		len += ucslen(ubuf);
		sprintf(cont,"%d",i_GetShort(entity,0));
		utf2ucs(&ubuf[len],cont);
		break;
	case iDConf:
		utf2ucs(ubuf,"(Conf)     ");
		len += ucslen(ubuf);
		sprintf(cont,"%d",i_GetConf(entity));
		utf2ucs(&ubuf[len],cont);
		break;
	case iDNo:
		utf2ucs(ubuf,"(No Data) ");
		len += ucslen(ubuf);
		break;
	case iDUserData:
		utf2ucs(ubuf,"(UserData) ");
		len += ucslen(ubuf);
		break;
	}
	drawUnicode(display,pixmap,gc,TRIMAR + SELWID + ((maxlen + XYLEN) * cw),
		yp,ubuf,ucslen(ubuf));
}

void 
HE::willMove()
{
}

/* 
*/
void 
HE::destroy()
{
	this->redraw();
}

void 
HE::update()
{
	HierPage *page = (HierPage *) this->page;
	
	page->refreshSel(entity);
}

void 
HE::move()
{
}

int HE::inTri(int x)
{
	return x < TRIMAR && i_GetChildEntity(entity);
}

void 
HE::select()
{
	HierPage *page = (HierPage *) this->page;
	
	page->refreshSel(entity);
}

void 
HE::unSelect()
{
	HierPage *page = (HierPage *) this->page;
	
	page->refreshSel(entity);
}

void 
HE::recognize()
{
}

MainWindow *
HE::getMainWindow()
{
	return this->page->mode->main_window;
}

