
/*
 * Copyright (c) 1994, 1995 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#define CORRECT_LINE_DEFINITION


#include "illum.h"

#include <X11/StringDefs.h>
#include <X11/Intrinsic.h> 
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>
#include <X11/cursorfont.h>

#include "rafutil.h"
#include "smemory.h"

#include "misc/Unicode.h"
#include "misc/XArg.h"

#include "scroll.h"
#include "font.h"
#include "MainWindow.h"

#include "ViewCanvas.h"
#include "View.h"
#include "EntityAux.h"
#include "MainWindow.h"
#include "Document.h"
#include "SelectionManager.h"

#include "HE.h"
#include "HierFrame.h"

#define TRISIZE 5
#define CHARW 9
#define CHARH 15

void HierFrame::initHEs(iEntityPtr en,int ind,int shown)
{
	HE *he;
	EntityAux *ea;
	iEntityPtr cur;
	int len;
	
	ea = getController(en);
	he = new HE(this,en,this->nlines * ch,ea->show,ind);
	eaSetFace(en,vtHIER,he);
	if (shown) this->nlines++;
	if (this->maxlev < ind) this->maxlev = ind;
	len = strlen(i_TypeToName(i_GetEntityType(en)));
	if (this->maxlen < len) this->maxlen = len;

	for (cur = i_GetChildEntity(en);cur;cur = i_GetNextEntity(cur))
		this->initHEs(cur,ind+1,shown && he->show);
}

HierFrame::HierFrame(View *input_view, 
					 const aString& label, 
					 iEntityPtr input_root_entity)
	: Frame(label, input_view),
	  initialized_canvas(0), 
	  root_entity(input_root_entity),
	  maxlev(0),
	  nlines(0),
	  maxlen(0)
{
	ViewCanvas *vc;
	iUCode uni_m[2] = { 'M', 0};
	long h;

	vc = this->getCanvas();
	vc->measureText(uni_m, 1, &this->cw, &this->ch, &this->asc);
	initHEs(input_root_entity,0,1);
	this->maxlen += 2; // what the heck
	
	h = (this->nlines + 1) * ch;
	// oh yeah this makes sense
	if (h > 32030) this->height = 32030;
	else this->height = h;
	
	// Gee! (6 * 4) + 5 + 30 -- I'll have to remember that next time
	this->width = (this->maxlen + (6 * 4) + 5 + 30) * cw;
	this->stale = 0;
}

void HierFrame::delHEs(iEntityPtr en)
{
	iEntityPtr cur;
	HE *he = (HE*)eaGetFace(en,vtHIER);
	EntityAux *ea = getController(en);
	
	ea->show = he->show;
	eaSetFace(en,vtHIER,0);
	delete he;
	for (cur = i_GetChildEntity(en);cur;cur = i_GetNextEntity(cur))
		delHEs(cur);
}

HierFrame::~HierFrame()
{
	this->flushRedrawRegion();
	delHEs(root_entity);
}

int
HierFrame::resizeNecessaryP()
{
	return 0;
}

void 
HierFrame::conditionallyInitialiseCanvas()
{
	ViewCanvas *vc;
	
	if(this->initialized_canvas == FALSE) {
		this->initialized_canvas = TRUE;
		this->configureCanvas(this,this->width,this->height);
	}
	
	vc = this->getCanvas();
	vc->setClient(this);
}

void 
HierFrame::redisplay(XmDrawingAreaCallbackStruct *call_data)
{
	Widget w = this->getCanvas()->canvas;
	XEvent *event = (XEvent *) call_data->event;
	
	static Region region = NULL;
	
	if(XtIsManaged(w) == FALSE)
		return;
	
	if(!region)
		region = XCreateRegion();
	
	XtAddExposureToRegion(event, region);
	
	if(event->xexpose.count != 0)
		return;
	
	iBox refresh_box;
	
	XRectangle rect;
	XClipBox(region, &rect);
	XDestroyRegion(region);
	region = 0;
	
	refresh_box.x = rect.x;
	refresh_box.y = rect.y;
	refresh_box.w = rect.width;
	refresh_box.h = rect.height;
	
	if(refresh_box.w == 0 || refresh_box.h == 0)
		return;
	
	this->drawRegion(refresh_box);
}

#define SORRY "This window can't scroll farther, Sorry for the inconvenience"

int
itsTheBogusNewline(iEntityPtr entity)
{
	iEntityPtr parent;
	iEntityPtr grand_parent;
	iEntityPtr bogus_newline;

	return (
		entity && 
		(parent = i_GetParentEntity(entity)) && 
		((grand_parent = i_GetParentEntity(parent)) == 0) && 
		((bogus_newline = i_GetLastEntity(parent)) == entity)
		);
}

int HierFrame::drawentities(Display *display,Pixmap pixmap,GC gc,
		iBox *bp,iEntityPtr en)
{
	iEntityPtr cur;
	HE *he = (HE*)eaGetFace(en,vtHIER);
	int keepgoing;
	
	if (he->y > 32000) {
		XDrawString(display,pixmap,gc,4,he->y + asc,SORRY,strlen(SORRY));
		return 0;
	}
	if (((he->y + ch) > bp->y) && itsTheBogusNewline(he->entity) == FALSE)
		he->drawAnEntity(display,pixmap,gc,cw,ch,asc,maxlen);
	if (he->y > (bp->y + bp->h))
		return 0;
	keepgoing = 1;
	if (he->show) {
		for (cur = i_GetChildEntity(en);cur && keepgoing;cur = i_GetNextEntity(cur))
			keepgoing = drawentities(display,pixmap,gc,bp,cur);
	}
	return keepgoing;
}

void
HierFrame::refreshSel(iEntityPtr en)
{
	ViewCanvas *vc = this->getCanvas();
	Widget widget = vc->canvas;
	HE *he,*ce;
	iEntityPtr cur;
	
	he = (HE*)eaGetFace(en,vtHIER);
	for (cur = i_GetParentEntity(en);cur;cur = i_GetParentEntity(cur)) {
		ce = (HE*)eaGetFace(cur,vtHIER);
		if (!ce->show) he = ce;
	}
	XFillRectangle(XtDisplay(widget),XtWindow(widget), 
				   vc->blankGC(),
				   0, he->y,
				   this->width, ch);
	he->drawAnEntity(XtDisplay(widget),XtWindow(widget),vc->blackGC(),
		cw,ch,asc,maxlen);
}

void
HierFrame::drawRegionNow(iBox abox)
{
	ViewCanvas *vc = this->getCanvas();
	Widget widget = vc->canvas;
	long h;
	
	XFillRectangle(XtDisplay(widget),XtWindow(widget), 
				   vc->blankGC(),
				   abox.x, abox.y,
				   abox.w, abox.h);
	if (stale) {
		this->nlines = 0;
		updateHEs(root_entity);
		h = (this->nlines + 1) * ch;
		if (h > 32030) this->height = 32030;
		else this->height = h;
		this->stale = 0;
		this->configureCanvas(this,this->width,this->height);
	}
	drawentities(XtDisplay(widget),XtWindow(widget),vc->blackGC(),
		&abox,this->root_entity);
}

int HierFrame::containsSelection(iEntityPtr en)
{
	iEntityPtr cur;
	
	if (getEntitySelection(en) == scITEM)
		return 1;
	for (cur = i_GetChildEntity(en);cur;cur = i_GetNextEntity(cur))
		if (containsSelection(cur))
			return 1;
	return 0;
}

void HierFrame::updateHEs(iEntityPtr en)
{
	HE *he;
	iEntityPtr cur;
	
	he = (HE*)eaGetFace(en,vtHIER);
	he->y = this->nlines * ch;
	this->nlines++;
	if (he->show)
		for (cur = i_GetChildEntity(en);cur;cur = i_GetNextEntity(cur))
			updateHEs(cur);
}

iEntityPtr HierFrame::FindEntity(iEntityPtr en,int y)
{
	iEntityPtr cur;
	HE *he = (HE*)eaGetFace(en,vtHIER);
	
	if (he->y > 32000) return 0;
	if (he->y < y && ((he->y + ch) > y)) return en;
	if (he->show)
		for (cur = i_GetChildEntity(en);cur; cur = i_GetNextEntity(cur))
			if (en = FindEntity(cur,y))
				return en;
	return 0;
}

iEntityPtr 
HierFrame::findEnclosingEntity(int x, int y)
{
	return FindEntity(root_entity,y);
}

void 
HierFrame::addNewEntity(iEntityPtr entity)
{
}

void
HierFrame::prevEntity()
{
	upEntity();
}

void
HierFrame::findNextFlagEntity()
{	
}

void
HierFrame::nextEntity()
{
	downEntity();
}

int 
parentsShow(HE *he)
{
	int show = TRUE;
	iEntityPtr parents;

	parents = i_GetParentEntity(he->entity);
	while(show && parents){
		he = (HE *) eaGetFace(parents, vtHIER);
		show = he && he->show;
		parents = i_GetParentEntity(parents);
	}

	return show;
}

void
centerIfNecessary(HierFrame *frame, iEntityPtr entity)
{
	HE *he;
	
	he = (HE *) eaGetFace(entity, vtHIER);
	if(he){
		Widget sw = frame->view->scrolled_window;
		iBox vr = visibleRectangle(sw);
		if(visiblePositionP(sw, vr.x, he->y) == FALSE ||
		   visiblePositionP(sw, vr.x, he->y + 20) == FALSE)
			frame->centerOnSelected(entity);
	}
}

void
HierFrame::upEntity()
{

	iEntityPtr sole_entity;
	iEntityPtr left_entity=0;
	Ref r;
	int good = FALSE;

	const DLList *selection=this->getSM()->getItems();

	r = dllistLast(selection);
	if (r)
		sole_entity = (iEntityPtr) dllistContents(selection, r);
	else
		sole_entity=0;

	if(sole_entity) {
		left_entity = sole_entity;
		do {
			left_entity = i_FindEntityBack(left_entity, 0, ALL, 0);
			if(left_entity){
				HE *he = (HE *) eaGetFace(left_entity, vtHIER);
				if(he && parentsShow(he)) {
					good = TRUE;
				}
			}
		} while(left_entity && good == FALSE);
		if (left_entity && good && itsTheBogusNewline(left_entity) == FALSE) {
			this->getSM()->setItem(left_entity);
			centerIfNecessary(this, left_entity);
		}
	}
}

void
HierFrame::downEntity()
{
	iEntityPtr sole_entity;
	iEntityPtr right_entity=0;
	Ref r;
	int good = FALSE;
	
	const DLList *selection=this->getSM()->getItems();

	r = dllistFirst(selection);
	if (r)
		sole_entity = (iEntityPtr) dllistContents(selection, r);
	else
		sole_entity=0;

	if(sole_entity) {
		right_entity = sole_entity;
		do {
			right_entity = i_FindEntityForward(right_entity, 0, ALL, 0);
			if(right_entity){
				HE *he = (HE *) eaGetFace(right_entity, vtHIER);
				if(he && parentsShow(he)) {
					good = TRUE;
				}
			}
		} while(right_entity && good == FALSE);
		if (right_entity && good && itsTheBogusNewline(right_entity) == FALSE) {
			this->getSM()->setItem(right_entity);
			centerIfNecessary(this, right_entity);
		} 
	}
}



void
HierFrame::centerOnSelected(iEntityPtr entity)
{
	HE *he,*ce;
	iEntityPtr cur;
	
	he = (HE*)eaGetFace(entity,vtHIER);
	for (cur = i_GetParentEntity(entity);cur;cur = i_GetParentEntity(cur)) {
		ce = (HE*)eaGetFace(cur,vtHIER);
		if (!ce->show) he = ce;
	}
	scrollToPosition(this->view->scrolled_window, 0, he->y);
}

void HierFrame::changeEntity(iEntityPtr en)
{
	HE *he = (HE*)eaGetFace(en,vtHIER);
	he->redraw();
}


int 
HierFrame::pressKey(iUCode key_value)
{
	if(this->getSM()->getState() == scNOT){
		ringTheBell();
		return 0;
	}		
	if (key_value == '\b') {
		this->view->main_window->deleteSelection(FALSE);
		return 1;
	} else if(key_value == 20){
		killContentsForSelection();
	} else {
		ringTheBell();
		return 0;
	}
}

void
HierFrame::pressButton(XEvent *event)
{
    iEntityPtr entity;
	HE *he;
	
    switch(event->xbutton.button) {
	case Button1:
		entity = this->findEnclosingEntity(event->xbutton.x, event->xbutton.y);
		if(itsTheBogusNewline(entity))
		   return;
		if(entity != 0) {
			if(event->xbutton.state & ShiftMask){
				this->toggleItem(entity);
			} else {
				he = (HE*)eaGetFace(entity,vtHIER);
				if (he->inTri(event->xbutton.x)) {
					he->show = !he->show;
					this->changeEntity(entity);
				} else this->setItem(entity);
			}
		} else {
			this->clear();
		}
		break;
	case Button2:
		break;
    }
}

void
HierFrame::releaseButton(XEvent *event)
{
}

void
HierFrame::moveButton(int state, int x, int y)
{
}

void
HierFrame::manage()
{
	this->Frame::manage();
}

void
HierFrame::unmanage()
{
	this->Frame::unmanage();
}

/* these useless magnification functions are virtual requirements */
void 
HierFrame::setMag(int x, int ox)
{
}

void 
HierFrame::_setMag(int x, int ox)
{
}
