
/*
 * Copyright (c) 1994, 1995 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"
#include "illum.h"

#include "smemory.h"
#include "ucs.h"
#include "rafutil.h"

#include "Mode.h"
#include "misc/bb.h"
#include "EntityAux.h"

#include "ImagePage.h"
#include "MainWindow.h"
#include "ChooseDialog.h"

#include "IE.h"

static iUCode *
getTextReally(const iEntityPtr entity)
{
	const iPosSet *ps;
	iPosSet mini_ps;
	int2 n;
	iUCode c;
	iUCode *result;
	int i;
	aString as;
	uString us;
	
	switch(i_GetEDataType(entity)) {
    case iDPosSet:
		ps = i_GetPosSet(entity, &n);
		if (n==0) {
			mini_ps.chr=NON_REC_CHAR;
			ps=&mini_ps;
			n=1;
		}
		result=(iUCode *)smalloc(sizeof(iUCode)*(n+1));
		if (!result) halt("Can't allocate memory");
		for(i = 0; i < n; i++)
			result[i]=ps[i].chr;
		result[n]=0;
		break;
    case iDString:
		as = i_GetText(entity,0);
		us = uString(as);
		result = (iUCode *) smalloc((us.length() + 1) * sizeof(iUCode));
		if (!result) halt("Can't allocate memory");
		ucscpy(result,us.string());
		break;
    case iD1PosSet:
		c = i_GetGlyph(entity,0);
		if(c == 0) c=NON_REC_CHAR;
		result=(iUCode *)smalloc(sizeof(iUCode)*(2));
		if (!result)
			halt("Can't allocate memory");
		result[0]=c;
		result[1]=0;
		break;
	case iDUString:
		us = i_GetUText(entity,0);
		result = (iUCode *) smalloc((us.length() + 1) * sizeof(iUCode));
		if (!result) halt("Can't allocate memory");
		ucscpy(result,us.string());
		break;
    default:
		result=NULL;
		break;
	}
	
	return result;
}

IE::IE(ImagePage *input_page, iEntityPtr input_entity)
	: EntityFace(input_page, input_entity)
{
	this->real_text = getTextReally(this->entity);
}


IE::~IE()
{
	if (this->real_text)
		smfree(this->real_text);
}

void
IE::deleteAndUpdate()
{
	delete this;
}

void
IE::deleteCheap()
{
	delete this;
}

/* 
// should blank
// bogus -- should do nothing??
*/
void 
IE::willUpdate()
{
	iBox redraw_box;
	
	ImagePage *ipage = (ImagePage *) this->page;
	
	redraw_box = ipage->entityBounds(entity); 
	
	ipage->drawRegion(promoteBox(redraw_box));
}


/* 
// should draw directly
*/

void 
IE::update()
{
	extern int ancestorFlagged(iEntityPtr);
	int parent_flagged=ancestorFlagged(i_GetParentEntity(this->entity));
	int chosen=this->getMainWindow()->chooser->qualify(this->entity,
													   &parent_flagged);
	ImagePage *ipage = (ImagePage *) this->page;
	
	if(chosen == FALSE)
		ipage->invalid_guy_added = TRUE;

	if(this->real_text)
		smfree(this->real_text);
	this->real_text = getTextReally(this->entity);
	
	if(this->page->managed){
		ImagePage *ipage = (ImagePage *) this->page;
		iBox redraw_box;
		
		redraw_box=ipage->entityBounds(entity); 
		
		ipage->drawRegion(promoteBox(redraw_box));
	}
}

void 
IE::willMove()
{
	ImagePage *ipage = (ImagePage *) this->page;
	ipage->tree.deleteItem(entity);
	this->willUpdate();
}

void 
IE::move()
{
	ImagePage *ipage = (ImagePage *) this->page;
	ipage->tree.addItem(entity);
	this->update();
}

/* 
// should blank
*/
void 
IE::destroy()
{
	ImagePage *ipage = (ImagePage *) this->page;
	
	// bogus refresh should be my own function
	ipage->refresh(this);
	ipage->deleteEntity(entity);
}


/*
// should draw directly
*/
void 
IE::select()
{
	ImagePage *ipage = (ImagePage *) this->page;
	
	ipage->refreshBorder(entity);
}

/*
// should draw directly
*/
void 
IE::unSelect()
{
	ImagePage *ipage = (ImagePage *) this->page;
	
	ipage->refreshBorder(entity);
}


void 
IE::recognize()
{
	if(i_GetEntityType(entity) == iGlyphT){
		// will reset the label
		// bogus hmmm -- why dont I directly call update
		this->update();
	} else {
		ImagePage *ipage = (ImagePage *) this->page;
		
		ipage->refreshBorder(entity);
	}
}

MainWindow *
IE::getMainWindow()
{
	return this->page->mode->main_window;
}


void 
IE::create()
{
	// do nothing
}

void
iefriend() {}
