
/*
 * Copyright (c) 1994, 1995 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"
#include "illum.h"

#include "c-type.h"

#include <X11/StringDefs.h>
#include <X11/Intrinsic.h> 
#include <X11/Xutil.h>
#include <Xm/Xm.h>
#include <Xm/DrawingA.h>
#include <Xm/PushB.h>

#include "smemory.h"
#include "ucs.h"

#include "rafutil.h"

#include "misc/color.h"
#include "misc/XArg.h"
#include "misc/Unicode.h"


#include "misc/scroll.h"
#include "misc/bb.h"
#include "mag.h"

#include "QuadTree.h"
#include "EntityAux.h"
#include "Mode.h"
#include "MainWindow.h"
#include "ViewCanvas.h"
#include "Document.h"
#include "ChooseDialog.h"
#include "ChooseDialog.h"
#include "SelectionManager.h"
#include "fudge.h"
#include "IE.h"
#include "ImagePage.h"


#ifdef RAF_INTERNAL

#include "form.h"
extern int raf_display;
#define SPS_LINE_D 1
#define SPS_PARA_D 2
#ifdef STEVE
#include "/raf/linus/home/steve/src/illum/sps.h"
#endif
#include "rafgraph.h"

#endif
/*
// these bounds are measured on the original because they are
// used in conjunction with the QuadTrees 
// we have to exaggerate the size of the label because
// the label stays the same size while the box grows or
// shrinks depending on the mag
// therefore we exaggerate the label size to be that
// when it is most dominating -- 1/4 mag
*/

iBox
boundsOnTheOriginal(ImagePage *page, iEntityPtr entity)
{
	iBox box;
	IE *ie; 

	if(!(ie = (IE *) eaGetFace(entity, vtIMAGE))) {
		box.h = box.w = 0;
		printf("in boundsOnTheOriginal, entity with no faces\n");
	} else {
		const iUCode *real_text = ie->real_text;

        	// compute width and height of label, including the box drawn around label
        	// when it is being edited, and the gap between the text and the box.

		// Note: this computation must match ImagePage::entityBounds, except
		// for the different scale factor used for text. 
        	int char_width=0, char_height=0, ascent=0;
        	if(real_text)
                	page->getCanvas()->measureText(real_text, ucslen(real_text),
                                                	&char_width, &char_height,
                                                	&ascent);
        	int tw = char_width  + 2*RECT_GAP + MAX_BOX_THICKNESS;
        	int th = char_height + 2*RECT_GAP + MAX_BOX_THICKNESS;

		// multiply tw and th by 4  to exaggerate label size as described above
		tw *= 4;  th *= 4;

		// get bounding box of entity without label
		box = i_GetEnclosingBox(entity);

        	// if label is wider than entity, enlarge box but keep same x center
        	if(tw > box.w) {
                	box.x = box.x + box.w/2 - tw/2;
                	box.w = tw;
        	}
        	// add height of label to bounding box
        	box.h += th;

		// add the refresh fudge factor, to be consistent with entityBounds
		// heinous - this should not be necessary, but probably won't hurt
        	box.x -= REFRESH_FUDGE;
        	box.y -= REFRESH_FUDGE;
        	box.w += 2*REFRESH_FUDGE;
        	box.h += 2*REFRESH_FUDGE;
	}
	
	return box;
}


// you are not allowed to create a new entity that is smaller than this
// in either dimension
#define LITTLE_PUNK_LIMIT 4

ImagePage::ImagePage(iImagePtr input_image, Mode *input_parent,
					   const aString& input_name)
	: Page(input_name, input_parent), image(input_image), 
	  tree((BoundsFunction) boundsOnTheOriginal, this),
	  first_x(-1), first_y(-1),
	  last_x(-1), last_y(-1),
	  thing(tgNOTHING),
	  invalid_guy_added(FALSE)
{
	this->xor_box.w = 0;
	this->go = 0;
	this->central_position.w = 0;
	dllistCtor(&this->select_list);
	iLBox ibox = {0,0,0,0};
	ibox.h = i_GetImageHeight(input_image);
	ibox.w = i_GetImageWidth(input_image);
	ibox.w += FUDGE_WIDTH;
	ibox.h += FUDGE_HEIGHT;
	this->setPicture(ibox);
}

ImagePage::~ImagePage()
{
	// the quad tree and the image should take care of themselves
	
	// the face list gets it in Page
	
	dllistDtor(&this->select_list);
}

// actually, it's itself or any ancestor
int ancestorFlagged(iEntityPtr entity)
{
	if (!entity) return FALSE;
	long flagged=(!i_GetLongProp(entity, "Flag", &flagged));
	if (flagged)
		return flagged;
	else
		return ancestorFlagged(i_GetParentEntity(entity));
}

void ImagePage::checkEntity(iEntityPtr entity)
{
	iEntityPtr member=0;
	DLList *list = &this->face_list;
	MainWindow *mw;
	Ref r;
	IE *ie;
	
	if(!(entity=pickpath(entity)))
		return;
	
	mw=this->mode->main_window;
	
	// bogus, it seems to me that the chooser should be
	// attached to the page or mode, not the main window - Steve
	int parent_flagged=ancestorFlagged(i_GetParentEntity(entity));
	int chosen=mw->chooser->qualify(entity,&parent_flagged);
	
	for(r = dllistFirst(list); r; dllistNext(list, &r)){
		ie = (IE *) dllistContents(list, r);
		member = ie->entity;
		if (member==entity) break;
	}
	iLBox box;
	if (member==entity && !chosen) {
		// delete member from page
		this->removeFace(ie);
		if(this->tree.initialized()) {
			this->tree.deleteItem(entity);
		}
		box = promoteBox(this->entityBounds(entity));
		this->drawRegion(box);
	} else if (member!=entity && chosen) {
		// good this will create a face
		this->addMember(entity);
		if(this->tree.initialized())
			this->tree.addItem(entity);
		// bogus warning will robinson -- could redisplay stuff piecemeal
		// lets check the drawRegion function for intelligence
		box = promoteBox(this->entityBounds(entity));
		this->drawRegion(box);
	}
	
	iEntityPtr child;
	// since flagging mode affects children, have to change all the children
	for (child=i_GetChildEntity(entity); child; child=i_GetNextEntity(child))
		this->checkEntity(child);
}

void
ImagePage::deleteMember(iEntityPtr entity)
{
	IE *face = (IE *) eaGetFace(entity, vtIMAGE);
	this->removeFace(face);
}

void
ImagePage::addMember(iEntityPtr member)
{
	IE *ie;
	ie = new IE(this, member);
	this->addFace(ie);
	eaSetFace(member, vtIMAGE, ie);
}

Ref
ImagePage::prevRef(iEntityPtr current)
{
	iEntityPtr prev_guy;
	IE *ie = 0;
	Ref r;

	prev_guy = current;
	while(prev_guy = i_FindEntityBack(prev_guy,0,PROPERTY,ILLUM_PROP)) {
		if(ie = (IE *) eaGetFace(prev_guy, vtIMAGE))
			break;
	}

	if (ie && (r=dllistSeek(&this->face_list, ie)))
		return r;
	
	return 0;
}



void
ImagePage::addNewMember(iEntityPtr member)
{
	
	int parent_flagged=ancestorFlagged(i_GetParentEntity(member));
	int chosen=this->mode->main_window->chooser->qualify(member,&parent_flagged);
	IE *ie;
	
	if(chosen == FALSE)
		this->invalid_guy_added = TRUE;
	
	ie = new IE(this, member);

	// prev ref is in the face list now
	Ref r=this->prevRef(member);

	
	if (r)
		dllistInsertAfter(&this->face_list, r, ie);
	else 
		dllistAppend(&this->face_list, ie);
	
	eaSetFace(member, vtIMAGE, ie);
}

int
ImagePage::pressKey(iUCode key_value)
{
	if(this->getSM()->getState() == scNOT){
		ringTheBell();
		return 0;
	}		
		
	if (key_value == '\b') {
		if (this->editing(this->edit_entity)) 
			this->reassignValues(this->getItems(), key_value);
		else 
			this->mode->main_window->deleteSelection(FALSE);
	} else if(key_value == 20) {
		if (this->editing(this->edit_entity)) 
			this->reassignValues(this->getItems(), key_value);
		else {
			killContentsForSelection();
		}
	} else {
		if(this->mode->main_window->text_disabled){
			ringTheBell();
			return 0;
		}
		if(key_value < ' ' && (key_value != '\b') && (key_value != 20)) {
			ringTheBell();
			return 0;
		}

                // bogus why does this make sense here?
		this->track(CENTER);
		this->reassignValues(this->getItems(), key_value);
	}

	return 1;
}

void
ImagePage::begin_edit(iEntityPtr entity)
{
	static char data[] = "ImagePage";

	if (!this->edit_entity)
        	this->edit_entity=entity;
	/* dont supply a data length, because it is transient */
	i_SetUserPtrProp(entity,EDITING_PROP, data, NULL);
}

int
ImagePage::editing(iEntityPtr entity)
{
	// assumes only editing one entity at a time
	// return (entity && (this->edit_entity == entity));

	// allows more than one entity to be edited at a time
	return (entity && i_FindProp(entity, EDITING_PROP));
}

void
ImagePage::end_edit(iEntityPtr entity)
{
	// heinous - this operation appears to duplicate one in
	// OOCPage.  We should reconcile the two and consider moving
	// the method up to the Page level.

	if(editing(entity)) 
		i_DisposeProp(entity, EDITING_PROP);
}

void
ImagePage::end_all_edits()
{
	if (this->edit_entity) {
		const DLList *list = this->getItems();
        	for(Ref r = dllistFirst(list); r; dllistNext(list, &r)){
                	iEntityPtr entity = (iEntityPtr) dllistContents(list, r);
			this->end_edit(entity);
        	}
        	this->edit_entity=0;
	}
}

void
ImagePage::reassignValues(const DLList *list, iUCode key_value)
{
	// override the Page level reassignValues with a more complicated
	// version which allows replacing a glyph's label with a string

        Ref r;
        iEntityPtr entity;
        Mode *v;
		iUCode buf[2];

        int changed=FALSE;

        v = this->mode;
        for(r = dllistFirst(list); r; dllistNext(list, &r)){
                entity = (iEntityPtr) dllistContents(list, r);
		switch (i_GetEntityType(entity)) {
			case iGlyphT: 
			default: {
				iUCode *txt;
				iUCode uc[2];
				if (editing(entity) && !i_GetUText(entity,0)) {
					// convert glyph label to text label
					short conf;
					uc[0] = i_GetGlyph(entity,&conf);
					uc[1] = '\0';
					i_SetUText(entity,uc,0);
					changed = TRUE;
				} else if (!editing(entity) && (txt=(iUCode*)i_GetUText(entity,0))) {
					// delete existing text string label
					uc[0] = '\0';
					i_SetUText(entity,uc,0);
					changed = TRUE;
				}
				if (txt=(iUCode*)i_GetUText(entity,0)) {
					// edit text label
					// for now just append what user types
					const int len = ucslen(txt);	
					if(!editing(entity)) begin_edit(entity);
					switch (key_value) {
						case '\b':  
							if (len == 0) {
								i_SetNoData(entity);
								end_edit(entity);
							} else if(len > 0) 
								i_DelUni(entity,len-1);
							break; 
						case 20:
							  buf[0] = 0;
							i_SetUText(entity,buf,0);
							break;
						default: 
							i_AddUni(entity,len,key_value);
							break;
					}
					changed = TRUE;
				} else {
					begin_edit(entity);
                        		i_SetGlyph(entity, key_value,CERTCONF);
                        		i_DisposeProp(entity, FLAG_PROP);
					changed = TRUE;
				}
				break;
			}
		}
        }
        // we may have been deleted from underneath
        if (changed)
                v->main_window->document->setDirty();
}



/*
// the input is two lists one is the list of things i have
// to refresh, the other is the list of all things selected
// i output things that i have to refresh that are selected
// and things that i have to refresh that arent selected
*/
void
splitSelected(DLList *list, 
			  const DLList *selection, 
			  DLList *selected,
#ifndef RAF_INTERNAL
			  DLList *unselected)
#else
DLList *unselected,
	DLList *grot_selected,
	DLList *grot_unselected)
#endif
{
	Ref r;
	VoidPtr item;
	
	for(r = dllistFirst(list); r; dllistNext(list, &r)) {
		item = dllistContents(list, r);
		
		if(dllistSeek(selection, item)) 
			dllistAppend(selected, item);
		else
			dllistAppend(unselected, item);
	}
}

void p_list(DLList *l)
{
	Ref r;
	int data;
	
	for(r = dllistFirst(l); r; dllistNext(l, &r)){
		data = (int) dllistContents(l, r);
		printf("0x%x  ",data);
	}
	printf("\n");
	
	
}

static int4
compareEntType(VoidPtr a, VoidPtr b)
{
	iEntityPtr ent_a = (iEntityPtr) a;
	iEntityPtr ent_b = (iEntityPtr) b;
	int ta = i_GetEntityType(ent_a);
	int tb = i_GetEntityType(ent_b);

	return ( ta < tb 
			 ? 1
			 : ( ta > tb 
				 ? -1
				: 0
				 )
			 );
}

void
ImagePage::drawRegionNow(iLBox refresh_box)
{
	ViewCanvas *vc = this->getCanvas();
	Widget w = vc->canvas;
	int mag_exp = this->getMagExp();
	int mag_image_exp = 16 + mag_exp;
	
	if(this->image == 0){
		return;
	}
	
	if(XtIsManaged(w) == FALSE)
		return;
	
	Display *display = XtDisplay(w);
	Drawable drawable = RootWindowOfScreen(XtScreen(w));
	Pixmap pixmap;
	int depth = DefaultDepthOfScreen(XtScreen(w));
	iImagePtr image = this->image;
	int2 dw;
	char *bits = 0;
	iLBox tbox,box,offb;
	// no longer scaled automatically for us
	int image_width =  magUpInt(i_GetImageWidth(image), mag_exp);
	int image_height = magUpInt(i_GetImageHeight(image), mag_exp);
	tbox = refresh_box;
	tbox.y -= TOPPAD;
	tbox.x -= LEFTPAD;
	
	pixmap = XCreatePixmap(display, drawable, 
						   refresh_box.w, refresh_box.h, 
						   depth);
	
	
	if (refresh_box.x < LEFTPAD) {
		box = refresh_box;
		box.w = LEFTPAD - box.x;
		XFillRectangle(display, pixmap, 
					   this->getCanvas()->grayGC(),
					   box.x-refresh_box.x,box.y-refresh_box.y,box.w, box.h);
	}
	if (refresh_box.y < TOPPAD) {
		box = refresh_box;
		box.h = TOPPAD - box.y;
		XFillRectangle(display, pixmap, 
					   this->getCanvas()->grayGC(),
					   box.x-refresh_box.x,box.y-refresh_box.y,box.w, box.h);
	}
	if (tbox.x < 0) {
		tbox.w += tbox.x;
		offb.x = tbox.x;
		tbox.x = 0;
	} else offb.x = 0;
	if (tbox.y < 0) {
		tbox.h += tbox.y;
		offb.y = tbox.y;
		tbox.y = 0;
	} else offb.y = 0;
	if (image_height < tbox.y+tbox.h) {
		tbox.h = image_height - tbox.y;
		box = refresh_box;
		box.y = image_height;
		box.h = refresh_box.y + refresh_box.h - box.y;
		XFillRectangle(display, pixmap, 
					   this->getCanvas()->grayGC(),
					   box.x-refresh_box.x,box.y-refresh_box.y,box.w, box.h);
	}
	if (image_width < tbox.x+tbox.w) {
		tbox.w = image_width - tbox.x;
		box = refresh_box;
		box.x = image_width;
		box.w = refresh_box.x + refresh_box.w - box.x;
		XFillRectangle(display, pixmap, 
					   this->getCanvas()->grayGC(),
					   box.x-refresh_box.x,box.y-refresh_box.y,box.w, box.h);
	}
	if (tbox.h > 0 && tbox.w > 0) {
		if (i_Image2BitMap(image, demoteBox(tbox), 1 << mag_image_exp, &dw, &bits))
			return;
		// create the image to be the size of the bitmap
		XImage *x_image = XCreateImage(display, 0, 1, XYBitmap, 
									   0, // tricky -- ignore initial bits
									   bits,
									   tbox.w, tbox.h, 
									   8, // quanta of scan line
									   dw);
		x_image->byte_order = MSBFirst;
		x_image->bitmap_bit_order = MSBFirst;
		_XInitImageFuncPtrs(x_image);
		
		// create the pixmap to same size
		
		// copy the entire image to the pixmap
		XPutImage(display,pixmap, this->getCanvas()->copyGC(), x_image, 
				  0,0, -offb.x,-offb.y, tbox.w, tbox.h);
		
		// we want to free the image ourselves, with smfree cause it was smalloced
		x_image->data=0;
		XDestroyImage(x_image);
		smfree(bits);
	}
	
	DLList redraw_list;
	DLList unselected_list;
	DLList selected_list;
#ifdef RAF_INTERNAL
	DLList grot_unselected_list;
	DLList grot_selected_list;
#endif
	
	dllistCtor(&redraw_list);
	dllistCtor(&selected_list);
	dllistCtor(&unselected_list);
#ifdef RAF_INTERNAL
	dllistCtor(&grot_selected_list);
	dllistCtor(&grot_unselected_list);
#endif
	

	iBox qbox;
	refresh_box.x -= LEFTPAD;
	refresh_box.y -= TOPPAD;
	qbox = demoteBox(refresh_box);
	bbScaleDown(qbox, this->getMagExp());
	qbox.x += -1;
	qbox.y += -1;
	qbox.w += 2;
	qbox.h += 2;
	
	this->tree.findOverlappingItems(qbox, &redraw_list);

	// sort them by type
	dllistSort(&redraw_list, compareEntType);
	
	// A we can speed up drawing but we haven't
	// B we can draw the selected guys last
	//    so they stand out usually
#ifndef RAF_INTERNAL
	splitSelected(&redraw_list, 
				  this->getSM()->getItems(),
				  &selected_list, 
				  &unselected_list);
#else
	splitSelected(&redraw_list, 
				  this->getSM()->getItems(),
				  &selected_list, 
				  &unselected_list,
				  &grot_selected_list, 
				  &grot_unselected_list);
#endif
	
	this->drawBoxes(display,
					pixmap, 
					refresh_box.x, refresh_box.y,
					&unselected_list, 
					FALSE);
	
#ifdef RAF_INTERNAL
	this->drawBoxes(display,
					pixmap, 
					refresh_box.x, refresh_box.y,
					&grot_unselected_list, FALSE);
#endif

	if(this->mode->main_window->text_disabled == FALSE) {
		this->drawText(display, 
					   pixmap, 
					   refresh_box.x, refresh_box.y,
					   &unselected_list, 
					   FALSE);
	
#ifdef RAF_INTERNAL
	this->drawText(display, 
				   pixmap, 
				   refresh_box.x, refresh_box.y,
				   &grot_unselected_list, 
				   FALSE);
#endif
	}
	
	this->drawPitch(display, 
					pixmap, 
					refresh_box.x, refresh_box.y,
					&unselected_list);
	this->redrawImages(display,
					   pixmap,
					   refresh_box.x, refresh_box.y,
					   &selected_list);
	this->drawBoxes(display, 
					pixmap, 
					refresh_box.x, refresh_box.y,
					&selected_list,
					TRUE);
	
#ifdef RAF_INTERNAL
	this->drawBoxes(display, 
					pixmap, 
					refresh_box.x, refresh_box.y,
					&grot_selected_list,
					TRUE);
#endif
	
	
	iEntityPtr sole_item = this->getSM()->getSoleItem();
	if(sole_item && dllistSeek(&selected_list, sole_item)){
		this->drawKnobs(display,
						pixmap,
						refresh_box.x, refresh_box.y,
						sole_item,
						TRUE);
	}
	
#ifdef RAF_INTERNAL
	if(sole_item && dllistSeek(&grot_selected_list, sole_item)){
		this->drawKnobs(display,
						pixmap,
						refresh_box.x, refresh_box.y,
						sole_item,
						TRUE);
	}
#endif
	
	if(this->mode->main_window->text_disabled == FALSE) {
		this->drawText(display, 
					   pixmap, 
					   refresh_box.x, refresh_box.y,
					   &selected_list,
					   TRUE);
#ifdef RAF_INTERNAL
		this->drawText(display, 
					   pixmap, 
					   refresh_box.x, refresh_box.y,

					   &grot_selected_list,
					   TRUE);
#endif
	}

	this->drawPitch(display, 
					pixmap, 
					refresh_box.x, refresh_box.y,
					&selected_list);
	
	dllistDtor(&redraw_list);
	dllistDtor(&selected_list);
	dllistDtor(&unselected_list);
	
	// redraw the XOR
	if(this->go)
		this->drawXOR(pixmap,refresh_box.x - vc->pix_hoffset + 10,
							refresh_box.y - vc->pix_voffset + 10);

	// this is the place we finally trim the
	// area to be exactly what the expose wants
	XCopyArea(display, pixmap, XtWindow(w), this->getCanvas()->copyGC(),
			  0, 0,
			  refresh_box.w, refresh_box.h,  
			 refresh_box.x-vc->pix_hoffset+LEFTPAD, refresh_box.y+TOPPAD-vc->pix_voffset);
	
	// cut way down on pixmaps
	// we are close to having no need for member variable
	XFreePixmap(display, pixmap);
	
#ifdef RAF_INTERNAL
	aString type;
		
	iEntityPtr line_entity = this->getSM()->getSoleItem();
	if(line_entity) {
		int entity_type = i_GetEntityType(line_entity);
		type = i_TypeToName(i_GetEntityType(line_entity));
		if((entity_type == iLineT || type=="LineSeg" ) && (raf_display & SPS_LINE_D)) {
			MakeRAFGraph(this,line_entity);
		}
		else if ((entity_type==iParagraphT || entity_type==iDocumentT) &&
		        (raf_display & SPS_PARA_D)) {
#ifdef STEVE
			MakeHorizSumsGraph(this,line_entity);
#endif
		}
	}
#endif
}

void
ImagePage::redisplay (XmDrawingAreaCallbackStruct *call_data)
{
	static Region region = NULL;
	XEvent *event = (XEvent *) call_data->event;
	iBox refresh_box;
	XRectangle rect;
	
	if (this->should_delete)
		return;
	
	Widget sw = this->mode->scrolled_window;
	iBox new_position;
	int was_moved = FALSE;
	
	new_position = demoteBox(visibleRectangle(sw));
	
	if(equalBoxes(this->central_position, new_position) == FALSE) {
		was_moved=TRUE;
		this->central_position = new_position;
	}
	
	if(!region)
		region = XCreateRegion();
	XtAddExposureToRegion(event, region);
	
	if(event->xexpose.count != 0)
		return;
	
	// erase the xor because it got uncovered somehow
	XClipBox(region, &rect);
	XDestroyRegion(region);
	region = 0;
	refresh_box.x = rect.x;
	refresh_box.y = rect.y;
	refresh_box.w = rect.width;
	refresh_box.h = rect.height;
	refresh_box.x += this->getCanvas()->pix_hoffset;
	refresh_box.y += this->getCanvas()->pix_voffset;
	// only draw in the visible area
	iBox window_region;
	iBox truncated_region;
	
	window_region = demoteBox(visibleRectangle(sw));
	i_IntersectBox(window_region, refresh_box, &truncated_region);
	if (truncated_region.w > 0 && truncated_region.h > 0)
		this->drawRegionNow(promoteBox(truncated_region));
	
	
}


static int4
compare_selection(VoidPtr a, VoidPtr b)
{
	iEntityPtr entity_a = (iEntityPtr) a;
	iEntityPtr entity_b = (iEntityPtr) b;
	iBox bounds;
	int a_size;
	int b_size;
	int a_type = i_GetEntityType(entity_a);
	int b_type = i_GetEntityType(entity_b);
	
	bounds = i_GetEnclosingBox(entity_a);
	a_size = bounds.w * bounds.h;
	bounds = i_GetEnclosingBox(entity_b);
	b_size = bounds.w * bounds.h;
	
	return( ((a_size < b_size) ?
			 -1 :
			 ((a_size > b_size) ?
			  1
			  : ((a_type > b_type) ?
				 -1 :
				 ((a_type < b_type) ?
				  1 :
				  0))))
		);
}
#define OVERLAP_DELTA 6
/*
// use type int so there is always a clear order
// and to avoid using aString too much
//
// be careful to make this useful for select and toggle
*/
void
ImagePage::findEnclosingEntities(int x, int y, DLList *overlap_list)
{   
	Ref r;
	iEntityPtr child;
	int mag_exp = this->getMagExp();
	DLList tree_list_temp;
	DLList *tree_list;
	iBox qbox;
	
	x = magDownInt(x, mag_exp);
	y = magDownInt(y, mag_exp);
	
	tree_list = &tree_list_temp;
	
	dllistCtor(tree_list);
	
	qbox.x = x;
	qbox.y = y;
	qbox.w = OVERLAP_DELTA;
	qbox.h = OVERLAP_DELTA;
	this->tree.findOverlappingItems(qbox, tree_list);
	
	for(r = dllistFirst(tree_list); r ; dllistNext(tree_list, &r)){
		iBox box;

		child = (iEntityPtr) dllistContents(tree_list, r);
		box = i_GetEnclosingBox(child);
		
		if(bbNullP(box) == FALSE){
			if(x + OVERLAP_DELTA >= box.x  &&
			   x - OVERLAP_DELTA <= (box.x + box.w) &&
			   y + OVERLAP_DELTA >= box.y &&
			   y - OVERLAP_DELTA <= (box.y + box.h)){
				dllistAppend(overlap_list, child);
			}
		}
	}
	
	dllistDtor(tree_list);

	dllistSort(overlap_list, compare_selection);
}

void
ImagePage::conditionallyInitialiseCanvas ()
{

	iImagePtr image;
	int mag_exp = this->getMagExp();

	if(this->getCanvas())
	        return;


	image = this->image;

	// no longer scaled automatically for us
	this->width =  magUpInt(i_GetImageWidth(image), mag_exp);
	this->height = magUpInt(i_GetImageHeight(image), mag_exp);

	// bogus fudge so we can write on the bottom of thescreen
	this->width+=FUDGE_WIDTH;
	this->height += FUDGE_HEIGHT;
	this->getCanvas()->initializeScrollBarValues();
	this->syncSize();
}


void 
ImagePage::_setMag(int mag_exp, int old_mag_exp)
{
	int mag_image_exp;
	int new_mag;
	iLBox b={0,0,0,0};
	mag_image_exp = 16;
	mag_image_exp += mag_exp;
	new_mag = 1 << mag_image_exp;

	// have to shift up
	this->width =  magUpInt(i_GetImageWidth(image), mag_exp);
	this->height = magUpInt(i_GetImageHeight(image), mag_exp);
	// bogus fudge so we can write on the bottom of the screen
	this->width +=  FUDGE_WIDTH;
	this->height += FUDGE_HEIGHT;
	b.w=this->width;
	b.h=this->height;
	this->setPicture(b);
	
	this->syncSize();
}

void 
ImagePage::setMag(int mag_exp, int old_mag_exp)
{
	int4 x, y;
	iEntityPtr sole_entity;
	
	sole_entity = this->getSM()->getSoleItem();
	
	getScrollPositionCenter(this->mode->scrolled_window, &x, &y);

	this->_setMag(mag_exp, old_mag_exp);
	
	// should do the one that loses information second
	x = magDownInt(magUpInt(x, mag_exp), old_mag_exp);
	y = magDownInt(magUpInt(y, mag_exp), old_mag_exp);
	
	// should make an update() member function??
	if(this->managedP()){
		this->unmanage();
		this->manage();
		if(sole_entity)
			centerOnSelected(sole_entity);
		else
			this->getCanvas()->scrollToPosition(this->mode->scrolled_window, x, y);
	}
}

static void setBox(iBox& rbox, int x, int y, int w, int h)
{
	rbox.x = x;
	rbox.y = y;
	rbox.w = w;
	rbox.h = h;
	if(x < 0) {
		rbox.x = 0;
		rbox.w += x;
	}
	if(y < 0){
		rbox.y = 0;
		rbox.h += y;
	}
}

static void
refreshText(ImagePage *page, Widget w, 
			IE *face, GContext font, iBox box)
{
	const iUCode *text;
	
	text = face->real_text;
	
	if(text && text[0]) {
		int tx, ty, tw, th;
		int font_ascent;
		
		page->getCanvas()->measureText(text, ucslen(text),
										&tw, &th,
										&font_ascent);
		
		tx = box.x + box.w/2 - tw/2;
		ty = box.y + box.h + th;
		
		iLBox refresh_box;
		refresh_box.x = tx - BLANK_FUDGE + LEFTPAD;
		refresh_box.y = ty - font_ascent - BLANK_FUDGE + TOPPAD;
		refresh_box.w = tw + 2 * BLANK_FUDGE;
		refresh_box.h = th + 2 * BLANK_FUDGE;
		page->drawRegion(refresh_box);
	}
}

static void
refreshTextConstant(ImagePage *page, Widget w, iEntityPtr entity, 
					GContext font, iBox box)
{
	const iUCode *text;
	
	IE *ie;
	ie = (IE *) eaGetFace(entity, vtIMAGE);
	
	text=ie->real_text;
	
	if(text && text[0]) {
		int tx, ty, tw, th;
		int font_ascent;
		
		page->getCanvas()->measureText(text,
										ucslen(text),
										&tw, &th,
										&font_ascent);
		
		tx = box.x + box.w/2 - tw/2;
		ty = box.y + box.h + th;
		
		iBox rbox;
		setBox(rbox,
			   tx - BLANK_FUDGE + LEFTPAD, 
			   ty - font_ascent - BLANK_FUDGE + TOPPAD,
			   tw + 2 * BLANK_FUDGE, th + 2 * BLANK_FUDGE);
		page->drawRegion(promoteBox(rbox));
	}
}


/*
// bogus this could be improved to use the actual text size 
*/
void
ImagePage::refreshBorder(iEntityPtr entity)
{
	
	iBox box;
	if(this->managedP() == 0)
		return;
	Widget w = this->getCanvas()->canvas;
	if(w == 0)
		return;
	box = i_GetEnclosingBox(entity);
	
	if(bbNullP(box) == TRUE)
		return;
	
	bbScaleUp(box, this->getMagExp());
	iBox rbox1,rbox2,rbox3,rbox4;
	
	setBox(rbox1,
		   box.x - REFRESH_FUDGE, box.y - REFRESH_FUDGE,
		   2*REFRESH_FUDGE, box.h + 2*REFRESH_FUDGE);
	setBox(rbox2,
		   box.x - REFRESH_FUDGE, box.y - REFRESH_FUDGE,
		   box.w + 2*REFRESH_FUDGE, 2*REFRESH_FUDGE);
	setBox(rbox3, 
		   box.x + box.w - REFRESH_FUDGE, box.y - REFRESH_FUDGE,
		   2*REFRESH_FUDGE, box.h + 2*REFRESH_FUDGE);
	setBox(rbox4, 
		   box.x - REFRESH_FUDGE, box.y + box.h - REFRESH_FUDGE,
		   box.w + 2*REFRESH_FUDGE, 2*REFRESH_FUDGE);
	
	if (overlapBoxes(rbox1,rbox3) || overlapBoxes(rbox2,rbox4)) {
		iBox box1,box2;
		iBox box3;
		i_UnionBox(rbox1,rbox3, &box1);
		i_UnionBox(rbox2,rbox4, &box2);
		i_UnionBox(box1,box2, &box3);
		this->drawRegion(promoteBox(box3));
	} else {
		this->drawRegion(promoteBox(rbox1));
		this->drawRegion(promoteBox(rbox2));
		this->drawRegion(promoteBox(rbox3));
		this->drawRegion(promoteBox(rbox4));
	}
	refreshTextConstant(this, w, entity, this->getCanvas()->getFont(), box);
}
extern iBox i_GetPBBBounds(iEntityPtr entity);

/* 
// bogus to N degrees
// Modified by Scott - only bogus to N-1 degrees now
// does compute label size
// doesnt consider children -- but should it?
*/
iBox
ImagePage::entityBounds(iEntityPtr entity)
{
	// get bounding box of entity without label and scale to current magnification
	iBox box=i_GetEnclosingBox(entity);
#ifdef RAF_INTERNAL
	iBox pbb_box = i_GetPBBBounds(entity);
	i_UnionBox(box,pbb_box,&box);
#endif
	bbScaleUp(box, this->getMagExp());

	// get entity label
	const iUCode *real_text = ((IE *) eaGetFace(entity, vtIMAGE))->real_text;

	// compute width and height of label, including the box drawn around label
	// when it is being edited, and the gap between the text and the box
	int char_width=0, char_height=0, ascent=0;
	if(real_text)
		this->getCanvas()->measureText(real_text, ucslen(real_text), 
				  		&char_width, &char_height,
					   	&ascent);
	const int tw = char_width  + 2*RECT_GAP + MAX_BOX_THICKNESS;
	const int th = char_height + 2*RECT_GAP + MAX_BOX_THICKNESS;

	// if label is wider than entity, enlarge box but keep same x center
	if(tw > box.w) {
		box.x = box.x + box.w/2 - tw/2;
		box.w = tw;
	}
	// add height of label
	box.h += th;

	// refresh an area slightly larger than the actual entity
	box.x -= REFRESH_FUDGE;
	box.y -= REFRESH_FUDGE;
	box.w += 2*REFRESH_FUDGE;
	box.h += 2*REFRESH_FUDGE;
	
	// clip negative x and y
	if(box.x < 0){
		box.w += box.x;
		box.x = 0;
	}
	if(box.y < 0){
		box.h += box.y;
		box.y = 0;
	}
	
	return box;
}
static int TRUE_func(iEntityPtr entity)
{
	return TRUE;
}



void
ImagePage::refresh(IE *face)
{
	if(this->managedP() == 0) return;
	
	Widget w = this->getCanvas()->canvas;
	if(w == 0) return;
	
	iBox box = this->entityBounds(face->entity);
	if(bbNullP(box)) return;
	
	iBox rbox;
	setBox(rbox,
		   box.x, box.y,
		   box.w, box.h);
	this->drawRegion(promoteBox(rbox));
	
	refreshText(this, w, face, this->getCanvas()->getFont(), box);
}
iEntityPtr 
getTopMostEntity(iEntityPtr en)
{
	iEntityPtr parent = en;
	
	while (en = i_GetParentEntity(en))
		parent = en;
	return parent;	
}

static iBox
rotate_one_box(iBox in, iBox relative_box)
{
  iBox out;
  out.x = relative_box.h - in.h - in.y;
  out.y = in.x;
  out.w = in.h;
  out.h = in.w;
  return out;
}

static iDAFSError 
rotate_one_entity(iEntityPtr entity, iBox relative_box)
{
  void *data;
  iATag tag;
  iBox in, out;

  i_GetArea(entity, &data, &tag);
  switch(tag) {
  case iABox:
	in = i_GetEnclosingBox(entity);
	out = rotate_one_box(in, relative_box);
	i_SetBox(entity, out);
	break;
  case iAOrthogonList:	
	i_ForceAreaType(entity, iABoxList);
	i_GetArea(entity, &data, &tag);
  case iABoxList:	
	iBoxListPtr in_bl;
	iBoxListPtr out_bl = i_NewBoxList();
	in_bl = (iBoxListPtr) data;
	int i;
	for(i = 0; i < in_bl->numRects; i++) {
	  BOX rect;
	  rect = in_bl->rects[i];
	  in.x = rect.x1;
	  in.y = rect.y1;
	  in.w = rect.x2 - rect.x1;
	  in.h = rect.y2 - rect.y1;
	  out = rotate_one_box(in, relative_box);
	  i_UnionBoxWithBoxList(&out, out_bl, out_bl);
	}
	i_SetBoxList(entity, out_bl);
	break;
  }

  iBox tmp,*pbb_box;
  void *vp;

  // these are relative to the original box
  if (!i_GetUserDataProp(entity, "PBB", &vp, 0, i_SwapUBox)){
	pbb_box = (iBox*)vp;
	tmp.x = pbb_box->x;
	tmp.y = pbb_box->y;
	tmp.w = pbb_box->h;
	tmp.h = pbb_box->w;
	i_SetUserDataProp(entity, "PBB", (void*)&tmp, sizeof(iBox));
  }
  if (!i_GetUserDataProp(entity, "LPBB", &vp, 0, i_SwapUBox)){
	pbb_box = (iBox*)vp;
	tmp.x = pbb_box->x;
	tmp.y = pbb_box->y;
	tmp.w = pbb_box->h;
	tmp.h = pbb_box->w;
	i_SetUserDataProp(entity, "LPBB", (void*)&tmp, sizeof(iBox));
  }

  return DAFSOK;
}
	

void
ImagePage::rotateImage()
{
	int mag_exp = this->mode->main_window->mag_exp;
	iLBox ibox = {0,0,0,0};
	
#if 0
	/* this didn't work for a variety of reasons
	   mostly flaws in i_RotateEntity
	   but the concept that there is a hierarchy that all
	   have the same image is flawed
	   */
	elem = dllistContents(&this->face_list,dllistFirst(&this->face_list));
	entity = getTopMostEntity(elem->entity);
	i_RotateEntity(entity,90);
#else
	Ref r;
	DLList *l = &this->face_list;
	iBox image_bounds;
	image_bounds = i_GetImageBounds(image);

	for(r = dllistFirst(l); r; dllistNext(l, &r)){
	  IE *elem = (IE *) dllistContents(l, r);
	  iEntityPtr en;
	  en = elem->entity;
	  
	  rotate_one_entity(en, image_bounds);
	}
	i_RotateImage(this->image,90);
#endif

	this->width =  magUpInt(i_GetImageWidth(image), mag_exp);
	this->height = magUpInt(i_GetImageHeight(image), mag_exp);

	this->width+=FUDGE_WIDTH;
	this->height += FUDGE_HEIGHT;
	ibox.w = this->width;
	ibox.h = this->height;
	
	this->setPicture(ibox);
	this->syncSize();
	this->mode->main_window->document->setDirty();

	this->drawRegionNow(visibleRectangle(this->mode->scrolled_window));

}
void
ImagePage::centerOnSelected(iEntityPtr entity)
{
	iBox box = i_GetEnclosingBox(entity);
	int x, y;

	bbScaleUp(box, this->getMagExp());
	
	x = box.x + box.w/2;
	y = box.y + box.h/2;
	iLBox lbox = promoteBox(box);	
	// check the size of the scroll window, don't scroll if too small
	iLBox check, visible = visibleRectangle(this->mode->scrolled_window);
	i_IntersectLBox(visible, lbox, &check); // returns check.w = 0 if no intersection

	if((visible.w > lbox.w || visible.h > lbox.h) || check.w == 0)
		this->getCanvas()->scrollToPosition(this->mode->scrolled_window, x, y);
}
void
ImagePage::scrollToUpperLeft(iEntityPtr entity)
{
	iBox box = i_GetEnclosingBox(entity);

	bbScaleUp(box, this->getMagExp());
	
	if (!visiblePositionP(this->mode->scrolled_window, box.x, box.y)
		|| !visiblePositionP(this->mode->scrolled_window, box.x, box.y))
		this->getCanvas()->scrollToPosition(this->mode->scrolled_window, box.x, box.y);
}

void
ImagePage::scrollToTop(iEntityPtr entity)
{
	iBox box = i_GetEnclosingBox(entity);
	int x, y;

	bbScaleUp(box, this->getMagExp());
	
	x = box.x + box.w/2;
	y = box.y;

	if (!visiblePositionP(this->mode->scrolled_window,box.x,box.y) 
			|| !visiblePositionP(this->mode->scrolled_window,box.x,box.y+box.h))
		this->getCanvas()->scrollToPosition(this->mode->scrolled_window, x, y);
}
void
ImagePage::scrollToBottom(iEntityPtr entity)
{
	iBox box = i_GetEnclosingBox(entity);
	int x, y;

	bbScaleUp(box, this->getMagExp());
	
	x = box.x + box.w/2;
	y = box.y + box.h;

	if (!visiblePositionP(this->mode->scrolled_window,box.x,y) 
			|| !visiblePositionP(this->mode->scrolled_window,box.x+box.w,y))
		this->getCanvas()->scrollToPosition(this->mode->scrolled_window, x, y);
}
void
ImagePage::scrollToLeft(iEntityPtr entity)
{
	iBox box = i_GetEnclosingBox(entity);
	int x, y;

	bbScaleUp(box, this->getMagExp());
	
	x = box.x;
	y = box.y + box.h/2;

	if (!visiblePositionP(this->mode->scrolled_window,box.x,box.y) 
			|| !visiblePositionP(this->mode->scrolled_window,box.x,box.y+box.h))
		this->getCanvas()->scrollToPosition(this->mode->scrolled_window, x, y);
}
void
ImagePage::scrollToRight(iEntityPtr entity)
{
	iBox box = i_GetEnclosingBox(entity);
	int x, y;

	bbScaleUp(box, this->getMagExp());
	
	x = box.x + box.w;
	y = box.y + box.h/2;

	if (!visiblePositionP(this->mode->scrolled_window,x,box.y) 
			|| !visiblePositionP(this->mode->scrolled_window,x,box.y+box.h))
		this->getCanvas()->scrollToPosition(this->mode->scrolled_window, x, y);
}

void
ImagePage::deleteEntity(iEntityPtr entity)
{
	IE *ie;
	ie = (IE *) eaGetFace(entity, vtIMAGE);
	
	// because the boxes are no longer those gotten from i_GetBounds
	tree.deleteItem(entity);
	// bogus this was order dependent because I needed to get real text
	this->removeFace(ie);
	ie->deleteCheap();
}


static aString
getType(iEntityPtr parent)
{
	iTypeIndex parent_type;
	iTypeIndex type;
	
	parent_type = i_GetEntityType(parent);
	
	switch(parent_type) {
    case iNullT:
    case iZoneT:
    case iRegionT:
    case iGlyphT:
    default:
		i_NameToType("Blob",&type);
		break;
    case iDocumentT:
    case iParagraphT: 
    case iLineT:
    case iWordT:
		type = parent_type + 1;
		break;
	}
	
	return i_TypeToName(type);
}


#ifdef NEW_ENTITY
void
ImagePage::createChild(iEntityPtr parent)
{
	iBox bounds;
	int x, y, w, h;
	aString type;
	iEntityPtr new_entity;
	
	
	this->first_x = MAX(0, this->first_x);
	this->first_y = MAX(0, this->first_y);
	this->last_x  = MAX(0, this->last_x);
	this->last_y  = MAX(0, this->last_y);
	
	Position p_width = this->width;
	Position p_height = this->height;

	
	this->first_x = MIN(this->first_x, p_width);
	this->first_y = MIN(this->first_y, p_height);
	this->last_x  = MIN(this->last_x,  p_width);
	this->last_y  = MIN(this->last_y,  p_height);
	
	if(this->first_x < this->last_x) {
		x = this->first_x;
		w = this->last_x - x;
	} else {
		x = this->last_x;
		w = this->first_x - x;
	}
	
	if(this->first_y < this->last_y) {
		y = this->first_y;
		h = this->last_y - y;
	} else {
		y = this->last_y;
		h = this->first_y - y;
	}
	
	bounds.x = x;
	bounds.y = y;
	bounds.w = w;
	bounds.h = h;
	bbScaleDown(bounds, this->getMagExp());
	
#if 1
	if(bounds.w < LITTLE_PUNK_LIMIT || bounds.h < LITTLE_PUNK_LIMIT)
		return;
#endif
	
	type = getType(parent);
	// bogus -- does the parent get notified when I
	// add a child like this ???

	// bogus -- lets just do things right
	this->mode->main_window->document->purgeInactiveModes();
	iTypeIndex itype;

	i_NameToType(type.string(),&itype);
	new_entity = i_NewEntity(itype, parent);

	i_SetBox(new_entity, bounds);
	i_SetImage(new_entity, this->image);
	
#if 1
	// now how do we add this to displays?
	// bogus dont do this here but in the callback
	// but it was useless because the image wasnt set
	
	this->mode->main_window->document->addNewEntity(new_entity, TRUE);
#endif
	
}
#endif

#define CORNER_DELTA 6

static int
manhattanDistance(int first_x, int first_y, int second_x, int second_y)
{
	return DIFF(first_x, second_x) + DIFF(first_y, second_y);
}
static int onBar(iBox item_bounds, int x, int y)
{
	return (((item_bounds.x + CORNER_DELTA) < x) && ((item_bounds.x + item_bounds.w - CORNER_DELTA) > x)
		&& ((item_bounds.y - CORNER_DELTA) < y) && ((item_bounds.y + CORNER_DELTA) > y));
}
static ThingGrabbed
decideOperation(iBox item_bounds, int x, int y)
{
	if(manhattanDistance(item_bounds.x, item_bounds.y, x, y) < CORNER_DELTA)
		return tgUL;
	if(manhattanDistance(item_bounds.x + item_bounds.w, item_bounds.y + item_bounds.h, x, y) < CORNER_DELTA)
		return tgLR;
	if(manhattanDistance(item_bounds.x + item_bounds.w, item_bounds.y, x, y) < CORNER_DELTA)
		return tgUR;
	if(manhattanDistance(item_bounds.x, item_bounds.y + item_bounds.h, x, y) < CORNER_DELTA)
		return tgLL;
	if (onBar(item_bounds, x, y))
		return tgAll;
	return tgNOTHING;
}
iEntityPtr 
ImagePage::findTheEntity(DLList *l, int x, int y)
{
	Ref r;
	iBox bounds;
	iEntityPtr cur;
	
	for (r = dllistFirst(l); r;	 dllistNext(l,&r)) {
		cur = (iEntityPtr)dllistContents(l,r);
		bounds = i_GetEnclosingBox(cur);
		bbScaleUp(bounds,this->getMagExp());
		if (x >= bounds.x && x <= (bounds.x + bounds.w)
			&& y >= bounds.y && y <= (bounds.y + bounds.h))
			return cur;
	}
	return 0;
}
void
ImagePage::pressButton(XEvent *event)
{
	DLList l;
	Ref r;
	iEntityPtr entity;
	iEntityPtr sole_item;
	ViewCanvas *vc = this->getCanvas();
	
	this->end_all_edits();
	dllistCtor(&l);
	this->go = 0;
	event->xbutton.x -= LEFTPAD;
	event->xbutton.y -= TOPPAD;
	event->xbutton.x += vc->pix_hoffset;
	event->xbutton.y += vc->pix_voffset;
        switch(event->xbutton.button) {
	case Button1:
		this->findEnclosingEntities(event->xbutton.x,
									event->xbutton.y, &l);
		if(dllistEmpty(&l) == FALSE){
			sole_item = this->getSM()->getSoleItem();
			entity = (r = dllistFirst(&l)) ? 
				(iEntityPtr) dllistContents(&l, r) : 0;
			if (sole_item && (dllistSeek(&l,sole_item))) {
			
			
				iBox item_bounds;
				item_bounds = i_GetEnclosingBox(sole_item);
				bbScaleUp(item_bounds, this->getMagExp());
				
				this->thing = decideOperation(item_bounds, event->xbutton.x, event->xbutton.y);
				if (this->thing != tgNOTHING) {
					switch(this->thing){
					case tgAll:
						
						this->last_x = this->first_x = event->xbutton.x;
						this->last_y = this->first_y = event->xbutton.y;
						break;
					case tgUL:
						// pin the LR
						
						this->first_x = item_bounds.x + item_bounds.w;
						this->first_y = item_bounds.y + item_bounds.h;
						this->last_x = event->xbutton.x;
						this->last_y = event->xbutton.y;
						break;
					case tgLR:
						// pin the UL
						
						this->first_x = item_bounds.x;
						this->first_y = item_bounds.y;
						this->last_x = event->xbutton.x;
						this->last_y = event->xbutton.y;
						break;
					case tgLL:
						// pin the UR
						
						this->first_x = item_bounds.x + item_bounds.w;
						this->first_y = item_bounds.y;
						this->last_x = event->xbutton.x;
						this->last_y = event->xbutton.y;
						break;
					case tgUR:
						// pin the LL
						
						this->first_x = item_bounds.x;
						this->first_y = item_bounds.y + item_bounds.h;
						this->last_x = event->xbutton.x;
						this->last_y = event->xbutton.y;
						break;
					default:
						break;
					}
					this->setupMoveXor();
					//this->drawXOR(XtWindow(vc->canvas));
					dllistDtor(&l);
					return;
				}
			
			}
			if(event->xbutton.state & ShiftMask){
				this->toggleItem(entity);
			} else {
				
				if(dllistEqual(&l, &this->select_list) && sole_item && dllistSeek(&l, sole_item)){
					// mr user has click twice in near enough the same spot
					for(r = dllistFirst(&l); r && (entity = (iEntityPtr) dllistContents(&l, r)) != sole_item; dllistNext(&l, &r))
						;
					entity = 0;
					if(r){
						dllistNext(&l, &r);
						if(r)
							entity = (iEntityPtr) dllistContents(&l, r);
					}
				} else {
					int x = event->xbutton.x;
					int y = event->xbutton.y;
					entity = findTheEntity(&l,x, y);
				}
				dllistAssign(&this->select_list, &l);
				if(entity)
					this->setItem(entity);
				else
					this->clear();
			}
		} else {
			this->clear();
		}
		break;
	case Button2:
#ifdef NEW_ENTITY
		if(this->getSM()->getSoleItem()) { 
			this->last_x = this->first_x = event->xbutton.x;
			this->last_y = this->first_y = event->xbutton.y;
			this->setupNewXor();
		} 
#endif
		break;
	case Button3:
		break;
    }
}

void
ImagePage::releaseButton(XEvent *event)
{
	iEntityPtr entity;
	ViewCanvas *vc = this->getCanvas();
	
	// we got a release without a press
	if(!this->go)
		return;
	
	event->xbutton.x -= LEFTPAD;
	event->xbutton.y -= TOPPAD;
	event->xbutton.x += vc->pix_hoffset;
	event->xbutton.y += vc->pix_voffset;
    switch(event->xbutton.button) {
	case Button3:
		break;
	case Button2:
#ifdef NEW_ENTITY
		if(entity = this->getSM()->getSoleItem()){ 
			this->eraseXOR();
			this->createChild(entity);
			
		}
#endif
		break;
	case Button1:
		if((entity = this->getSM()->getSoleItem())
				 && this->thing != tgNOTHING) {
			this->eraseXOR();
			
			iBox bounds;
			iBox *pbb;
			iBox old_bounds;
			old_bounds = i_GetEnclosingBox(entity);
			bounds = this->xor_box;
			if (bounds.w <= 2 || bounds.h <= 2) 
				goto end;
			bbScaleDown(bounds, this->getMagExp());

			// duh! move the pbb first so that
			// it is changed by the time setbounds
			// calls the update
			if (!i_GetUserDataProp(entity, "PBB",(void**) &pbb , 
							   0, i_SwapUBox)){
				pbb->x += old_bounds.x - bounds.x;
				pbb->y += old_bounds.y - bounds.y;
			}
			i_SetBox(entity, bounds);

			this->mode->main_window->document->setDirty();
		}
		break;
    }
end:
	this->first_x =  this->last_x = this->first_y = this->last_y = -1;
	this->xor_box.w = 0;
	this->go = 0;
	this->thing = tgNOTHING;
	return;
}

void
ImagePage::moveButton(int state, int x, int y)
{
	iImagePtr sole_image = 0;
	iEntityPtr sole_entity;
	ViewCanvas *vc = this->getCanvas();
		
	x -= LEFTPAD;
	y -= TOPPAD;
	sole_entity = this->getSM()->getSoleItem();
	if(sole_entity)
		sole_image = i_GetImage(sole_entity);
	
	if(sole_image != this->image)
		return;
	
	if(state & Button2Mask) {
#ifdef NEW_ENTITY
		if (this->go)
			this->eraseXOR();
		
		
		this->last_x = x;
		this->last_y = y;
	    
		this->setupNewXor();
		this->drawXOR(XtWindow(vc->canvas),0,0);
		this->go = 1;
#endif
	} else if (state & Button1Mask) {
		if (this->thing == tgNOTHING)
			return;
		if (this->go)
			this->eraseXOR();
		//printf("continued a move\n");	 
		this->last_x = x;
		this->last_y = y;
		
		this->setupMoveXor();
		this->drawXOR(XtWindow(vc->canvas),0,0);
		this->go = 1;
	}
}


void
ImagePage::prevEntity()
{
	this->prevEntity(TRUE_func);
}

void
ImagePage::nextEntity()
{
	this->nextEntity(TRUE_func);
}

void
ImagePage::findNextFlagEntity()
{
	iEntityPtr sole_item = this->getSM()->getSoleItem();

	if(sole_item)
		entityClearFlag(this->mode->main_window,sole_item);
	this->nextEntity(entityFlaggedP);
}

void
ImagePage::nextEntity(EntityCheckFunc checker)
{
	iEntityPtr sole_entity;
	iEntityPtr right_entity=0;
	int good=0;
	Ref r;
	IE *ie = 0;
	
	const DLList *selection=this->getSM()->getItems();
	this->end_all_edits();
	r = dllistLast(selection);
	if (r)
		sole_entity = (iEntityPtr) dllistContents(selection, r);
	else
		sole_entity=0;

	if(sole_entity)
		ie = (IE *) eaGetFace(sole_entity, vtIMAGE);

	if(ie){
		DLList *l = &this->face_list;
		Ref r = dllistSeek(l, ie);
		while(r && !good) {
			dllistNext(l, &r);
			if(r){
				ie = (IE *) dllistContents(l, r);
				if (ie) {
					right_entity = ie->entity;
					good = checker(right_entity);
				}
			}
		}
		if (right_entity && good) this->getSM()->setItem(right_entity);
	} else
		this->firstEntity(checker);
}

void
ImagePage::prevEntity(EntityCheckFunc checker)
{
	iEntityPtr sole_entity;
	iEntityPtr right_entity=0;
	int good=0;
	Ref r;
	IE *ie = 0;
	
	// I guess this no longer operates on sole selected entities
	// but rather on the first guy of a group of selected guys
	/*   sole_entity = this->getSM()->getSoleItem();*/
	
	const DLList *selection=this->getSM()->getItems();

	this->end_all_edits();
	r = dllistFirst(selection);
	if (r)
		sole_entity = (iEntityPtr) dllistContents(selection, r);
	else
		sole_entity=0;

	if(sole_entity)
		ie = (IE *) eaGetFace(sole_entity, vtIMAGE);
	
	if(ie){
		DLList *l = &this->face_list;
		Ref r = dllistSeek(l, ie);
		while(r && !good) {
			dllistPrev(l, &r);
			if(r){
				ie = (IE *) dllistContents(l, r);
				if(ie){
					right_entity = ie->entity;
					good=checker(right_entity);
				}
			}
		}
		if (right_entity && good) this->getSM()->setItem(right_entity);
	} else {
		this->lastEntity(checker);
	}
}

/*
// hmm checking for selection when clearflagcb KNOWS there
// isnt one, is selection ever important
// is this called elsewhere?
*/
void
ImagePage::firstEntity(EntityCheckFunc checker)
{
	iEntityPtr right_entity=0;
	int good=0;
	IE *ie;
	
	DLList *l = &this->face_list;
	Ref r = dllistFirst(l);
	this->end_all_edits();
	while(r && !good) {
		ie = (IE *) dllistContents(l, r);
		if(ie) {
			right_entity = ie->entity;
			good=checker(right_entity);
		}
		dllistNext(l, &r);
	}
	if (right_entity && good) 
		this->getSM()->setItem(right_entity);
}

void
ImagePage::lastEntity(EntityCheckFunc checker)
{
	iEntityPtr right_entity=0;
	int good=0;
	IE *ie;
	
	DLList *l = &this->face_list;
	Ref r = dllistLast(l);
	this->end_all_edits();
	while(r && !good) {
		ie = (IE *) dllistContents(l, r);
		if(ie) {
			right_entity = ie->entity;
			good=checker(right_entity);
		}
		dllistPrev(l, &r);
	}
	if (right_entity && good) 
		this->getSM()->setItem(right_entity);
}



/*
// I see two choices for moving
// 1. make it not really do xor to erase and use drawRegionnow
// 2. or maybe redisplay should use drawRegionnow
//
// I choose 2 because it will be generally faster and 
// handling special cases for drawRegionNow 
//   like making it bigger
//   avoiding 0 size were a pain
*/

void
ImagePage::drawXOR(Pixmap drawable,int x_off, int y_off)
{
	ViewCanvas *vc = this->getCanvas();
	Widget widget = this->getCanvas()->canvas;
	Display *display = XtDisplay(widget);
	GC gc;
	iLBox drawbox;
	iBox xorbox = this->xor_box;
	
	
	gc = this->getCanvas()->xorGC();
	iLBox vir = visibleRectangle(this->mode->scrolled_window);
	if (vc->pix_hoffset) vir.x -= LEFTPAD;
	if (vc->pix_voffset) vir.y -= TOPPAD;
	
	i_IntersectLBox(promoteBox(xorbox),vir,&drawbox);
	XDrawRectangle(display, drawable, gc,
				   drawbox.x + LEFTPAD - vc->pix_hoffset - x_off, 
				   drawbox.y + TOPPAD-vc->pix_voffset - y_off,
				   drawbox.w, drawbox.h);
}

void
ImagePage::eraseXOR()
{
	Widget w = this->getCanvas()->canvas;
	this->drawXOR(XtWindow(w),0,0);
}


void
ImagePage::setupNewXor()
{
	int mag_exp = this->mode->main_window->mag_exp;
	int image_height;
	int image_width;
	
	
	image_height = magUpInt(i_GetImageHeight(image), mag_exp);
	image_width  = magUpInt(i_GetImageWidth(image), mag_exp);
	
	this->first_x = MAX(0, this->first_x);
	this->first_y = MAX(0, this->first_y);
	this->last_x  = MAX(0, this->last_x);
	this->last_y  = MAX(0, this->last_y);
	
	this->first_x = MIN((int) this->first_x, (int) image_width);
	this->first_y = MIN((int) this->first_y, (int) image_height);
	this->last_x  = MIN((int) this->last_x,  (int) image_width);
	this->last_y  = MIN((int) this->last_y,  (int) image_height);
	if (mag_exp>0){
		// bogus what is this doing
		this->first_x = (this->first_x>>mag_exp)<<mag_exp;
		this->first_y = (this->first_y>>mag_exp)<<mag_exp;
		this->last_x = (this->last_x>>mag_exp)<<mag_exp;
		this->last_y = (this->last_y>>mag_exp)<<mag_exp;
	}
	if(this->first_x < this->last_x) {
		this->xor_box.x = this->first_x;
		this->xor_box.w = this->last_x - this->xor_box.x;
	} else {
		this->xor_box.x = this->last_x;
		this->xor_box.w = this->first_x - this->xor_box.x;
	}
	
	if(this->first_y < this->last_y) {
		this->xor_box.y = this->first_y;
		this->xor_box.h = this->last_y - this->xor_box.y;
	} else {
		this->xor_box.y = this->last_y;
		this->xor_box.h = this->first_y - this->xor_box.y;
	}
	bbScaleDown(this->xor_box, this->getMagExp());
	bbScaleUp(this->xor_box,this->getMagExp());
}

void
ImagePage::setupMoveXor()
{
	iEntityPtr sole_entity;
	int image_height;
	int image_width;
	int mag_exp = this->mode->main_window->mag_exp;
	
	image_height = magUpInt(i_GetImageHeight(image), mag_exp);
	image_width  = magUpInt(i_GetImageWidth(image), mag_exp);
	
	sole_entity = this->getSM()->getSoleItem();
	iBox bounds;
	iBox ebox = i_GetEnclosingBox(sole_entity);
	
	switch(this->thing){
    case tgAll:
		// moving the whole thing
		bounds = i_GetEnclosingBox(sole_entity);
		
		// bogus what is this 4 stuff
		this->xor_box.x = 
			this->last_x - this->first_x + ((bounds.x << (mag_exp + 4)) >> 4);
		this->xor_box.y = this->last_y - this->first_y + ((bounds.y << (mag_exp + 4)) >> 4);
		this->xor_box.w = (bounds.w << (mag_exp + 4)) >> 4;
		this->xor_box.h = (bounds.h << (mag_exp + 4)) >> 4;
		
		this->xor_box.x = MAX(0, this->xor_box.x);
		this->xor_box.y = MAX(0, this->xor_box.y);
		
		this->xor_box.x = MIN(image_width - this->xor_box.w, this->xor_box.x);
		this->xor_box.y = MIN(image_height - this->xor_box.h, this->xor_box.y);
		
		break;
    default:
		// moving a corner
		this->xor_box.x = MIN(this->first_x, this->last_x);
		this->xor_box.y = MIN(this->first_y, this->last_y);
		
		this->xor_box.x = MAX(0, this->xor_box.x);
		this->xor_box.y = MAX(0, this->xor_box.y);
		this->xor_box.x = MIN(this->xor_box.x, image_width  - 1);
		this->xor_box.y = MIN(this->xor_box.y, image_height - 1);
		
		this->xor_box.w = DIFF(this->first_x, this->last_x);
		this->xor_box.h = DIFF(this->first_y, this->last_y);
		
		this->xor_box.w = MAX(1, this->xor_box.w);
		this->xor_box.h = MAX(1, this->xor_box.h);
		
		this->xor_box.w = MIN(this->xor_box.w, image_width  - this->xor_box.x);
		this->xor_box.h = MIN(this->xor_box.h, image_height - this->xor_box.y);
		
		break;
	}
	bbScaleDown(this->xor_box, this->getMagExp());


	// if we don't do the following code,
	// rounding can cause the stationary end to move
	switch(this->thing){
	  // restore original y
	case tgLR:
	case tgLL:
	  if(this->first_y < this->last_y) 
		xor_box.y = ebox.y;
	  else
		xor_box.h = ebox.y - xor_box.y;
	  break;
	case tgUR:
	case tgUL:
	  // restore original y+h
	  // assert ebox.y + ebox.h == xor_box.y + xor_box.h
	  if(this->first_y > this->last_y) 
		xor_box.h = ebox.y + ebox.h - xor_box.y;
	  else
		xor_box.y = ebox.y + ebox.h;
	  break;
	case tgAll:
	default:
	  break;
	}

	switch(this->thing){
	  // restore original x
	case tgUR:
	case tgLR:
	  if(this->first_x < this->last_x) 
		xor_box.x = ebox.x;
	  else
		xor_box.w = ebox.x - xor_box.x;
	  break;
	case tgUL:
	case tgLL:
	  // restore original x+w
	  // assert ebox.x + ebox.w == xor_box.x + xor_box.w
	  if(this->first_x > this->last_x) 
		xor_box.w = ebox.x + ebox.w - xor_box.x;
	  else
		xor_box.x = ebox.x + ebox.w;
	  break;
	case tgAll:
	default:
	  break;
	}

	bbScaleUp(this->xor_box,this->getMagExp());
}

/*
// Some methods of Page that we override
*/

void
ImagePage::manage()
{
	Widget sw = this->mode->scrolled_window;
	
	this->Page::manage();
	this->central_position = demoteBox(visibleRectangle(sw));
}

void
ImagePage::unmanage()
{
	this->Page::unmanage();
	this->central_position.w = 0;
}

void
ImagePage::initialise()
{
	int length;
	int log;
	iBox box;


	box.x = 0;
	box.y = 0;

	// this is for the quad tree, so we must
	// get the dimension without screwing up the mag
	box.w = i_GetImageWidth(image);
	box.h = i_GetImageHeight(image);

	box.h = MAX(box.h, box.w);
	box.w = box.h;
		
	length = dllistLength(&this->face_list);
	log = 0;
	while(length >>= 1)
		log++;
	log /= 2;
	// test size
	
	this->tree.createTree(box, log);
	DLList list;
	Ref r;
	IE *ie;
		
	// bogus change QuadTree to take the IE's
	dllistCtor(&list);
	for(r = dllistFirst(&this->face_list);r; dllistNext(&this->face_list, &r)){
		ie = (IE *) dllistContents(&this->face_list, r);
		dllistAppend(&list, ie->entity);
	}
	this->tree.addItems(&list);
	dllistDtor(&list);
}
