
/*
 * Copyright (c) 1994, 1995 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */


#include <X11/StringDefs.h>
#include <X11/Intrinsic.h> 
#include <X11/Xutil.h>
#include <Xm/Xm.h>
#include <Xm/DrawingA.h>
#include <Xm/PushB.h>

#include <stdio.h>

#include "illum.h"

#include "rafutil.h"

#include "misc/stringVoidPtrMap.h"

#include "scroll.h"
#include "bb.h"

#include "EntityAux.h"
#include "MainWindow.h"
#include "ChooseDialog.h"
#include "Document.h"
#include "SelectionManager.h"

#include "ImageView.h"
#include "ImageFrame.h"
#include "IE.h"

static char *
makeKey(iImagePtr image, char buffer[24])
{
	sprintf(buffer, "%d", image);
	return buffer;
}

void
ImageView::createFrames(DLList *image_list, DLList *entity_list)
{
	stringVoidPtrMap map;
	int frame_count = 0;
	Ref r;
	iEntityPtr entity;
	iImagePtr image;
	char buffer[24];
	char buffer2[24];
	char *key;
	ImageFrame *frame;
	
	svmapInit(&map, 0, 10001);
	
	for(r = dllistFirst(image_list); r; dllistNext(image_list, &r)){
		image = (iImagePtr) dllistContents(image_list, r);
		key = makeKey(image, buffer);
		
		sprintf(buffer2, "%d", ++frame_count);
		frame = new ImageFrame(image, this, buffer2);
		svmapSet(&map, key, frame);
		this->addFrame(frame);
	}
	
	for(r = dllistFirst(entity_list); r; dllistNext(entity_list, &r)) {
		entity = (iEntityPtr) dllistContents(entity_list, r);
		
		image = i_GetImage(entity);
		if(image) {
			key = makeKey(image, buffer);
			frame = (ImageFrame *) svmapGet(&map, key);
			frame->addMember(entity);
		}
	}
	
	for(r = dllistFirst(image_list); r; dllistNext(image_list, &r)){
		image = (iImagePtr) dllistContents(image_list, r);
		key = makeKey(image, buffer);
		frame = (ImageFrame *) svmapGet(&map, key);
		
		frame->initialise();
	}
	
	svmapCleanup(&map);
}

void ImageView::imageDeleted(iImagePtr image)
{
	Ref r;
	DLList *fl = &this->frame_list;
	ImageFrame *frame;
	
	for(r = dllistFirst(fl); r; dllistNext(fl, &r)) {
		frame = (ImageFrame *) dllistContents(fl, r);
		if (frame->image==image) {
			frame->image=0;
			frame->should_delete=TRUE;
		}
	}
}


void
ImageView::findImagesSub(iEntityPtr entity, stringVoidPtrMap *map, DLList *image_list)
{
	iImagePtr image;
	char buffer[24];
	char *key;
	iEntityPtr child;
	
	image = i_GetImage(entity);
	
	if(image) {
		key = makeKey(image, buffer);
		if(svmapContainsP(map, key) == FALSE){
			svmapSet(map, key, image);
			dllistAppend(image_list, image);
		}
	}
	
	for(child = i_GetChildEntity(entity); child; child = i_GetNextEntity(child)) {
		this->findImagesSub(child, map, image_list);
	}
}

void
ImageView::findImages(iEntityPtr input_root_entity, DLList *image_list)
{
	stringVoidPtrMap map;
	svmapInit(&map, 0, 10001);
	
	this->findImagesSub(input_root_entity, &map, image_list);
	
	svmapCleanup(&map);
}

void
ImageView::createFrames(iEntityPtr input_root_entity)
{
	DLList entity_list;
	DLList image_list;
	
	dllistCtor(&entity_list);
	dllistCtor(&image_list);
	
	this->findImages(input_root_entity, &image_list);
	this->main_window->chooser->choose(input_root_entity, &entity_list);
	this->createFrames(&image_list, &entity_list);
	
	dllistDtor(&image_list);
	dllistDtor(&entity_list);
}

ImageView::ImageView(MainWindow *input_main_window,
					 iEntityPtr input_root_entity)
	: View(input_main_window)
{
	this->createFrames(input_root_entity);
}

ImageView::~ImageView()
{
}

void
ImageView::centerOnSelected(iEntityPtr entity)
{
	IE *ie;
	ie = (IE *) eaGetFace(entity, vtIMAGE);
	DLList *l = &this->frame_list;
	Ref r;
	if (!ie) return;
	r = dllistSeek(l, ie->frame);
	if(r){
		setCurrentFrame(r);
		ie->frame->centerOnSelected(entity);
	}      
}

void
ImageView::reconsider(iEntityPtr entity)
{
	DLList *l = &this->frame_list;
	Ref r;
	ImageFrame *frame;
	iImagePtr image = i_GetImage(entity);
	
	for(r = dllistFirst(l); r; dllistNext(l, &r)) {
		frame = (ImageFrame *) dllistContents(l, r);
		if(image == frame->image)
			break;
	}
	
	if(r == 0)
		halt("couldnt find the frame.\n");
	
	frame->checkEntity(entity);
}

void
ImageView::addNewEntity(iEntityPtr new_entity)
{
	// bogus
	DLList *l = &this->frame_list;
	Ref r;
	ImageFrame *frame;
	iImagePtr image = i_GetImage(new_entity);
	
	for(r = dllistFirst(l); r; dllistNext(l, &r)) {
		frame = (ImageFrame *) dllistContents(l, r);
		if(image == frame->image)
			break;
	}
	
	if(r == 0)
		halt("couldnt find the frame.\n");
	
	frame->addNewMember(new_entity);
	
	if(frame->tree.initialized())
		frame->tree.addItem(new_entity);
}

void
ImageView::choose(int buttons_changed)
{
	Ref r;
	DLList *l = &this->frame_list;
	ImageFrame *frame;
	ImageFrame *current_frame;
	iImagePtr image;
	Position x, y;
	int managed;

	// find the frame in view and remember its image
	// then carefully junk all the current frames
	
	current_frame = (ImageFrame *) this->getCurrentFrame();
	managed = current_frame && current_frame->managedP();
	
	if(((current_frame && current_frame->invalid_guy_added) || buttons_changed) == FALSE)
		return;

	// dont clear -- we used to do this because we really hadn't fix
	// the delete of the frame
#if 0
	if(current_frame)
		current_frame->clear();
#endif
	
	if(managed) {
		image = current_frame->image;
		getScrollPositionCenter(current_frame->view->scrolled_window, &x, &y);
		current_frame->unmanage();
	} else {
		image = 0;
	}
	
	this->current_frame_ref = 0;
	
	for(r = dllistFirst(l); r; dllistNext(l, &r)){
		frame = (ImageFrame *) dllistContents(l, r);
#if 0	
		// this is a cheese -- it deletes the frame -- and delete frame
		// doesn't kill the ea->faces 
		// we should check out what changeMode does
		delete frame;
#else
		frame->purge();
#endif
	}
	dllistClear(l);
	iEntityPtr root_entity = this->main_window->document->root_entity;
	this->createFrames(root_entity);
	r = 0;
	// if there used to be an image in view, let's find it
	// among the new frames
	
	if(managed) {
		for(r = dllistFirst(l); r; dllistNext(l, &r)){
			frame = (ImageFrame *) dllistContents(l, r);
			if(image == frame->image)
				break;
		}
	}
	
	if(managed) {
		// we found the new frame that corresponds to the image in the
		// one that used to be in view
		if(r){
			this->setCurrentFrame(r);
			scrollToPosition(frame->view->scrolled_window, x, y);
		} else {
			this->start();
		}
	}

	// go through the list of entities selected
	// and unselect the ones that no longer have IE's
	const DLList *selected;
	DLList copy;
	dllistCtor(&copy);
	selected = this->getSM()->getItems();
	// have to have a copy because we will modify the
	// list in the loop
	dllistAssign(&copy, selected);

	for(r = dllistFirst(&copy); r; dllistNext(&copy, &r)){
		iEntityPtr entity;
		IE *ie;
		entity = (iEntityPtr) dllistContents(&copy, r);
		ie = (IE *) eaGetFace(entity, vtIMAGE);
		if(ie == 0)
			this->getSM()->toggleItem(entity);
	}
	
	dllistDtor(&copy);
	// bogus do I have to do this
	this->main_window->refresh_dialogs();
}

iBox
ImageView::getBounds(iEntityPtr entity)
{
	iBox bounds;
	
	// -- check that the frame matches
	ImageFrame *current_frame;
	ImageFrame *entitys_frame;
	IE *ie;
	
	bounds.w = 0;
	ie = (IE *) eaGetFace(entity, vtIMAGE);
	if(ie == 0)
		return bounds;
	entitys_frame = (ImageFrame *) ie->frame;
	current_frame = (ImageFrame *) this->getCurrentFrame();
	
	if(entitys_frame == current_frame) {
		bounds = i_GetBounds(entity);
		int mag_exp = current_frame->getMagExp();
		
		// account for mag
		bbScaleUp(bounds, mag_exp);
	} else {
		bounds.w = 0;
	}
	
	
	return bounds;
}
