
/*
 * Copyright (c) 1994, 1995 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */


#include "rafport.h"
#include "illum.h"

#include "ucs.h"
#include "ListDialog.h"
#include "misc/XArg.h"
#include "rafutil.h"
#include <Xm/DialogS.h>
#include <X11/Xutil.h>

#include <Xm/Text.h>
#include <Xm/TextF.h>
#include <Xm/FileSB.h>
#include <Xm/SelectioB.h>
#include <Xm/SelectioBP.h>

#include "Application.h"

/*
// There are two ways to use this dialog
// modally, using showmodal as in the type selection in entity dialog
// or
// non-modally, as in the page selection in  illum
// the former returns a selected string
// the latter calls a user supplied callback 
*/



static void OkCB(Widget w, ListDialog *sl, void *call)
{
	sl->okCB((XmFileSelectionBoxCallbackStruct *)call);
}

static void CancelCB(Widget w, ListDialog *sl, void *call)
{
	sl->cancelCB((XmFileSelectionBoxCallbackStruct *)call);
}

#if 0
static void ApplyCB(Widget w, ListDialog *sl, void *call)
{
	sl->applyCB((XmFileSelectionBoxCallbackStruct *)call);
}
#endif

static void beepAtHim(Widget w, ListDialog *sl, void *call)
{
	ringTheBell();
}

int 
modal_callback(ListDialog *sl, char *name)
{
	sl->done = TRUE;
	sl->answer = name;

	return TRUE;
}

ListDialog::ListDialog(MainWindow *input_mw, Widget input_app_shell, 
					   XtAppContext(input_app_context),
					   XmString input_title, selCBProc CB, void *CBdata)
	: ModalDialog(input_mw, input_app_shell, input_app_context),
	  user_data(CBdata),
	  call_back(CB),
	  answer(""),
	  title(XmStringCopy(input_title))
{		
	if(call_back == 0 && input_app_context != 0) {
		call_back = modal_callback;
		user_data = this;
	}
}

ListDialog::~ListDialog()
{
	XmStringFree(this->title);
}

void
ListDialog::create()
{
	XArg args;
	args.reset();
	args.setArg(XmNdialogTitle,title);
	if(this->app_context)
		args.setArg(XmNdialogStyle, XmDIALOG_FULL_APPLICATION_MODAL);
	else {
		args.setArg(XmNmustMatch, TRUE);
	}
	
	
	this->dialog = XmCreateSelectionDialog(this->app_shell, 
										   "select list",
										   args.getArgs(), 
										   args.getSize());
	XtUnmanageChild(XmSelectionBoxGetChild(this->dialog, XmDIALOG_HELP_BUTTON));
	XtAddCallback(this->dialog, XmNokCallback, OkCB, this);
	XtAddCallback(this->dialog, XmNcancelCallback, CancelCB, this);
	if(this->app_context){
		XtUnmanageChild(XmSelectionBoxGetChild(this->dialog, XmDIALOG_APPLY_BUTTON));
	} else {
		XtUnmanageChild(XmSelectionBoxGetChild(this->dialog, XmDIALOG_APPLY_BUTTON));
		XtAddCallback(this->dialog, XmNnoMatchCallback, beepAtHim, this);
		//XtAddCallback(this->dialog, XmNapplyCallback, ApplyCB, this);
	}
}

/*
// non modal version of show
// don't copy thing John
*/
void
ListDialog::show(XmStringTable table, int size)
{
	if(this->dialog == 0)
		this->create();

	// what do I have to free? Anything?
	XArg args;
	
	args.reset();
	args.setArg (XmNlistItemCount, size); 
	args.setArg (XmNlistItems, table); 
	args.setValues(this->dialog);

	Widget text_input;
	text_input = XmSelectionBoxGetChild(this->dialog, XmDIALOG_TEXT);

	XmTextFieldSetString(text_input, "");
	
	this->Dialog::show();
}


/*
// dont come back until you have an answer
*/
aString
ListDialog::showModal(XmStringTable table, int size)
{
	if(this->dialog == 0)
		this->create();

	// what do I have to free? Anything?
	XArg args;
	
	args.reset();
	args.setArg (XmNlistItemCount, size); 
	args.setArg (XmNlistItems, table); 
	args.setValues(this->dialog);

	Widget text_input;
	text_input = XmSelectionBoxGetChild(this->dialog, XmDIALOG_TEXT);

	XmTextFieldSetString(text_input, "");
	
	XtManageChild(this->dialog);

	// contains the event loop
	this->ModalDialog::showModal();

	return this->answer;
}

void
ListDialog::change(XmStringTable table, int size)
{
	XArg args;
	
	// hmm... I wonder if I have to worry about it being managed?
	args.reset();
	args.setArg (XmNlistItemCount, size); 
	args.setArg (XmNlistItems, table); 
	args.setValues(this->dialog);
}

char *
ListDialog::getString(XmFileSelectionBoxCallbackStruct *cbs)
{
	char *string;
	// this is bogus old code
	if (XtIsSubclass(this->dialog, xmTextFieldWidgetClass)) {
		string = XmTextFieldGetString(this->dialog);
	} else {
		/* item was selected from list ? */
		XmStringGetLtoR(cbs->value, XmSTRING_DEFAULT_CHARSET, &string);
	}
	return string;
}

void
ListDialog::okCB(XmFileSelectionBoxCallbackStruct *cbs)
{
	char *s;
	if (this->call_back) {
		s=this->getString(cbs);
		if (s && this->call_back(user_data,s))
			this->hide();
		XtFree(s);
	} else
		this->hide();
}

void
ListDialog::applyCB(XmFileSelectionBoxCallbackStruct *cbs)
{
	if (call_back) {
		this->call_back(user_data,this->getString(cbs));
	}
}

void
ListDialog::cancelCB(XmFileSelectionBoxCallbackStruct *cbs)
{
	if (call_back) {
		this->call_back(user_data,"");
	}

	this->hide();
}

