
/*
 * Copyright (c) 1994, 1995 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"

#include <X11/StringDefs.h>
#include <X11/Intrinsic.h> 
#include <X11/Xutil.h>
#include <Xm/Xm.h>
#include <Xm/DrawingA.h>
#include <Xm/PushB.h>

#include <stdio.h>

#include "illum.h"

#include "Mode.h"
#include "Page.h"
#include "MainWindow.h"
#include "ViewCanvas.h"
#include "smemory.h"


Mode::Mode(MainWindow *input_main_window)
	: current_page_ref(0),
	  scrolled_window(0), main_window(input_main_window)
{
	dllistCtor(&this->page_list);
}


Mode::~Mode()
{
	Page *page;
	DLList *fl = &this->page_list;
	Ref r;
	
	for(r = dllistFirst(fl); r; dllistNext(fl, &r)){
		page = (Page *) dllistContents(fl, r);
		if(page->managed) {
			if(this->main_window->view_canvas)
				this->main_window->view_canvas->unmanage();
		}
		delete page;
	}
	dllistDtor(fl);
}

void
Mode::purge()
{
	Page *page;
	DLList *fl = &this->page_list;
	Ref r;
	
	for(r = dllistFirst(fl); r; dllistNext(fl, &r)){
		page = (Page *) dllistContents(fl, r);
		if(page->managed) {
			if(this->main_window->view_canvas)
				this->main_window->view_canvas->unmanage();
		}
		page->purge();
	}
	dllistClear(fl);

	delete this;
}

iBox 
Mode::getBounds(iEntityPtr entity)
{ 
	iBox box;
	box.w = 0;
	return box;
}


Page *
Mode::getCurrentPage ()
{
	Ref r;
	Page *page = 0;
	r = this->current_page_ref;
	if (r)
		page = (Page *) dllistContents(&this->page_list, r);
	return page;
}


void
Mode::show()
{
	Page *page;
	DLList *fl = &this->page_list;
	
	if(current_page_ref != 0){
		page = (Page *) dllistContents(fl, this->current_page_ref);
		page->manage();
	}
}

void 
Mode::makePageList(XmStringTable *list, int *list_size)
{
	Ref r;
	int i;
	Page *page;
	
	*list_size=0;
	DLList *l = &this->page_list;
	
	for(r = dllistFirst(l), i = 0; r; dllistNext(l, &r))
		i++;
	*list = (XmStringTable)smalloc(i * sizeof(XmString *));
	for(i = 0, r = dllistFirst(l); r; dllistNext(l, &r)) {
		page = (Page *) dllistContents(l, r);
		if (!page->should_delete)
			(*list)[i++] = XmStringCreateLocalized((char *) page->name.string());
	}
	*list_size=i;
}

void
Mode::hide()
{
	Page *page;
	DLList *fl = &this->page_list;
	
	if(current_page_ref != 0){
		page = (Page *) dllistContents(fl, this->current_page_ref);
		page->unmanage();
	}
}

void
Mode::emptied(Page *page)
{
	Ref r,new_r=0;
	
	DLList *fl = &this->page_list;
	r=dllistSeek(fl,page);
	
	if (r) {
		Ref r1=r;
		Ref r2=r;
		dllistNext(fl,&r1);
		new_r=r1;
		if (!r1) {
			dllistPrev(fl,&r2);
			new_r=r2;
		}
	}
	if (page->managedP()) {
		this->setCurrentPage(new_r);
	} else {
		if(r == current_page_ref)
			current_page_ref = 0;
	}
	if (r) 
		dllistRemove(fl,r);
	if (page) {
		//delete page;
	}
	
	
}

void
Mode::setCurrentPage(Ref r)
{
	Page *page;
	DLList *fl = &this->page_list;
	
	if(current_page_ref != 0){
		page = (Page *) dllistContents(fl, this->current_page_ref);
		page->unmanage();
	}
	this->current_page_ref = r;
	if(this->current_page_ref != 0) {
		page = (Page *) dllistContents(fl, this->current_page_ref);
		page->conditionallyInitialiseCanvas();
		page->manage();
		this->main_window->standardTitle();
	}
}

void
Mode::start()
{
	Page *page;
	DLList *fl = &this->page_list;
	Ref r;
	
	for(r = dllistFirst(fl); 
		r && (page = (Page *) dllistContents(fl, r))->should_delete;
		dllistNext(fl, &r));

	this->setCurrentPage(r);
}

int
Mode::countPages()
{
	int i;
	Ref r;
	DLList *fl=&this->page_list;
	
	for (i=0, r=dllistFirst(fl); r; dllistNext(fl,&r))
		if (!((Page *) dllistContents(fl, r))->should_delete) i++;
	return i;
//   return dllistLength(&this->page_list);
}

Ref
Mode::forwardPageRef()
{
	DLList *fl = &this->page_list;
	Ref r;
	Page *page;
	
	if((r = this->current_page_ref))
		dllistNext(fl, &r);
	
	for(;
       r && (page = (Page *) dllistContents(fl, r))->should_delete;
       dllistNext(fl, &r));
	
	return r;
}

Ref
Mode::backwardPageRef()
{
	DLList *fl = &this->page_list;
	Ref r;
	Page *page;
	
	if((r = this->current_page_ref))
		dllistPrev(fl, &r);
	
	for(;
       r && (page = (Page *) dllistContents(fl, r))->should_delete;
       dllistPrev(fl, &r));
	
	return r;
}


int
Mode::forward()
{
	Ref r;
	
	r=this->forwardPageRef();
	if(r) {
		this->setCurrentPage(r);
		return TRUE;
	} else
		return FALSE;
}

int
Mode::back()
{
	Ref r;
	
	r=this->backwardPageRef();
	if(r) {
		this->setCurrentPage(r);
		return TRUE;
	} else
		return FALSE;
}

void
Mode::setParentWidget (Widget parent)
{
	this->scrolled_window = parent;
}

void
Mode::setMag(int mag_exp, int old_mag)
{
	Page *page;

	page = this->getCurrentPage();
	if (page) {
		page->setMag(mag_exp, old_mag);
	}
}

int 
Mode::managedP() 
{
	Page *current_page = this->getCurrentPage();
	if(current_page == 0)
		return 0;
	return current_page->managedP();
}

SelectionManager *Mode::getSM()
{
	return this->main_window->sm;
}

void
Mode::addPage(Page *page)
{
	Page *current_page = this->getCurrentPage();
	dllistAppend(&this->page_list, page);
	this->current_page_ref = dllistSeek(&this->page_list, current_page);
}

Ref
Mode::addPageBefore(Ref p_r, Page *page)
{
	Page *current_page = this->getCurrentPage();

	p_r = dllistInsertBefore(&this->page_list, p_r, page);

	this->current_page_ref = dllistSeek(&this->page_list, current_page);

	return p_r;
}
