
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

// *******************************************************
// This pragma forces inclusion of debugging code
// from the named header file or header or same name
#ifdef __GNUC__
#pragma implementation
#endif

#include <stdio.h>

#include "illum.h"

#include "misc/defines.h"

#include "OOCFrame.h"
#include "EntityAux.h"

#include "OOCView.h"
#include "MainWindow.h"
#include "Document.h"
#include "CE.h"
#include "bb.h"
#include "uString.h"
#include "ViewCanvas.h"

extern "C" int compareCETree(void *a, void *b);
extern "C" int compareEntTree(void *a, void *b);


OOCView::OOCView(MainWindow *input_main_window, iEntityPtr root_entity,int type, int flagged)
	: View(input_main_window)
{
	this->ooc_type=type;
	this->ooc_flagged=flagged;
	if (this->ooc_type==iGlyphT) this->view_type=vtOOC;
	else this->view_type=vtFLAG;
	
	this->createFrames(root_entity);
}

OOCView::~OOCView()
{
	// not b---- because we have no local members
}

void 
OOCView::chooseMembers(iEntityPtr entity, DLList *m)
{
	iEntityPtr child;
	int ok;
	iBox bounds;
	
	if (!(entity = pickpath(entity))) 
		return;
	bounds = i_GetBounds(entity);
	int null_bounds = bbNullP(bounds);	

	ok = (null_bounds==0);
	if (ok && this->ooc_type)    ok = (i_GetEntityType(entity)==this->ooc_type);
	if (ok && this->ooc_flagged) ok = entityFlaggedP(entity);
	if (ok) dllistAppend(m, entity);

	for(child = i_GetChildEntity(entity); child; child = i_GetNextEntity(child))
		this->chooseMembers(child, m);
}

int
compareEnts(VoidPtr a, VoidPtr b)
{
	iEntityPtr entity_a = (iEntityPtr) a;
	iEntityPtr entity_b = (iEntityPtr) b;
	
	int result;
	uString ua(getFirstOfPosset(entity_a));
	uString ub(getFirstOfPosset(entity_b));
	
	if (ua < ub) result=-1;
	else if (ua == ub) result=0;
	else result=1;
	
	return result;
}

int
compareEntType(VoidPtr a, VoidPtr b)
{
	int int_a = i_GetEntityType((iEntityPtr) a);
	int int_b = i_GetEntityType((iEntityPtr) b);
	
	if (int_a<int_b) return -1;
	else if (int_a==int_b) return 0;
	else return 1;
}

int
compareFrameType(VoidPtr a, VoidPtr b)
{
	iEntityPtr ent_a = ((OOCFrame *) a)->getFirstMember();
	iEntityPtr ent_b = ((OOCFrame *) b)->getFirstMember();
	if (ent_a==ent_b) 
		return 0; 
	// really just handling case where ent_a=ent_b=0
	if (ent_a==0) return -1;
	if (ent_b==0) return +1;
	return compareEntType(ent_a,ent_b);
}

int 
compareFrame(VoidPtr a, VoidPtr b)
{
	iEntityPtr ent_a = ((OOCFrame *) a)->getFirstMember(),
	           ent_b = ((OOCFrame *) b)->getFirstMember();

	if(ent_a == ent_b)
		return 0;
	if(ent_a == 0) 
		return -1;
	if(ent_b == 0)
		return 1;

	return compareEnts(ent_a, ent_b);
}

void
OOCView::createFrames(iEntityPtr root_entity)
{
	DLList m;
	
	dllistCtor(&m);
	this->chooseMembers(root_entity, &m);
	
	this->createList(&m);
	dllistDtor(&m);
}

// in flag mode, we want to preserve the order,
// therefore we can't sort as in OOC mode
void OOCView::mergeInListFlag(DLList *m)
{
	Ref p_r;
	Ref e_r;
	OOCFrame *p;
	aString p_name;
	aString e_name;
	iEntityPtr first_entity;
	
	for(e_r = dllistFirst(m); e_r; dllistNext(m, &e_r)) {
		iEntityPtr entity = (iEntityPtr) dllistContents(m, e_r);
		for (p_r = dllistFirst(&this->frame_list); p_r; dllistNext(&this->frame_list,&p_r)) {
			p = (OOCFrame *) dllistContents(&this->frame_list, p_r);
			first_entity=p->getFirstMember();
			if (compareEntType(entity,first_entity)==0) break;
		}
		if (!p_r) {
			// new frame
			e_name = i_TypeToName(i_GetEntityType(entity));
			p_name = e_name + ": "+"flagged";
			
			
			p = new OOCFrame(
				e_name, p_name, this, this->view_type,i_GetEntityType(entity));
			first_entity=entity;
			this->addFrame(p);
			
		}
		p->addMember(entity);
	}
	if (p) {
		Frame *frame = this->getCurrentFrame();
		dllistSort(&this->frame_list, compareFrameType);
		this->current_frame_ref = dllistSeek(&this->frame_list, frame);
	}
}

aString getOOCFrameName(iEntityPtr entity)
{
	const iUCode *c;
	uString us(getFirstOfPosset(entity));
	aString as;
	char name_str[8];
	
	for(c = us.string(); *c; c++)
		if (*c<32 || *c>127) {
			sprintf(name_str,"U+%x",*c);
			as += name_str;
		} 
		else {
			name_str[0] = *c;
			name_str[1] = 0;
			as += name_str;
		}

	return as;
}

void
OOCView::mergeInList(DLList *m)
{
	Ref e_r;
	Ref p_r;
	OOCFrame *p;
	aString e_name;
	int (*sorter)(VoidPtr,VoidPtr);
	
	if (this->ooc_flagged) {
		this->mergeInListFlag(m);
		return;
	}
	
	if (this->ooc_type==iGlyphT) sorter=compareEnts;
	else sorter=compareEntType;
	dllistSort(m, sorter);
	
	p_r = dllistFirst(&this->frame_list);
	if(p_r)
		p = (OOCFrame *) dllistContents(&this->frame_list, p_r);
	else
		p = 0;
	
	for(e_r = dllistFirst(m); e_r; dllistNext(m, &e_r)) {
		iEntityPtr entity = (iEntityPtr) dllistContents(m, e_r);
		if (this->ooc_type!=iGlyphT) {
			e_name = i_TypeToName(i_GetEntityType(entity));
		} else {
			e_name=getOOCFrameName(entity);
		}

		while(p_r && sorter(p->getFirstMember(),entity) < 0) {
			dllistNext(&this->frame_list, &p_r);
			if(p_r)
				p = (OOCFrame *) dllistContents(&this->frame_list, p_r);
			else 
				p = 0;
		}
		
		if(p && (e_name == p->e_name)){
			iEntityPtr prev;

			p->addMember(entity);
			if(this->ooc_type != iGlyphT)
				prev = i_FindEntityBack(entity, i_GetEntityType(entity), ALL, NULL);
			else{
				int top_char = i_GetGlyph(entity, NULL);
				prev = i_FindEntityBack(entity, iGlyphT, POSSET, (void *) top_char);
			}
			p->sortMember(entity, prev);
		}
		else {
			aString p_name;
			if(e_name == "")
				p_name = "Non-rec";
			else
				p_name = e_name;
			p = new OOCFrame(e_name, p_name, this, this->view_type,iGlyphT);
			p_r = this->addFrameBefore(p_r, p);
			p->addMember(entity);
		}
	}
}

static const char *
makeKey(iEntityPtr entity, OOCView *that)
{
	static aString key;

	if (that->ooc_type != iGlyphT) 
		key = i_TypeToName(i_GetEntityType(entity)); 
	else 
		key = getOOCFrameName(entity);

	return key.string();
}

void
OOCView::createList(DLList *entity_list)
{
	// blowout if flagged
	if (this->ooc_flagged) {
		this->mergeInListFlag(entity_list);
		return;
	}

	// create a hash table of the frame_list as entries
	stringVoidPtrMap map;
	svmapInit(&map, 0, 10001);

	Ref r;
	OOCFrame *frame;
	aString key;
	iEntityPtr entity;

	for(r = dllistFirst(entity_list); r; dllistNext(entity_list, &r)) {
		entity = (iEntityPtr) dllistContents(entity_list, r);

		key = makeKey(entity, this);
		if(svmapContainsP(&map, key.string())) {
			frame = (OOCFrame *) svmapGet(&map, key.string());
			frame->addMember(entity);
		}
		else {
			aString f_name, e_name = key;
			if(e_name == "")
				f_name = "Non-rec";
			else
				f_name = e_name;

			frame = new OOCFrame(e_name, f_name, this, this->view_type, iGlyphT);
			svmapSet(&map, key.string(), frame);
			frame->addMember(entity);
			dllistAppend(&this->frame_list, frame);
		}
	}

	// preserve the current frame reference
	Frame *current_frame = this->getCurrentFrame();
	if(this->ooc_type != iGlyphT)
		dllistSort(&this->frame_list, compareFrameType);
	else
		dllistSort(&this->frame_list, compareFrame);
	this->current_frame_ref = dllistSeek(&this->frame_list, current_frame);

	svmapCleanup(&map);
}

void
OOCView::configure()
{
	OOCFrame *frame;
	DLList *fl = &this->frame_list;
	Ref r;
	
	r = this->current_frame_ref;
	if(r != 0){
		frame = (OOCFrame *) dllistContents(fl, r);
		frame->configure();
	}
}

void
OOCView::addNewEntity(iEntityPtr new_entity)
{
	DLList l;
	dllistCtor(&l);
	dllistAppend(&l, new_entity);
	this->mergeInList(&l);
	dllistDtor(&l);
}

void
OOCView::choose()
{
	
}



void
OOCView::centerOnSelected(iEntityPtr entity)
{
	// bogus, find the frame
	// I'm on not the current frame
	
	OOCFrame *frame;
	CE *ce;
	ce = (CE *) eaGetFace(entity, this->view_type);
	frame = (OOCFrame *) (ce ? ce->frame : 0);
	DLList *l = &this->frame_list;
	Ref r;
	r = dllistSeek(l, frame);
	if(r){
		this->setCurrentFrame(r);
#if 1
		frame->centerOnSelected(entity);
#endif
	}      
}

iBox
OOCView::getBounds(iEntityPtr entity)
{
	iBox bounds;
	
	// -- check that the frame matches
	bounds.w=0;
	
	OOCFrame *current_frame;
	OOCFrame *entitys_frame;
	CE *ce;
	
	ce = (CE *) eaGetFace(entity, this->view_type);
	if (ce) {
		entitys_frame = (OOCFrame *) ce->frame;
		current_frame = (OOCFrame *) this->getCurrentFrame();
		
		if(entitys_frame == current_frame) {
			bounds=ce->computeBounds();
		}
	}
	
	return bounds;
}


void
OOCView::setCurrentFrame(Ref r)
{
	OOCFrame *frame;
	this->View::setCurrentFrame(r);
	
	frame = (OOCFrame *) this->getCurrentFrame();
	if(frame)
		this->main_window->document->last_ooc_page = frame->name;
}


int
OOCView::forward()
{
	Ref r;
	
	OOCFrame *current_frame = (OOCFrame *) this->getCurrentFrame();
	if(current_frame->current_page + 1 < current_frame->num_pages) {
		current_frame->setPage(current_frame->current_page + 1);
		return TRUE;
	}

	r=this->forwardFrameRef();
	if(r) {
		this->setCurrentFrame(r);
		return TRUE;
	} else
		return FALSE;
}

int
OOCView::back()
{
	Ref r;
	
	OOCFrame *current_frame = (OOCFrame *) this->getCurrentFrame();
	if(current_frame->current_page - 1 >= 0) {
		current_frame->setPage(current_frame->current_page - 1);
		return TRUE;
	}
	
	r=this->backwardFrameRef();
	if(r) {
		this->setCurrentFrame(r);
		current_frame = (OOCFrame *) this->getCurrentFrame();
		current_frame->setPage(current_frame->num_pages - 1);
		return TRUE;
	} else
		return FALSE;
}


