
/*
 * Copyright (c) 1995 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"
#include "illum.h"
#include "rafassert.h"

#include "misc/XArg.h"
#include "lte/oscroll.h"

#include "RTEUniEditor.h"
#include "RTElement.h"
#include "RTEPage.h"
#include "MainWindow.h"
#include "SelectionManager.h"
#include "Mode.h"
#include "Application.h"
#include "IllCanvas.h"


RTEUniEditor::RTEUniEditor(Canvas *input_canvas,
						   int input_scroll_flag,
						   MainWindow *input_mw,
						   RTElement *pe)
  : IllEditor(input_canvas, input_scroll_flag, input_mw),
	parent_element(pe)
{
}

void
RTEUniEditor::setSize(int w, int h)
{
	this->parent_element->setSize(w, h);
}


int 
RTEUniEditor::setCursor(int line, int column)
{
  this->parent_element->page->setActiveEditor(this);
  int result = this->UniEditor::setCursor(line, column);

  iEntityPtr entity = 0;
  int index = 0;
  SelectionManager *sm;

  this->convertLine2DAFS(line, column, 
						 &entity, &index);

  assert(entity != 0 && index >= 0);

  sm = this->mw->sm;
  sm->setRange(entity, index, entity, index);

  return result;
}

int 
RTEUniEditor::setSelection(int begin_line, int begin_column, 
						int end_line, int end_column)
{
  this->parent_element->page->setActiveEditor(this);
  int result = this->UniEditor::setSelection(begin_line, begin_column, 
											 end_line, end_column);
  iEntityPtr begin_entity = 0;
  int begin_index;
  iEntityPtr end_entity = 0;
  int end_index;
  SelectionManager *sm;

  this->convertLine2DAFS(begin_line, begin_column, 
						 &begin_entity, &begin_index);
  this->convertLine2DAFS(end_line, end_column, 
						 &end_entity, &end_index);

  assert(begin_entity != 0 && end_entity != 0 && begin_index >= 0 && end_index >= 0);
  sm = this->mw->sm;
  sm->setRange(begin_entity, begin_index, end_entity, end_index);

  return result;
}

int 
RTEUniEditor::clearSelection(void)
{
  this->parent_element->page->setActiveEditor(this);
  SelectionManager *sm;
  int result = this->UniEditor::clearSelection();
  sm = this->mw->sm;
  sm->clear();

  return result;
}

void
RTEUniEditor::getBasePosition(Position *x, Position *y)
{
  Widget canvas = this->getWidget();
  XArg args;
  args.setArg(XmNx, x);
  args.setArg(XmNy, y);
  args.getValues(canvas);
}

int
RTEUniEditor::isVisible(int x, int y)
{
  Widget canvas = this->getWidget();
  Widget bb = XtParent(canvas);
  Widget sw = XtParent(XtParent(bb));
  Position base_x, base_y;
  // are we managed
  if(XtIsManaged(bb) == FALSE)
	return FALSE;

  this->getBasePosition(&base_x, &base_y);

#if 0
  printf("is visible\n");
  printf("widget coord:      %3d,%3d\n", x, y);
  printf("widget offset:     %3d,%3d\n", base_x, base_y);
  printf("\n");
#endif

  return visiblePositionPAuto(sw, base_x + x, base_y + y);	
}

void
RTEUniEditor::calculateVisible(int *first, int *last)
{
  int foo;
  Widget canvas = this->getWidget();
  Widget bb = XtParent(canvas);
  Widget sw = XtParent(XtParent(bb));
  Position base_x, base_y;
  // are we managed
  if(XtIsManaged(bb) == FALSE) {
	*first = -1;
	*last = -1;
	return;
  }

  iBox vr;

  vr = visibleRectangleAuto(sw);
  this->getBasePosition(&base_x, &base_y);
#if 0
  printf("visible rectangle: %3d,%3d %dx%d\n", vr.x, vr.y, vr.w, vr.h);
  printf("widget offset:     %3d,%3d\n", base_x, base_y);
#endif
  vr.x -= base_x;
  vr.y -= base_y;
  this->coordToPosition(vr.x, vr.y, first, &foo);
  this->coordToPosition(vr.x + vr.w, vr.y + vr.h, last, &foo);
}

void
RTEUniEditor::centerOnCoord(int x, int y)
{
  Widget canvas = this->getWidget();
  Widget bb = XtParent(canvas);
  Widget sw = XtParent(XtParent(bb));
  Position base_x, base_y;
  // are we managed
  if(XtIsManaged(bb) == FALSE) {
	return;
  }

#if 0
  printf("center\n");
  printf("widget coord:      %3d,%3d\n", x, y);
  printf("widget offset:     %3d,%3d\n", base_x, base_y);
  printf("\n");
#endif

  this->getBasePosition(&base_x, &base_y);
  x += base_x;
  y += base_y;

  scrollToPositionAuto(sw, x, y);
}


void
RTEUniEditor::convertLine2DAFS(int line_index, int column, 
			    iEntityPtr *entity_ptr, int *index_ptr)
{
  iEntityPtr entity;
  int index;
  Reference *ref;
  Line *line = this->lines.line[line_index];

  int ref_index = line->reference(column);
  ref = line->getRef(ref_index);
  entity = ref->en;
  index = ref->en_index;

  *entity_ptr = entity;
  *index_ptr = index;
}

int
RTEUniEditor::convertDAFS2Line(iEntityPtr entity, int index,
							int *line_ptr, int *col_ptr)
{
  int line = 0, col = 0;
  Reference *ref;
  Reference *prev_ref = 0;
  int prev_line;
  int i;
  int r;
  int success = TRUE;

  for(i = 0; i < this->lines.size; i++) {
	Line *lp = this->lines.line[i];
	for(r = 0; r < lp->num_refs; r++) {
	  ref = lp->refs + r;
	  if(prev_ref && prev_ref->en == entity && (ref->en != entity || ref->en_index > index))
		break;
	  prev_ref = ref;
	  prev_line = i;
	}
	if(r < lp->num_refs)
	  break;
  }

  if(prev_ref->en == entity){
	line = prev_line;
	col = prev_ref->text_index + (index - prev_ref->en_index);
  } else {
	success = FALSE;
  }
	
  *line_ptr = line;
  *col_ptr = col;

  return success;
}



int 
RTEUniEditor::pressKey(iUCode key_value)
{
  int initial;
  int final;

  this->insertAtSelection(key_value, &initial, &final);
  this->conditionallyResize();
  this->drawRange(initial, final);

  return TRUE;
}

void
RTEUniEditor::processDelete(void)
{
  int initial = 0, 
	final  = this->lastLine();

  this->deleteAtSelection(&initial, &final);

  this->conditionallyResize();
  this->drawRange(initial, final);
}


void
RTEUniEditor::compare()
{
  UniEditor *other;
  RTEPage *rte_page = (RTEPage *) this->parent_element->page;
  MainWindow *mw = rte_page->mode->main_window;
  IllCanvas *nc = new IllCanvas(0, mw->app->app_context,
								rte_page->bb, 0, 0,
								mw);
  other = new RTEUniEditor(nc, FALSE, mw, this->parent_element);
  other->setEntity(this->root_entity);
  this->UniEditor::compare(other);
  delete other;
}
