
/*
 * Copyright (c) 1994, 1995 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"
#include "c-type.h"
extern "C" {
#include "../dafslib/entity.h"
};
#include "illum.h"


#include <X11/Xutil.h>

#include <Xm/MainW.h>
#include <Xm/Frame.h>
#include <Xm/Label.h>
// really used for menus
#include <Xm/PushB.h>
#include <Xm/PushBG.h>
#include <Xm/ToggleBG.h>
// really used for creating x window stuff
#include <Xm/RowColumn.h>
#include <Xm/PanedW.h>
#include <Xm/ScrolledW.h>
#include <Xm/ScrollBar.h>

#include "rafutil.h"
#include "smemory.h"


#include "misc/XArg.h"
#include "misc/color.h"
#include "misc/misc.h"

#include "TypePrefDialog.h"

#include "config.h"

extern "C" void i_DisposeTypeTab(iTypeTablePtr tab);
extern "C" iDAFSError i_PNameToType(iTypeTablePtr tab,const char *type, iTypeIndex *val);

extern "C" void i_DisposeTypeTab(iTypeTablePtr tab);

iDAFSError readSessionProps(const char *fileName);
iDAFSError writeSessionProps(char *fn);


#define STD_TYPE_COLOR "blue"
#define STD_TYPE_LENGTH 256
#define STD_TYPE_FIELD FALSE
#define STD_TYPE_PAGE FALSE
#define STD_TYPE_VISIBLE TRUE


static char *config_file = 0;

void 
readConfig(char *file)
{
  int fail = FALSE;
  iDAFSError err;
  iTypeIndex t;
  int4 v = 1;
  i_NameToType("PO", &t);
  fail = fail || (err=i_SetSessionBoolProp(t, PAGE_PROP, v));
  i_NameToType("Document", &t);
  fail = fail || (err=i_SetSessionBoolProp(t, PAGE_PROP, v));
  fail = fail || (err=i_SetSessionBoolProp(t, FIELD_PROP, v));
  t = iParagraphT;
  fail = fail || (err=i_SetSessionBoolProp(t, FIELD_PROP, v));
  t = iZoneT;
  fail = fail || (err=i_SetSessionBoolProp(t, FIELD_PROP, v));
  i_NameToType("Column", &t);
  fail = fail || (err=i_SetSessionBoolProp(t, FIELD_PROP, v));
  i_NameToType("Field", &t);
  fail = fail || (err=i_SetSessionBoolProp(t, FIELD_PROP, v));

	::config_file=file;
	readSessionProps(file);
}

void
writeConfig(void)
{
  writeSessionProps(::config_file);
}

void
setTypeColor(char *type, char *color)
{
  iTypeIndex t;
  iDAFSError err;
	
  if (type && color){
	if (!type[0]) 
	  type="<NULL>";
	err = i_NameToType(type, &t);
	assert(err == DAFSOK);
	i_SetSessionStringProp(t, COLOR_PROP, color);
  }
}

const char*
getTypeColor(const char * type)
{	

  iDAFSError err;
  const char *color = STD_TYPE_COLOR;

  assert(type != 0);

  disableCallbacks();

  if (!type[0])
	type="<NULL>";
		

  iTypeIndex t;
  err = i_NameToType(type, &t);

  if(err == DAFSOK) {
	iEntityPtr foo;
	foo = i_NewEntity(t, 0);
	color = getColor(foo);
	i_DisposeEntity(foo);
  }

  enableCallbacks();
	
  return color;
}

const char*
getColor(iEntityPtr entity)
{	
  assert(entity != 0);
  iDAFSError err;
  const char *color;

  err = i_GetStringProp(entity, COLOR_PROP, &color);
  if(err == DAFSOK) {
	return color;
  } else {
	return STD_TYPE_COLOR;
  }
}

void
setTypePage(char *type, int value)
{
  iTypeIndex t;
  iDAFSError err;
	
  if (type){
	if (!type[0]) 
	  type="<NULL>";
	err = i_NameToType(type, &t);
	assert(err == DAFSOK);
	i_SetSessionBoolProp(t, PAGE_PROP, value);
  }
}

int
getTypePage(const char * type)
{	

  iDAFSError err;
  int page = STD_TYPE_PAGE;

  assert(type != 0);

  disableCallbacks();

  if (!type[0])
	type="<NULL>";
		

  iTypeIndex t;
  err = i_NameToType(type, &t);

  if(err == DAFSOK) {
	iEntityPtr foo;
	foo = i_NewEntity(t, 0);
	page = getPage(foo);
	i_DisposeEntity(foo);
  }

  enableCallbacks();
	
  return page;
}


void
setTypeField(char *type, int value)
{
  iTypeIndex t;
  iDAFSError err;
	
  if (type){
	if (!type[0]) 
	  type="<NULL>";
	err = i_NameToType(type, &t);
	assert(err == DAFSOK);
	i_SetSessionBoolProp(t, FIELD_PROP, value);
  }
}

int
getTypeField(const char * type)
{	

  iDAFSError err;
  int field = STD_TYPE_FIELD;

  assert(type != 0);

  disableCallbacks();

  if (!type[0])
	type="<NULL>";
		

  iTypeIndex t;
  err = i_NameToType(type, &t);

  if(err == DAFSOK) {
	iEntityPtr foo;
	foo = i_NewEntity(t, 0);
	field = getField(foo);
	i_DisposeEntity(foo);
  }

  enableCallbacks();
	
  return field;
}


void
setTypeVisible(char *type, int value)
{
  iTypeIndex t;
  iDAFSError err;
	
  if (type){
	if (!type[0]) 
	  type="<NULL>";
	err = i_NameToType(type, &t);
	assert(err == DAFSOK);
	i_SetSessionBoolProp(t, VISIBLE_PROP, value);
  }
}
static void
set_all_nonvisible(iTypeIndex type,iEntityPtr entity)
{
	iEntityPtr cur;
	
	if (type != i_GetEntityType(entity))
		i_SetBoolProp(entity,VISIBLE_PROP,0);
	else 
		i_SetBoolProp(entity, VISIBLE_PROP, 1);
	
	for (cur = i_GetChildEntity(entity);cur;cur = i_GetNextEntity(cur))
		set_all_nonvisible(type,cur);
}
int 
turn_off_others(const char *type, iEntityPtr root_entity)
{
	iTypeIndex itype;
	
	i_NameToType(type, &itype);

	set_all_nonvisible(itype,root_entity);
	
}
int
getTypeVisible(const char * type)
{	

  iDAFSError err;
  int visible = STD_TYPE_VISIBLE;

  assert(type != 0);

  disableCallbacks();

  if (!type[0])
	type="<NULL>";
		

  iTypeIndex t;
  err = i_NameToType(type, &t);

  if(err == DAFSOK) {
	iEntityPtr foo;
	foo = i_NewEntity(t, 0);
	visible = getVisible(foo);
	i_DisposeEntity(foo);
  }

  enableCallbacks();
	
  return visible;
}

static iTypeTable Local;


iDAFSError
readHeader(RafFile fo, iDAFSWriteMode *mode)
{
	char b[2];
	
	if (riRead(b,1,2,fo) != 2)
		return ReadError;
	if (!b[0] && !b[1]) {
		*mode = DAFSB;
		return DAFSOK;
	}
	if (b[0]=='<' && (isascii(b[1]))){
#ifdef IMPLEMENT_DAFSA
		*mode = DAFSA;
		riSeek(fo,0,0);
#else
		halt("dafsa not implemented.\n");
#endif
		return DAFSOK;
	}
	riSeek(fo,0,0);
	return ReadError;
}


extern "C" iDAFSError 
ReadProps(RafFile fo,iEntity *me, iGlobid *gid,iTypeIndex et);

static iDAFSError 
_readSessionProps(RafFile fo, iGlobid *gid)
{
	iDAFSError e;
	char *type=0;
	iToken t;
	uint2 size;
	iTypeIndex et;
	iDAFSWriteMode mode;
	
	Local.size=0;
	disableCallbacks();
	if (e = readHeader(fo,&mode))
		goto error;

	switch (mode) {
	case DAFSB:
		if (e = i_ReadToken(fo,&t,&gid->DAFSBSwap))
			goto error;
		while ((t & iTMask) != BeginEntity) {
			if ((t & iTMask) == TString){
				if (e = i_ReadItemAlloc(fo,t,&size,(void**)&type,&gid->DAFSBSwap))
					goto error;
				if (e=i_PNameToType(&Local,type,0))
					goto error;
				if (e = i_NameToType(type,&et))
					goto error;
				if (e = i_ReadToken(fo,&t,&gid->DAFSBSwap))
					goto error;
				if ((t & iTMask) == BeginProp) {
					if (e = ReadProps(fo, 0, gid,et))
						goto error;
					if (e = i_ReadToken(fo,&t,&gid->DAFSBSwap))
					  goto error;
				}
				smfree(type);
				type=0;
			} else {
			  // curious -- seems to throw away whatever it is and
			  // suck up the next token
			  if (e = i_ReadItem(fo,t,0,&size,0,&gid->DAFSBSwap))
				goto error;
			  if (e = i_ReadToken(fo,&t,&gid->DAFSBSwap))
				goto error;
			}
		}
		break;
	}

  error:
	i_DisposeTypeTab(&Local);
	if (type) smfree(type);
	enableCallbacks();
	return e;
}


iDAFSError
readSessionProps(const char *fileName)
{
	iGlobid gid={0,0,0,0,0};
	iDAFSError e;
	RafFile fo;
	
	gid.im_flag=1;
	fo = riOpen(fileName,"rb");
	if (!fo) {
	  return NonExistentFileError;
	}
	e = _readSessionProps(fo,&gid);
	riClose(fo);
	if (gid.ims) smfree(gid.ims);

	return e;
}

extern "C" iDAFSError 
WriteProps(RafFile fo,iPropPtr props);

static iDAFSError 
WriteTypeTab(RafFile fo)
{
  iDAFSError e=DAFSOK;
  char *type;
  int4 i;
  iPropPtr prop;
  
  for (i=0;i<i_GlobalType.size;i++) {
	if (type= i_GlobalType.table[i].name) { 
	  e = i_WriteItem(fo,TString,strlen(type)+1,type);
	  if (!e && (prop = i_GlobalType.table[i].props))
		e = WriteProps(fo, prop);
	  if (e)
		return e;
	}
  }
  return DAFSOK;
}


static iDAFSError 
writeHeader(RafFile fo)
{
	iDAFSError e = DAFSOK;
	short s = 1;
	
	if (e = i_WriteItem(fo,NullTok,0,0))
	  return e;
	if (e = i_WriteItem(fo,NullTok,0,0))
	  return e;
	if (e = i_WriteItem(fo,ByteOrder,2,&s))
	  return e;

	return e;
}

iDAFSError 
writeSessionProps(char *fn)
{
  RafFile fo;

  fo = fopen(fn, "w");
  iDAFSError e;
  iGlobid gid={0,0,0,0,0};
	
  if (e = writeHeader(fo)) 
	goto error;

  gid.en_id = 1;
  gid.im_id = 1;
  e = WriteTypeTab(fo);
error:
  riClose(fo);
  return e;
}


int 
getPage(iEntityPtr en)
{
  iDAFSError e;
  int4 val;

  e = i_GetBoolProp(en, PAGE_PROP, &val);
  return (e == DAFSOK) ? val : 0;
}


int 
getField(iEntityPtr en)
{
  iDAFSError e;
  int4 val;

  e = i_GetBoolProp(en, FIELD_PROP, &val);
  return (e == DAFSOK) ? val : 0;
}

int isVisibleSet(iEntityPtr en)
{
	return (!i_GetBoolProp(en, VISIBLE_PROP, 0));
}
int 
getVisible(iEntityPtr en)
{
  iDAFSError e;
  int4 val;

  e = i_GetBoolProp(en, VISIBLE_PROP, &val);
  return (e == DAFSOK) ? val : 1;
}

void
disposeColorTable(void)
{
}
