
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"

#include <stdio.h>
#include <sys/types.h>
#include <stdlib.h>

#include <X11/StringDefs.h>
#include <X11/Intrinsic.h> 
#include <X11/Xutil.h>


#include <Xm/ScrollBar.h>

#include "rafutil.h"
#include "aString.h"
#include "misc/XArg.h"

#ifdef RAF_INTERNAL
#include "form.h"
#include "ellipse.h" 
#endif

#include "illum.h"

#include "Application.h"
#include "Document.h"
#include "MainWindow.h"
#include "EntityAux.h"
#include "ViewCanvas.h"
#include "SelectionManager.h"
#include "misc/Unicode.h"

#ifdef RAF_INTERNAL
/* bogus, shouldnt declare MAXNGRAM more than once */
#define MAXNGRAM 3
fDictionaryPtr di = 0; 
fNGramPtr gp[MAXNGRAM];
fCharSetPtr cp = NULL;
int nngrams = 0;
#endif

// MainWindow *the_main_window = 0;  no longer supported


// sets the text measurement function to be used by
// TextElement class
extern void initializeMeasureFunction( textMeasureFunction f );
extern void illumMeasureText(const iUCode *s, int l, int *tw, int *th, int *ascent)
{
        measureUnicode(XtDisplay(the_application->app_shell),s,l,tw,th,ascent);
}

void initializeText()
{
	// Initialize the TextElement class for use by this application.
	// Currently all we need to do is tell it which text measurement
	// function to use.

	initializeMeasureFunction(&illumMeasureText);
}

#ifdef RAF_INTERNAL
void
initializeRecognition()
{
	static int was_here = FALSE;
		
	char *probfile = "ngram%d.db";
	char *wordfile = "words.di";
	int i;
	char name[64];
	iDAFSError e;
	
	if(was_here == TRUE){
		if (cp)
			f_DisposeCharSet(cp);
		e = f_ReadCharSet(&cp,ellipse_file);
		if (!e && i_SetSessionUserPtrProp(iWordT,"RAFData.0",cp,NULL))
			halt("Failed to set property for charset\n");
	} else {
		f_InitLib();
		
		for (i=0;i<MAXNGRAM;i++) gp[i] = 0;
		nngrams = 0;
		
		if (probfile) {
			for (nngrams = 0; nngrams < MAXNGRAM; nngrams++) {
				sprintf(name, probfile, nngrams);
				e = f_ReadNGram(&gp[nngrams],name);
				if (e)
					break;
				sprintf(name, "RAFData.%d", nngrams);
				if (e=i_SetSessionUserPtrProp(iLineT,name,gp[nngrams],NULL))
					halt("Failed to set property int %s\n",name);
			}
		}
		if (wordfile) {
			e = f_ReadDict(&di,wordfile);
			if (e) {
#if 0
				fprintf(stderr, "Error reading %s, so it's being ignored\n", wordfile);
#endif
				di = NULL;
			} else if (e=i_SetSessionUserPtrProp(iWordT,"RAFData.0",di,NULL))
				halt("Failed to set property for dictionary\n");
		}
		
		e = f_ReadCharSet(&cp,ellipse_file);
		if (!e && i_SetSessionUserPtrProp(iGlyphT,"RAFData.0",cp,NULL))
			halt("Failed to set property for charset\n");
		
		was_here = TRUE;
	}
}
#endif



/*
// bogus -- shouldnt have different versions
// of this scattered everywhere
// should have a standard way of making this happen
*/
iEntityPtr 
pickpath(iEntityPtr doc)
{
#ifndef IGNORE_ORS
	iEntityPtr sub;
	
	if (i_GetLogicalType(doc) == iAnd) 
		return doc;
	
	sub = i_GetChildEntity(doc);
	return sub;
#else
	return doc;
#endif
}

void
ringTheBell()
{
	XBell(XtDisplay(the_application->app_shell), 0);
}

/*
// dont forget to handle swapping
*/
iBox 
i_GetPBBBounds(iEntityPtr entity)
{
	iBox pbb;
	void *vp;
	
	pbb.x = pbb.y = pbb.w = pbb.h = 0;
	
	if (!i_GetUserDataProp(entity, "PBB", &vp, 0, i_SwapUBox)){
		pbb = *((iBox *) vp);
		iBox cbb = i_GetEnclosingBox(entity);
		pbb.x += cbb.x;
		pbb.y += cbb.y;
	}
	
	return pbb;
}
void 
i_SetPBBBounds(iEntityPtr entity, iBox pbb)
{
	iBox cbb = i_GetEnclosingBox(entity);
	pbb.x -= cbb.x;
	pbb.y -= cbb.y;
	
	i_SetUserDataProp(entity, "PBB", (void*)&pbb, sizeof(iBox));
}

/*
// bogus this should be part of SelectionManager or MainWindow
*/
void
killContentsForSelection()
{
  const DLList *items;
  iEntityPtr en;
  Ref r;
  SelectionManager *sm = the_window->sm;
  items = sm->getItems();

  for(r = dllistFirst(items); r; dllistNext(items, &r)) {
	en = (iEntityPtr) dllistContents(items, r);
	i_SetNoData(en);
  }
  the_window->document->setDirty();
}


void
whereIsInputFocus()
{
	Display *display = XtDisplay(the_application->app_shell);
	Window window = 0;
	Widget widget = 0;
	int foo;
	void *user_data = 0;
	XArg args;

	XGetInputFocus(display, &window, &foo);
	if(window) {
		widget = XtWindowToWidget(display, window);
		if(widget) {
			args.setArg(XmNuserData, &user_data);
			args.getValues(widget);
		}
	}

	printf("input focus\n");
	printf("\twindow: %d\n", window);
	printf("\twidget: %d\n", widget);
	printf("\tuser data: %d\n", user_data);
}
