
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#ifndef DLLIST_H
#define DLLIST_H

#include "rafport.h"
#include "compare.h"

struct DLListNode {
	struct DLListNode *next;
	struct DLListNode *prev;
	VoidPtr value;
};

#ifndef __cplusplus
typedef struct DLListNode DLListNode;
#endif


typedef int dllError;
#define DLLIST_OK 0
/* error is always memory fail error */
#define DLLIST_MEMERROR 1

struct DLList {
	DLListNode start;
	DLListNode end;
	int4 count;
};

#ifndef __cplusplus
typedef struct DLList DLList;
#endif

#ifdef __cplusplus
extern "C" {
#endif
	
/* Singly Linked Stuff */
	void dllistCtor(DLList *);
	void dllistDtor(DLList *);
	dllError dllistAssign(DLList *lvalue, const DLList *rvalue);
	
	int dllistEmpty(const DLList *that);
	int4 dllistLength(const DLList *that);
	
	dllError dllistPrepend(DLList *that, VoidPtr value);
	dllError dllistAppend(DLList *that, VoidPtr value);
	
	void dllistJoin(DLList *that, DLList *other);
	
	VoidPtr dllistFront(const DLList *that);
	VoidPtr dllistRear(const DLList *that);
	VoidPtr dllistPopFront(DLList *that);
	void dllistDeleteFront(DLList *that);
	
	void dllistClear(DLList *that);
	
	
	void dllistDeleteAfter(DLList *, Ref r);
	
/* Doubly linked stuff */
	VoidPtr dllistPopRear(DLList *that);
	void dllistDeleteRear(DLList *that);
	Ref dllistInsertAfter (DLList *that, Ref r, VoidPtr item);
	Ref dllistInsertBefore(DLList *that, Ref r, VoidPtr item);
#define DLLIST_FORWARD 0
#define DLLIST_BACKWARD 1
	void dllistDelete(DLList *, Ref *r, int4 direction);
	
/* generic container stuff */
	Ref  dllistFirst(const DLList *that);
	void dllistNext(const DLList *that, Ref *ref);
	VoidPtr dllistContents(const DLList *that, Ref ref);
	
	Ref  dllistLast(const DLList *that);
	void dllistPrev(const DLList *that, Ref *ref);
	
	Ref dllistSeek(const DLList *that, VoidPtr value);
	
	dllError dllistSort(DLList *that, ComparisonFunction compare_function);
	int dllistEqual(const DLList *left, const DLList *right);
	
#if 1
	void dllistRemove(DLList *that, Ref);
#endif
	
#ifdef __cplusplus
};
#endif


#endif
