
/*
 * Copyright (c) 1995 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#ifndef _Line_h_
#define _Line_h_

#include "uString.h"
class LineList;

struct Reference {
  iEntityPtr en;
  int en_index, text_index;
};

class Line {
  uString text, context;
  iUCode *indices;

  LineList *manager;

  int render_length;

  void parseContext(void);
  void addUnicode(const iUCode *uni, const iEntityPtr en, int index);
  void removeChar(int index);

public:
  int num_refs, alloc_refs;
  Reference *refs;
  
public:
  Reference *getRef(int ref_index) { return this->refs + ref_index; }
  Line(LineList *lines);
  Line(const Line &);
  ~Line(void);

  Line &operator +=(Line &other);

  int textLength(void) { return text.length(); };
  const iUCode *getText(void) { return text.string(); };
  int text2base(int raw);
  int base2text(int base);
  int nonspacing(unsigned raw);
  void setText(const iUCode *new_text);
  void remove(int start, int length);
  void insert(iUCode uni, int pos);
  uDirection direction(int base);

  int contextLength(void) { return context.length(); };
  const iUCode *getContext(void) { return context.string(); };
  const iUCode getContext(int i) { return context[i]; };
  int context2base(int pos);
  int base2context(int base);

  int baseLength(void) { return text2base(text.length()); };
  int screen2base(int screen);
  int base2screen(int pos);
  int renderLength(void) { return render_length; };
  int renderLength(Display *display, int x);	
  int calculateLength(Display *display);

  friend class LineList;
  int reference(int index);
  int offset(int index);
  int referenceEnd(int ref);
  int referenceLength(int ref);
  void splitReference(int reference, int index);
  void insertReference(int reference, iEntityPtr en, int offset, int length);
  void growReference(int reference, int amount);
  void changeReference(int reference, iEntityPtr en, int amount);
  void deleteReference(int reference);

  friend void prlines(LineList *lines);
public:
  void pr();
  int compare(Line *other);
};

#endif



