
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

/*
// This file defines the PriotityQueue class which is just what it
// sounds like -- an abstract data type that has well defined operations:
//    insert an element
//    delete the largest element
*/

#ifndef PQ_H
#define PQ_H

#include "rafport.h"
#include "c-lib.h"
#include "compare.h"

struct PriorityQueue {
	VoidPtr *array;
	int max_elements;
	int num_elements;
	ComparisonFunction comparison_function;
};

#ifndef __cplusplus
typedef struct PriorityQueue PriorityQueue;
#endif

#ifdef __cplusplus
extern "C" {
#endif
	
	void pqCtor(PriorityQueue *that, ComparisonFunction);
	void pqDtor(PriorityQueue *that);
	int pqEmpty(PriorityQueue *that);
	int pqFull(PriorityQueue *that);
	int pqLength(PriorityQueue *that);
	void pqEnqueue(PriorityQueue *that, VoidPtr);
	VoidPtr pqDequeue(PriorityQueue *that);
	VoidPtr pqFront(PriorityQueue *that);
	void pqRemoveFront(PriorityQueue *that);
	void pqClear(PriorityQueue *that);
	
#ifdef __cplusplus
};
#endif

#endif
