#ifndef _RTEMisc_
#define _RTEMisc_

#include "Rectangle.h"
#include "RTEView.h"
#include "RTESet.h"
#include "Compare.h"


// -------- Forward declarations of classes from other modules --------

class Rectangle;
class ChildRectangle;
class PackedRectangle;
class CharacterElement;
class ParentRectangle;
class Interval;



// --------- Prototypes for routines in RTEMisc.cc --------

Direction opposite(Direction);

int compare(Interval *, Interval *);
int compare(ChildRectangle *child_a, ChildRectangle *child_b);
int compare(Rectangle *rect_a, Rectangle *rect_b);
int compare(Coord *, Interval *);

int compareII(const void *v1, const void *v2);
int compareCRCR(const void *v1, const void *v2);
int compareRR(const void *v1, const void *v2);

int compareInterval(void *v1, void *v2);
int compareChildRectangle(void *v1, void *v2);
int compareRectangle(void *v1, void *v2);

int compareCI(const void *v1, const void *v2);

Coord nearestMultiple(Coord, Coord);
Coord nearestPositiveMultiple(Coord, Coord);
Coord nearestMultipleGT(Coord, Coord);

// templates used to be here
#include "Generic.h"


// ---------------- Miscellaneous Classes -------------------

class RTESelectionManager {
	ElementSet selected_set;
	ChildRectangle *first, *last;
public:
	void select(ChildRectangle *child);
	void select(ElementSet *);
	void deselect(ChildRectangle *child);
	void deselect(ElementSet *);
	void toggleSelection(ChildRectangle *child);
	void toggleSelection(ElementSet *);
	void clear();
	void getSelected(ElementSet *);
	int isSelected(ChildRectangle *child);
	ElementSet *getSet(void) { return &selected_set; };
	int numSelected(void);

	ChildRectangle *getFirst(void) { return first; };
	ChildRectangle *getLast(void) { return last; };
	void setFirst(ChildRectangle *child) { first = child; };
	void setLast(ChildRectangle *child) { last = child; };
};


class TraversalManager {
        int at_end;
public:
	TraversalManager();
	~TraversalManager();
	ChildRectangle *checkRight(ChildRectangle *child, int goal = -1);
	ChildRectangle *checkLeft(ChildRectangle *child, int goal = -1);
	ChildRectangle *checkUp(ChildRectangle *child, int goal = -1);
	ChildRectangle *checkDown(ChildRectangle *child, int goal = -1);
	int atEnd(void) { return at_end; };
	void clearFlags(void) { at_end = FALSE; };
};


class DeletedPoolManager {
	ElementSet deleted_pool;
public:
	~DeletedPoolManager();
	void sendToDeleted(ChildRectangle *child);
	void sendToDeleted(ElementSet *);
	void getFromDeleted(ChildRectangle *child);
	void getFromDeleted(ElementSet *);
	int isDeleted(ChildRectangle *child);
};


class SimpleInterval {
	friend class IntervalList;
protected:
	Coord x;  // start
	Coord dx; // width
public:
	SimpleInterval(Coord, Coord);
	Coord getX() { assert(x >= 0); return x; };
	Coord getDX() { assert(dx >= 0); return dx; };
};

class Interval : public SimpleInterval {
	ElementSet *elements_present;
	int portionToLeftOf(Interval *, Interval **);
	int portionToRightOf(Interval *, Interval **);
	int intersectionWith(Interval *, Interval **);
	void removeIntersection(Interval *);
	friend class IntervalList;
public:
	Interval(Coord, Coord);
	Interval(Coord, Coord, ChildRectangle *child);
	~Interval();
	void debugPrint(int);
	void debugPrint(char *);
	int isNull() { assert(this->dx >= 0); return this->dx == 0; };
	int isEmpty() { return this->elements_present->isEmpty(); };  // no elements present
	int compare(Coord) ;      // less than, equal(inside), greater_than
	int compare(Interval *) ; // less than, contained, equal, greater_than (rel. to me)
	ElementSet *getElementsPresent();
};


class IntervalList {
	Interval **members;
	int size;
	int insertInterval(Interval *);
	void deleteAtLoc(int);
protected:
	int noOverlaps();
public:
	IntervalList();
	~IntervalList();
	void debugPrint(char *);
	ElementSet * findHit(Coord);
	ElementSet * findHit(Interval *);      // fully or partially included
	ElementSet * findIncluded(Interval *); // fully included
	int add(Interval **);
	void deleteElement(ChildRectangle *child);
	void deleteElement(Interval*, ChildRectangle *child);
	void deleteElements(ElementSet *);
	void deleteElements(Interval *, ElementSet *);
};


#endif
