
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#ifndef _TextElement_h_
#define _TextElement_h_

/*
// the representation of text in illum
//
*/

#include "text.h"

#include "dllist.h"

#include "EntityFace.h"

class VirtualPage;
class TextElement;
typedef TextElement *TextElementPtr;
typedef void ( *textMeasureFunction ) (const iUCode *s, int l, int *tw, int *th, int *ascent);


class TextElement : public EntityFace {
	// entity is in EntityFace
	//    void *user_data; would have been the TextPage in the case
	//    of illum, I dont know what it did for dafs2text
public:
	
	// -- is this necessary
	// is it better left in the entity and computed	?
	iUCode *text;      
	
	// -- is this necessary
	// I think the answer is -- sometimes I make these guys lie
	iPackType *pack;   
	
	// better left in entity?
	// or do I make it lie
	int type;          
	
	// I used to say these were relative to your parent
	// are they now relative to the change in direction?
	short rx;          
	short ry;
	
	short w;
	short h;
	
	short position_valid;
	short dimension_valid;
	
	TextElement(Page *user_data, iEntityPtr entity, 
				int type, TextElement *parent, 
				iPackType *pack);
private:
	virtual ~TextElement();
public:
	/*virtual*/ void deleteAndUpdate();
	/*virtual*/ void deleteCheap();
	
	void initializeText();

	/* traverse functions */
	TextElement *getParentElement();
	
	iEntityPtr getNextChild(iEntityPtr prev_child);
	iEntityPtr getPrevChild(iEntityPtr prev_child);
	
	iPackType *childP(iEntityPtr child);
	iPackType *parentP(iEntityPtr parent);
	
	
	// these routines are for faking lists
	// the actual hierarchy is in the entities
	void insert(VoidPtr input_child);
	
	void insertAfter(VoidPtr new_element, 
					 VoidPtr old_element);
	void remove(Ref r);
	Ref first();
	VoidPtr contents(Ref r);
	void next(Ref *r);
	void prev(Ref *r);
	Ref last();
	Ref seek(VoidPtr sought);
	int length();

	void preLayoutCB();
	void postLayoutCB();
	/* layout functions */

	void setPosition(short new_rx, short new_ry);
	void setDimensions(short new_w, short new_h);

	void getPosition(TextElement *root, int index, short& ax, short& ay);
	void getDimensions(short& rw, short &rh);
	void calculatePosition(TextElement *parent);
	void calculatePosition() { 
		this->calculatePosition(this->getParentElement());
	}

	void calculateDimensions();
	
	void getElementAtPosition(int x, int y, 
							  TextElementPtr& return_element, 
							  int& return_index,
							  int is_root = FALSE,
							  int over_end=FALSE);
	
	void change();
	
	/* hmm -- this is for dafs2text */
	VirtualPage *makePage();
	char *toAscii();
	int indexLength();
	
	/* find next possible input point */
	TextElement *findPrevElement();
	TextElement *findNextElement();
	
	/* callbacks -- EntityFace virtuals */
	/*virtual*/ void create();
	
	/*virtual*/ void willUpdate();
	/*virtual*/ void update();
	
	/*virtual*/ void willMove();
	/*virtual*/ void move();
	
	/*virtual*/ void destroy();
	
	/*virtual*/ void select();
	/*virtual*/ void unSelect();
	
	/*virtual*/ void recognize();
	
	/*virtual*/ MainWindow *getMainWindow();
	friend void textfriend();
};


inline void 
TextElement::insert(VoidPtr input_child)
{
	// do nothing
			  }

inline void 
TextElement::insertAfter(VoidPtr new_element, 
						 VoidPtr old_element)
{
	// do nothing
			  }

inline VoidPtr 
TextElement::contents(Ref r)
{
	return (VoidPtr) r;
}

inline Ref 
TextElement::seek(VoidPtr sought)
{
	return sought;
}

inline void
TextElement::remove(Ref r)
{
	// do nothing
			  }

inline iPackType *
TextElement::childP(iEntityPtr child)
{
	iPackType *pack;
	
	iDAFSError e=i_GetUserDataProp(child,PACK_NAME,(void **)&pack,0,i_SwapPackType);
	if(e == DAFSOK)
		return pack;
	else
		return 0;
}

inline iPackType *
TextElement::parentP(iEntityPtr parent)
{
	iPackType *pack;
	
	iDAFSError e=i_GetUserDataProp(parent,PACK_NAME,(void **)&pack,0,i_SwapPackType);
	if(e == DAFSOK)
		return pack;
	else
		return 0;
}


#endif


