
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */
#include "rafport.h"
#include "str-lib.h"

#include <X11/keysym.h>
#include <Xm/PushB.h>
#include <Xm/PushBG.h>
#include <Xm/BulletinB.h>

#include "dafs.h"
#include "misc/XArg.h"
#include "misc/color.h"

#include "KeyBoard.h"
#include "KeyMap.h"
#include "misc/Unicode.h"

static KeySymPair first_row[] = {
	{XK_Escape, XK_Escape},
	{XK_1, XK_exclam},
	{XK_2, XK_at},
	{XK_3, XK_numbersign},
	{XK_4, XK_dollar},
	{XK_5, XK_percent},
	{XK_6, XK_asciicircum},
	{XK_7, XK_ampersand},
	{XK_8, XK_asterisk}, 
	{XK_9, XK_parenleft},
	{XK_0, XK_parenright},
	{XK_minus, XK_underscore},
	{XK_equal, XK_plus},
	{XK_backslash, XK_bar},
	{XK_quoteleft, XK_asciitilde},
};

static KeySymPair second_row[] = {
	{XK_q, XK_Q},
	{XK_w, XK_W},
	{XK_e, XK_E},
	{XK_r, XK_R},
	{XK_t, XK_T},
	{XK_y, XK_Y}, 
	{XK_u, XK_U},
	{XK_i, XK_I},
	{XK_o, XK_O},
	{XK_p, XK_P},
	{XK_bracketleft, XK_braceleft},
	{XK_bracketright, XK_braceright},
};

static KeySymPair third_row[] = {
	{XK_a, XK_A},
	{XK_s, XK_S},
	{XK_d, XK_D},
	{XK_f, XK_F},
	{XK_g, XK_G},
	{XK_h, XK_H},
	{XK_j, XK_J},
	{XK_k, XK_K},
	{XK_l, XK_L},
	{XK_semicolon, XK_colon},
	{XK_quoteright, XK_quotedbl},
};

static KeySymPair fourth_row[] = {
	{XK_z, XK_Z},
	{XK_x, XK_X},
	{XK_c, XK_C},
	{XK_v, XK_V},
	{XK_b, XK_B},
	{XK_n, XK_N},
	{XK_m, XK_M},
	{XK_comma, XK_less},
	{XK_period, XK_greater},
	{XK_slash, XK_question},
};

static KeySymPair fifth_left_row [] = {
	
	{XK_Alt_L, XK_Alt_L},
//	{XK_Meta_L, XK_Meta_L},
};

/*
// Alt_R and Super_R are actually bogus
// x doesnt have keysyms
*/

static KeySymPair fifth_right_row[] =  {
//	{XK_Meta_R, XK_Meta_R},
	{XK_Alt_R, XK_Alt_R},           // These are bogus
	{XK_Super_R, XK_Super_R},       // ..
};


KeyBoard::KeyBoard(Widget parent, 
                   KeyMap *input_key_map,
                   KeyBoardHandler input_handler, 
                   void *input_user_data)
	: key_map(input_key_map),
	  shift_on(0),
	  control_on(0),
	  handler(input_handler),
	  user_data(input_user_data),
	  font_list(0)
{
	svmapInit(&this->button_map, 0, 1000);
	this->createKeyBoard(parent);
	this->setKeyMap(input_key_map);
}

KeyBoard::~KeyBoard()
{
	svmapCleanup(&this->button_map);
}

void 
KeyBoard::setKeyMap(KeyMap *input_key_map)
{
	this->key_map = input_key_map;
	this->setKeyNames();
	//this->setTitle(this->key_map->getName());
}


void
KeyBoard::toggleShift()
{
	Widget lshift_button = this->findButton(XK_Shift_L);
	Widget rshift_button = this->findButton(XK_Shift_R);

	XArg args;
	Pixel background;
	
	if(this->shift_on){
		args.reset();
		args.setArg(XmNbackground, &background);
		args.getValues(sub_bb);
	} else {
		background = getColorByName(sub_bb, "wheat");
	}
	this->shift_on = !this->shift_on;
	args.reset();
	args.setArg(XmNbackground, background);
	args.setValues(lshift_button);
	args.setValues(rshift_button);
    
	this->setKeyNames();
}

void
KeyBoard::toggleControl()
{
	Widget lcontrol_button = this->findButton(XK_Meta_L);
	Widget rcontrol_button = this->findButton(XK_Meta_R);

	XArg args;
	Pixel background;
	
	if(this->control_on){
		args.reset();
		args.setArg(XmNbackground, &background);
		args.getValues(sub_bb);
	} else {
		background = getColorByName(sub_bb, "wheat");
	}
	this->control_on = !this->control_on;
	args.reset();
	args.setArg(XmNbackground, background);
	args.setValues(lcontrol_button);
	args.setValues(rcontrol_button);
	
	this->setKeyNames();
}

void
KeyBoard::show()
{
	XtManageChild(this->bb);
	
	XtPopup(XtParent(this->bb), XtGrabNone);
	
	XMapRaised(XtDisplay(this->bb), XtWindow(XtParent(this->bb)));
}

void
KeyBoard::hide() 
{
	XtPopdown(XtParent(this->bb));
}


static aString
keyFromSym(KeySym sym)
{
	char buffer[20];
	aString key;
	
	sprintf(buffer, "%d", sym);
	key = buffer;
	return key;
}

Widget 
KeyBoard::findButton(KeySym keysym)
{
	aString key;
	key = keyFromSym(keysym);
	if(svmapContainsP(&this->button_map, key.string())){
		return (Widget) svmapGet(&this->button_map, key.string());
	}
	
	return 0;
}

void
KeyBoard::value(Widget w)
{
	KeySymPair *pair;
	XArg args;
	iUCode v;
	KeySym sym;
	
	args.setArg(XmNuserData, &pair);
	args.getValues(w);
	sym = this->shift_on ? pair->upper : pair->lower;
	if(this->control_on)
		v = this->key_map->lookupControlValue(sym);
	else
		v = this->key_map->lookupValue(sym);
	if(this->handler) 
		this->handler(this->user_data, sym, v);
}

static void
valueCB(Widget w, KeyBoard *kb, void *call)
{
	kb->value(w);
}

static void
toggleControlCB(Widget w, KeyBoard *kb, void *call)
{
	kb->toggleControl();
}

static void
toggleShiftCB(Widget w, KeyBoard *kb, void *call)
{
	kb->toggleShift();
}

void
KeyBoard::initialiseModKeys()
{
	this->shift_on = 0;
	this->control_on = 0;
	Widget lshift_button = this->findButton(XK_Shift_L);
	Widget rshift_button = this->findButton(XK_Shift_R);
	Widget rcontrol_button = this->findButton(XK_Meta_R);
	Widget lcontrol_button = this->findButton(XK_Meta_L);
	
	if(lshift_button)
		XtAddCallback(lshift_button, 
					  XmNactivateCallback,
					  (XtCallbackProc) toggleShiftCB,
					  this);
	if(rshift_button)
		XtAddCallback(rshift_button, 
					  XmNactivateCallback,
					  (XtCallbackProc) toggleShiftCB,
					  this);
	if(lcontrol_button)
		XtAddCallback(lcontrol_button, 
					  XmNactivateCallback,
					  (XtCallbackProc) toggleControlCB,
					  this);
	if(rcontrol_button)
		XtAddCallback(rcontrol_button, 
					  XmNactivateCallback,
					  (XtCallbackProc) toggleControlCB,
					  this);
}

void
KeyBoard::setButton(KeySym keysym, Widget button)
{
	aString key;
	
	key = keyFromSym(keysym);
	svmapSet(&this->button_map, key.string(), button);
}

void
KeyBoard::addKey(Widget parent,
				 int x,
				 int y,
				 int w,
				 int h,
				 KeySymPair* pair)
{
	XArg args;
	//char *string;
	aString a_string;
	uString u_string;
	
	XmString xm_string;
	Widget button;
	
	
	
	u_string = this->key_map->lookupName(pair->lower);
	xm_string=makeCompoundString(XtDisplay(parent),(iUCode *)u_string.string(),u_string.length(),&font_list);
	args.reset();
	x *= SCALE_FACTOR;
	y *= SCALE_FACTOR;
	w *= SCALE_FACTOR;
	h *= SCALE_FACTOR;
	x += 5;
	y += 5;
	w -= 10;
	h -= 10;
	args.setArg(XmNx,      x);
	args.setArg(XmNy,      y);
	args.setArg(XmNwidth,  w);
	args.setArg(XmNheight, h);
	args.setArg(XmNlabelString, xm_string);
	args.setArg(XmNrecomputeSize, False);
	args.setArg(XmNuserData, pair);
	if(font_list)
		args.setArg(XmNfontList, font_list);
	// bugs set this font
	a_string=u_string;

	KeySym ks = pair->lower;
	if(ks == XK_Shift_L || ks == XK_Shift_R || ks == XK_Meta_L || ks == XK_Meta_R)
	  button = XmCreatePushButton(parent, 
				      a_string.string(),
				      args.getArgs(),
				      args.getSize());
	else
	  button = XmCreatePushButtonGadget(parent,
					    a_string.string(),
					    args.getArgs(),
					    args.getSize());
	XmStringFree(xm_string);
	XtManageChild(button);
	XtAddCallback(button, 
				  XmNarmCallback,
				  (XtCallbackProc) valueCB,
				  this);
	//XtSetSensitive(button, FALSE);
	
	this->setButton(pair->lower, button);
	this->setButton(pair->upper, button);
}

void 
KeyBoard::createRow(Widget parent, 
					int x, int y, 
					KeySymPair *list, int n)
{
	int i;
	KeySymPair *p;
	
	for(i = 0; i < n; i++){
		p = list + i;
		this->addKey(parent, x, y, NORMAL_KEY_W, KEY_H,
					 p);
		x += NORMAL_KEY_W;
	}
}

KeySymPair delete_pair =     {XK_Delete, XK_Delete};
KeySymPair backspace_pair =  {XK_BackSpace, XK_BackSpace};
KeySymPair tab_pair = 	     {XK_Tab, XK_Tab};
KeySymPair control_pair =    {XK_Control_L, XK_Control_L};
KeySymPair lshift_pair =     {XK_Shift_L, XK_Shift_L};
KeySymPair rshift_pair =     {XK_Shift_R, XK_Shift_R};
KeySymPair space_pair =      {XK_space, XK_space};
KeySymPair return_pair =     {XK_Return, XK_Return};
KeySymPair caps_lock_pair =  {XK_Caps_Lock, XK_Caps_Lock};
KeySymPair lmeta_pair =      {XK_Meta_L, XK_Meta_L};
KeySymPair rmeta_pair =      {XK_Meta_R, XK_Meta_R};


void
KeyBoard::createKeyBoard(Widget parent)
{
	/*
// this describes a one variant of sparc keyboard
// only the main region of keys
// the qwerty region
// should be easily hackable to describe another keyboard
*/
	stringVoidPtrMap key_w;
	svmapInit(&key_w, 0, 1000);
	
	svmapSet(&key_w, "BackSpace", (void *) 15);
	svmapSet(&key_w, "Shift_L",   (void *) 23);
	svmapSet(&key_w, "Shift_R",   (void *) 27);
	svmapSet(&key_w, "Tab",       (void *) 15);
	svmapSet(&key_w, "Return",    (void *) 22);
	svmapSet(&key_w, "Control_L", (void *) 18);
	svmapSet(&key_w, "space",     (void *) 76);
	svmapSet(&key_w, "Caps_Lock",     (void *) 14);
	svmapSet(&key_w, "Meta_L",    (void *) 15);
	svmapSet(&key_w, "Meta_R",    (void *) 15);
	
	Dimension win_height = SCALE_FACTOR * 50;
	Dimension win_width  = SCALE_FACTOR * 150;
	
	Position x = 0;
	Position y = 0;
	
	// set the window so it can't be resized;
	
	XArg args;
	XmString title;
	
	
	args.reset();
	args.setArg(XmNmaxHeight, win_height);
	args.setArg(XmNminHeight, win_height);
	args.setArg(XmNmaxWidth, win_width);
	args.setArg(XmNminWidth, win_width);
	args.setArg(XmNmarginWidth, 0);
	args.setArg(XmNmarginHeight, 0);
	args.setArg(XmNautoUnmanage, FALSE);
	args.setArg(XmNresizePolicy, XmRESIZE_GROW);
	
	title = XmStringCreateLocalized("Keyboard");
	args.setArg(XmNdialogTitle, title);
	args.setArg(XmNnoResize, TRUE);
	
	this->bb = XmCreateBulletinBoardDialog(parent, 
										   "bb",
										   args.getArgs(),
										   args.getSize());
	
	XmStringFree(title);
	
	args.reset();
	args.setArg(XmNmaxHeight, win_height);
	args.setArg(XmNminHeight, win_height);
	args.setArg(XmNmaxWidth, win_width);
	args.setArg(XmNminWidth, win_width);
	args.setArg(XmNautoUnmanage, FALSE);

	
	this->sub_bb = XmCreateBulletinBoard(this->bb,
										 "sub_bb",
										 args.getArgs(),
										 args.getSize());
	
	int first_row_length = sizeof(first_row) / sizeof(KeySymPair);
	this->createRow(this->sub_bb, x, y, first_row, first_row_length);
	x += first_row_length * NORMAL_KEY_W;
	
	y += KEY_H;
	x = 0;
	
#if 1
	int second_row_length = sizeof(second_row) / sizeof(KeySymPair);
	
	this->addKey(this->sub_bb, x, y, (int) svmapGet(&key_w, "Tab"), KEY_H, 
				 &tab_pair);
	x +=  (int) svmapGet(&key_w, "Tab");
	this->createRow(this->sub_bb, x, y, second_row, second_row_length);
	x +=  second_row_length * NORMAL_KEY_W;
	
	this->addKey(this->sub_bb, x, y, (int) svmapGet(&key_w, "BackSpace"), KEY_H,
				 &backspace_pair);
	x +=  (int) svmapGet(&key_w, "BackSpace");
	
	
	y += KEY_H;
	x = 0;
	
	int third_row_length = sizeof(third_row) / sizeof(KeySymPair);
	
	this->addKey(this->sub_bb, x, y, (int) svmapGet(&key_w, "Control_L"), KEY_H,
				 &control_pair);
	x +=  (int) svmapGet(&key_w, "Control_L");
	this->createRow(this->sub_bb, x, y, third_row, third_row_length);
	x +=  third_row_length * NORMAL_KEY_W;
	this->addKey(this->sub_bb, x, y, (int) svmapGet(&key_w, "Return"), KEY_H, 
				 &return_pair);
	
	y += KEY_H;
	x = 0;
	
	int fourth_row_length = sizeof(fourth_row) / sizeof(KeySymPair);
	
	this->addKey(this->sub_bb, x, y, (int) svmapGet(&key_w, "Shift_L"), KEY_H, 
				 &lshift_pair);
	x +=  (int) svmapGet(&key_w, "Shift_L");
	this->createRow(this->sub_bb, x, y, fourth_row, fourth_row_length);
	x +=  fourth_row_length * NORMAL_KEY_W;
	
	this->addKey(this->sub_bb, x, y, (int) svmapGet(&key_w, "Shift_R"), KEY_H, 
				 &rshift_pair);
	x +=  (int) svmapGet(&key_w, "Shift_R");
	
	y += KEY_H;
	x = 0;
	
	this->addKey(this->sub_bb, x, y, (int) svmapGet(&key_w, "Caps_Lock"), KEY_H, 
				 &caps_lock_pair);
	x +=  (int) svmapGet(&key_w, "Caps_Lock");
	int fifth_left_row_length = sizeof(fifth_left_row) / sizeof(KeySymPair);
	this->createRow(this->sub_bb, x, y, fifth_left_row, fifth_left_row_length);
	x +=  fifth_left_row_length * NORMAL_KEY_W;
	this->addKey(this->sub_bb, x, y, (int) svmapGet(&key_w, "Meta_L"), KEY_H,
				 &lmeta_pair);
	x += (int) svmapGet(&key_w, "Meta_L");
	this->addKey(this->sub_bb, x, y, (int) svmapGet(&key_w, "space"), KEY_H, 
				 &space_pair);
	x +=  (int) svmapGet(&key_w, "space");
	this->addKey(this->sub_bb, x, y, (int) svmapGet(&key_w, "Meta_R"), KEY_H,
				 &rmeta_pair);
	x += (int) svmapGet(&key_w, "Meta_R");
	int fifth_right_row_length = sizeof(fifth_right_row) / sizeof(KeySymPair);
	this->createRow(this->sub_bb, x, y, fifth_right_row, fifth_right_row_length);
	x +=  fifth_right_row_length * NORMAL_KEY_W;
#endif
	
	//this->show();
	
	this->initialiseModKeys();
	svmapCleanup(&key_w);
}


void
KeyBoard::setKeyNames ()
{
        XtUnmanageChild(this->sub_bb);

	Ref r;
	Widget button;
	KeySymPair *pair;
	KeySym sym;
	XArg args;
	XArg args2;
	XmString xm_string;
	uString u_string, v_string;
	Dimension w, h;
	Position x, y;
	XmString blank_string;
	
	
	args.reset();
	args.setArg(XmNuserData, &pair);
	args.setArg(XmNx,      &x);
	args.setArg(XmNy,      &y);
	args.setArg(XmNwidth,  &w);
	args.setArg(XmNheight, &h);
	
	blank_string = XmStringCreateLocalized("");
	
	for(r = svmapFirst(&this->button_map); r; r = svmapNext(&this->button_map, r)){
		button = (Widget) svmapContents(&this->button_map, r);
		args.getValues(button);
		sym = this->shift_on ? pair->upper : pair->lower;
		if(this->control_on){
			u_string = this->key_map->lookupControlName(sym);
			v_string = this->key_map->lookupName(sym);
			// this causes meta and return and that lot to still show up
			if(v_string.length() > 1)
				u_string = v_string;
		} else {
			u_string = this->key_map->lookupName(sym);
			v_string = this->key_map->lookupControlName(sym);
		}
		
		if(u_string.length() == 0)
			xm_string = blank_string;
		else {
			if(unicodeDrawableP(XtDisplay(this->bb), 
								(iUCode *) u_string.string()))
				xm_string=makeCompoundString(XtDisplay(this->bb),
											 (iUCode *)u_string.string(),
											 u_string.length(), 
											 &font_list);
			else {
				// I cant render this key so stop it from being
				// inputtable
				if(this->control_on){
					this->key_map->disableControlTable(sym);
				} else {
					this->key_map->disableTable(sym);
				}

				xm_string=blank_string;
			}
		}
		args2.reset();
		args2.setArg(XmNlabelString, xm_string);
		args2.setArg(XmNfontList, font_list);
		args2.setValues(button);
		if(xm_string != blank_string)
			XmStringFree(xm_string);
	}
	
	XmStringFree(blank_string);
	
	if(!XtIsManaged(this->sub_bb))
	  XtManageChild(this->sub_bb);
}
