
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */
#include "rafport.h"
#include "str-lib.h"

#include <X11/keysym.h>

#include "KeyMap.h"

#define ESC 27

KeyMap *
initGreek()
{
    KeyMap *greek;
    KeyMap *keymap;
	
    greek = new KeyMap("Greek");
    keymap = greek;
    keymap->setTable(XK_backslash, uString('?'),  '?', uString('?'));
    keymap->setTable(XK_Delete, uString("Del"), '\b', uString());
	
    keymap->setTable(XK_Escape, uString(), '\0', uString());
    keymap->setTable(XK_1, uString('1'), '1', uString('1'));
    keymap->setTable(XK_2, uString('2'), '2', uString('2'));
    keymap->setTable(XK_3, uString('3'), '3', uString('3'));
    keymap->setTable(XK_4, uString('4'), '4', uString('4'));
    keymap->setTable(XK_5, uString('5'), '5', uString('5'));
    keymap->setTable(XK_6, uString('6'), '6', uString('6'));
    keymap->setTable(XK_7, uString('7'), '7', uString('7'));
    keymap->setTable(XK_8, uString('8'), '8', uString('8'));
    keymap->setTable(XK_9, uString('9'), '9', uString('9'));
    keymap->setTable(XK_0, uString('0'), '0', uString('0'));
    keymap->setTable(XK_minus, uString(0x00b7), 0x00b7, uString(0x00b7));
    keymap->setTable(XK_equal, uString(), 0, uString());
    keymap->setTable(XK_BackSpace, uString("BS"), '\b', uString());
	
    keymap->setTable(XK_Tab, uString("Tab"), '\t', uString('\t'));
    keymap->setTable(XK_q, uString(0x00a3), 0x00a3, uString(0x00a3));
    keymap->setTable(XK_w, uString(0x03c2), 0x03c2, uString(0x03c2));
    keymap->setTable(XK_e, uString(0x03b5), 0x03b5, uString(0x03b5));
    keymap->setTable(XK_r, uString(0x03c1), 0x03c1, uString(0x03c1));
    keymap->setTable(XK_t, uString(0x03c4), 0x03c4, uString(0x03c4));
    keymap->setTable(XK_y, uString(0x03c5), 0x03c5, uString(0x03c5));
    keymap->setTable(XK_u, uString(0x03b8), 0x03b8, uString(0x03b8));
    keymap->setTable(XK_i, uString(0x03b9), 0x03b9, uString(0x03b9));
    keymap->setTable(XK_o, uString(0x03bf), 0x03bf, uString(0x03bf));
    keymap->setTable(XK_p, uString(0x03c0), 0x03c0, uString(0x03c0));
    keymap->setTable(XK_bracketleft, uString(0x007e), 0x007e, uString(0x007e));
    keymap->setTable(XK_bracketright, uString(), 0, uString());
    keymap->setTable(XK_Return, uString("Ret"), '\n', uString('\n'));
	
    keymap->setTable(XK_Control_L, uString("Control"), '\0', uString());
    keymap->setTable(XK_a, uString(0x03b1), 0x03b1, uString(0x03b1));
    keymap->setTable(XK_s, uString(0x03c3), 0x03c3, uString(0x03c3));
    keymap->setTable(XK_d, uString(0x03b4), 0x03b4, uString(0x03b4));
    keymap->setTable(XK_f, uString(0x03c6), 0x03c6, uString(0x03c6));
    keymap->setTable(XK_g, uString(0x03b3), 0x03b3, uString(0x03b3));
    keymap->setTable(XK_h, uString(0x03b7), 0x03b7, uString(0x03b7));
    keymap->setTable(XK_j, uString(0x03be), 0x03be, uString(0x03be));
    keymap->setTable(XK_k, uString(0x03ba), 0x03ba, uString(0x03ba));
    keymap->setTable(XK_l, uString(0x03bb), 0x03bb, uString(0x03bb));
    keymap->setTable(XK_colon, uString('`'), '`', uString('`'));
    keymap->setTable(XK_quoteright, uString(0x2019), 0x2019, uString(0x2019));
    keymap->setTable(XK_quoteleft, uString('\\'), '\\', uString('\\'));
	
    keymap->setTable(XK_Shift_L, uString("Shift"), '\0', uString());
    keymap->setTable(XK_z, uString(0x03b6), 0x03b6, uString(0x03b6));
    keymap->setTable(XK_x, uString(0x03c7), 0x03c7, uString(0x03c7));
    keymap->setTable(XK_c, uString(0x03c8), 0x03c8, uString(0x03c8));
    keymap->setTable(XK_v, uString(0x03c9), 0x03c9, uString(0x03c9));
    keymap->setTable(XK_b, uString(0x03b2), 0x03b2, uString(0x03b2));
    keymap->setTable(XK_n, uString(0x03bd), 0x03bd, uString(0x03bd));
    keymap->setTable(XK_m, uString(0x03bc), 0x03bc, uString(0x03bc));
    keymap->setTable(XK_less, uString(';'), ';', uString(';'));
    keymap->setTable(XK_greater, uString(':'), ':', uString(':'));
    keymap->setTable(XK_question, uString('_'), '_', uString('_'));
    keymap->setTable(XK_Shift_R, uString("Shift"), '\0', uString());
    keymap->setTable(XK_Linefeed, uString("LF"), '\0', uString());
	
    keymap->setTable(XK_Caps_Lock, uString(), '\0', uString());
    keymap->setTable(XK_Alt_L, uString("Alt"), '\0', uString());
    keymap->setTable(XK_Meta_L, uString("Meta"), '\0', uString());
    keymap->setTable(XK_space, uString(), ' ', uString(' '));
    keymap->setTable(XK_Meta_R, uString("Meta"), '\0', uString());
    keymap->setTable(XK_Alt_R, uString(), '\0', uString());
    keymap->setTable(XK_Super_R, uString(), '\0', uString());
	
    keymap->setTable(XK_Control_R, uString(), '\0', uString("xo"));
	
    keymap->setTable(XK_bar, uString(0x0021), 0x0021, uString(0x0021));
	
    keymap->setTable(XK_exclam, uString(0x2022), 0x2022, uString(0x2022));
    keymap->setTable(XK_at, uString(0x002b), 0x002b, uString(0x002b));
//  keymap->setTable(XK_numbersign, uString('_'), '_', uString('_'));
    keymap->setTable(XK_dollar, uString(0x003d), 0x003d, uString(0x003d));
    keymap->setTable(XK_percent, uString(0x0028), '%', uString('%'));
    keymap->setTable(XK_asciicircum, uString(0x0029), 0x0029, uString(0x0029));
    keymap->setTable(XK_ampersand, uString(0x0025), 0x0025, uString(0x0025));
    keymap->setTable(XK_asterisk, uString(0x002f), 0x002f, uString(0x002f));
    keymap->setTable(XK_parenleft, uString(0x00ad), 0x00ad, uString(0x00ad));
    keymap->setTable(XK_parenright, uString(0x0023), 0x0023, uString(0x0023));
    keymap->setTable(XK_underscore, uString(0x002a), 0x002a, uString(0x002a));
    keymap->setTable(XK_plus, uString(0x005e), 0x005e, uString(0x005e));
	
    keymap->setTable(XK_Q, uString(0x0026), 0x0026, uString(0x0026));
    keymap->setTable(XK_W, uString(0x0024), 0x0024, uString(0x0024));
    keymap->setTable(XK_E, uString(0x0395), 0x0395, uString(0x0395));
    keymap->setTable(XK_R, uString(0x03a1), 0x03a1, uString(0x03a1));
    keymap->setTable(XK_T, uString(0x03a4), 0x03a4, uString(0x03a4));
    keymap->setTable(XK_Y, uString(0x03a5), 0x03a5, uString(0x03a5));
    keymap->setTable(XK_U, uString(0x0398), 0x0398, uString(0x0398));
    keymap->setTable(XK_I, uString(0x0399), 0x0399, uString(0x0399));
    keymap->setTable(XK_O, uString(0x039f), 0x039f, uString(0x039f));
    keymap->setTable(XK_P, uString(0x03a0), 0x03a0, uString(0x03a0));
    keymap->setTable(XK_braceleft, uString(0x2013), 0x2013, uString(0x2013));
    keymap->setTable(XK_braceright, uString(), 0, uString());
	
    keymap->setTable(XK_A, uString(0x0391), 0x0391, uString(0x0391));
    keymap->setTable(XK_S, uString(0x03a3), 0x03a3, uString(0x03a3));
    keymap->setTable(XK_D, uString(0x0394), 0x0394, uString(0x0394));
    keymap->setTable(XK_F, uString(0x03a6), 0x03a6, uString(0x03a6));
    keymap->setTable(XK_G, uString(0x0393), 0x0393, uString(0x0393));
    keymap->setTable(XK_H, uString(0x0397), 0x0397, uString(0x0397));
    keymap->setTable(XK_J, uString(0x039e), 0x039e, uString(0x039e));
    keymap->setTable(XK_K, uString(0x039a), 0x039a, uString(0x039a));
    keymap->setTable(XK_L, uString(0x039b), 0x039b, uString(0x039b));
    keymap->setTable(XK_quotedbl, uString(0x2018), 0x2018, uString(0x2018));
    keymap->setTable(XK_asciitilde, uString(), 0, uString());
	
    keymap->setTable(XK_Z, uString(0x0396), 0x0396, uString(0x0396));
    keymap->setTable(XK_X, uString(0x03a7), 0x03a7, uString(0x03a7));
    keymap->setTable(XK_C, uString(0x03a8), 0x03a8, uString(0x03a8));
    keymap->setTable(XK_V, uString(0x03a9), 0x03a9, uString(0x03a9));
    keymap->setTable(XK_B, uString(0x0392), 0x0392, uString(0x0392));
    keymap->setTable(XK_N, uString(0x039d), 0x039d, uString(0x039d));
    keymap->setTable(XK_M, uString(0x039c), 0x039c, uString(0x039c));
    keymap->setTable(XK_comma, uString(','), ',', uString(','));
    keymap->setTable(XK_period, uString('.'), '.', uString('.'));
    keymap->setTable(XK_slash, uString('-'), '-', uString('-'));
	
/*
		CONTROL CHARACTERS
*/

	keymap->setControlTable(XK_backslash, uString(0x2014), 0x2014, uString(0x2014));
	keymap->setControlTable(XK_bar, uString(0x2013), 0x2013, uString(0x2013));
	
	keymap->setControlTable(XK_e, uString(0x03ad), 0x03ad, uString(0x03ad));
	keymap->setControlTable(XK_t, uString(0x03cb), 0x03ad, uString(0x03cb));
	keymap->setControlTable(XK_y, uString(0x03cd), 0x03ad, uString(0x03cd));
	keymap->setControlTable(XK_u, uString(0x03ca), 0x03ca, uString(0x03ca));
	keymap->setControlTable(XK_i, uString(0x03af), 0x03af, uString(0x03af));
	keymap->setControlTable(XK_o, uString(0x03cc), 0x03cc, uString(0x03cc));
	keymap->setControlTable(XK_a, uString(0x03ac), 0x03ac, uString(0x03ac));
	keymap->setControlTable(XK_g, uString(0x03b0), 0x03b0, uString(0x03b0));
	keymap->setControlTable(XK_h, uString(0x03ae), 0x03ae, uString(0x03ae));
	keymap->setControlTable(XK_j, uString(0x0390), 0x0390, uString(0x0390));
	keymap->setControlTable(XK_z, uString('<'), '<', uString('<'));
	keymap->setControlTable(XK_x, uString('>'), '>', uString('>'));
	keymap->setControlTable(XK_c, uString('|'), '|', uString('|'));
	keymap->setControlTable(XK_v, uString(0x03ce), 0x03ce, uString(0x03ce));

	keymap->setControlTable(XK_9, uString(0x2121), 0x2121, uString(0x2121));
	keymap->setControlTable(XK_0, uString(0x00d7), 0x00d7, uString(0x00d7));
	keymap->setControlTable(XK_minus, uString(0x00f7), 0x00f7, uString(0x00f7));
	keymap->setControlTable(XK_equal, uString(0x00b1), 0x00b1, uString(0x00b1));

	keymap->setControlTable(XK_at, uString(0x00b2), 0x00b2, uString(0x00b2));
	keymap->setControlTable(XK_numbersign, uString(0x00b3), 0x00b3, uString(0x00b3));
	keymap->setControlTable(XK_ampersand, uString('{'), '{', uString('{'));
	keymap->setControlTable(XK_asterisk, uString('['), '[', uString('['));
	keymap->setControlTable(XK_parenleft, uString(']'), ']', uString(']'));
	keymap->setControlTable(XK_parenright, uString('}'), '}', uString('}'));
	keymap->setControlTable(XK_Q, uString('@'), '@', uString('@'));
	keymap->setControlTable(XK_E, uString(0x0388), 0x0388, uString(0x0388));
	keymap->setControlTable(XK_T, uString(0x03d4), 0x03d4, uString(0x03d4));
	keymap->setControlTable(XK_Y, uString(0x038e), 0x038e, uString(0x038e));
	keymap->setControlTable(XK_U, uString(0x03aa), 0x03aa, uString(0x03aa));
	keymap->setControlTable(XK_I, uString(0x038a), 0x038a, uString(0x038a));
	keymap->setControlTable(XK_O, uString(0x038c), 0x038c, uString(0x038c));
	keymap->setControlTable(XK_A, uString(0x0386), 0x0386, uString(0x0386));
	keymap->setControlTable(XK_H, uString(0x0389), 0x0389, uString(0x0389));
	keymap->setControlTable(XK_V, uString(0x038f), 0x038f, uString(0x038f));
	keymap->setControlTable(XK_M, uString(0x00b5), 0x00b5, uString(0x00b5));
	keymap->setControlTable(XK_less, uString(0x00ab),0x00ab, uString(0x00ab));	
	keymap->setControlTable(XK_greater, uString(0x00bb),0x00bb, uString(0x00bb));	
	keymap->setControlTable(XK_comma, uString(0x2039),0x2039, uString(0x2039));	
	keymap->setControlTable(XK_period, uString(0x203a),0x203a, uString(0x203a));	

	keymap->setControlTable(XK_1, uString(0x201a),0x201a, uString(0x201a));	
	keymap->setControlTable(XK_2, uString(0x201e),0x201e, uString(0x201e));	
	keymap->setControlTable(XK_3, uString(0x2018),0x2018, uString(0x2018));	
	keymap->setControlTable(XK_4, uString(0x2019),0x2019, uString(0x2019));	
	keymap->setControlTable(XK_5, uString(0x201c),0x201c, uString(0x201c));	
	keymap->setControlTable(XK_6, uString(0x201d),0x201d, uString(0x201d));	
    return keymap;
}
