
/*
 * Copyright (c) 1995 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"

#include "UniEditor.h"
#include "Line.h"
#include "misc/Unicode.h"
#include "rafutil.h"


void 
UniEditor::insertAtSelection(iUCode c, int *initial, int *final)
{
  *initial = 10000;
  *final = -1;
  Widget da = this->getWidget();

	if(this->getSelectionType() == stSELECTION) {
		// assume result is the text is deleted
		// and we switch to a cursor at the right point
	  this->setCursorMode(cmBEFORE);
	  this->deleteAtSelection(initial, final);
	}
	
	if(this->getSelectionType() == stNONE) {
		XBell(XtDisplay(da), 0);
		*initial = -1;
		*final = -1;
		return;
	}
	
	// assert selection_type == stCURSOR
	int line, col;
	cMode mode;
	this->getCursor(&line, &col, &mode);
	if(mode == cmAFTER)
	  col++;
	
	if(c == '\n') {
		lines.splitLine(line, lines.line[line]->base2text(col));
		lines.line[line]->calculateLength(XtDisplay(da));
		lines.line[line+1]->calculateLength(XtDisplay(da));

		this->setCursor(line + 1, 0);
		this->setCursorMode(cmBEFORE);
		*initial = MIN(*initial, line);
		*final   = MAX(this->lastLine(), *final);
	} 
	else {
		// add new character
		if(c) {
		        int raw = lines.line[line]->base2text(col);
			lines.line[line]->insert(c, raw);

			if(uniSpacing(c)) {
#if 1
			  uDirection old_dir, new_dir; 
			  new_dir = lines.line[line]->direction(col);			  
			  old_dir = lines.line[line]->direction(col-1);

			  if(new_dir != old_dir && mode == cmBEFORE) {
			    this->setCursorMode(cmAFTER);
			    mode = cmAFTER;
			  }

			  if(mode == cmBEFORE)
			    this->setCursor(line, col+1);
			  else
			    this->setCursor(line, col);
#else
			  if(lines.line[line]->direction(col) == dRIGHT)
			    this->moveCursor(dLEFT);
			  else
			    this->moveCursor(dRIGHT);
#endif
			  lines.line[line]->calculateLength(XtDisplay(da));
			}
			*initial = MIN(line, *initial);
			*final = MAX(line, *final);
		}
		else
			XBell(XtDisplay(da), 0);
	}
}

void
UniEditor::blammoSelection(void)
{
	int left_line, left_col;
	int right_line, right_col;
	int i;
	unsigned start, stop;
	Widget da = this->getWidget();
	
	this->getSelection(&left_line, &left_col, &right_line, &right_col);
	/* this prevents from having some selected entities deleted in illum*/
	this->clearSelection();

	if(left_line == right_line) {
		start = lines.line[left_line]->base2text(left_col);
		stop  = lines.line[left_line]->base2text(right_col);
		stop += lines.line[left_line]->nonspacing(right_col);
		lines.line[left_line]->remove(start, stop - start);
	} else {
		start = lines.line[left_line]->base2text(left_col);
		stop = lines.line[left_line]->textLength();
		lines.line[left_line]->remove(start, stop - start);
		start = 0;
		stop  = lines.line[right_line]->base2text(right_col);
		stop += lines.line[right_line]->nonspacing(right_col);
		lines.line[right_line]->remove(start, stop - start);
		for(i = left_line + 1; i < right_line; i++)
			lines.deleteLine(left_line+1);
		*lines.line[left_line] += *lines.line[left_line+1];
	}
	this->setCursor(left_line, left_col);
	lines.line[left_line]->calculateLength(XtDisplay(da));
}

void 
UniEditor::backspaceAtSelection(int *inital, int *final)
{
  Widget da = this->getWidget();

	switch (this->getSelectionType()) {
	case stSELECTION:
		this->blammoSelection();
		break;
	case stNONE:
		XBell(XtDisplay(da), 0);
		break;
	case stCURSOR: {
		unsigned start, stop;
		start = stop = 0;
		int line, col;
		cMode mode;
		this->getCursor(&line, &col, &mode);
		// avoid problems in illum with unselecting deleted entities
		this->clearSelection();
		this->selection_type = stCURSOR;
		if(mode == cmAFTER)
		  col++;		

		if(col > 0){
			// not backing up over newline, so delete base
			// character before and its diacriticals
			start = lines.line[line]->base2text(col - 1);
			stop  = lines.line[line]->base2text(col);
			stop += lines.line[line]->nonspacing(stop);

			// bogus -- when is this not true?
			if(stop != start)
				lines.line[line]->remove(start, stop - start);
			// set cursor back 1 base
			if(mode == cmAFTER)
			  if(col > 1)
			    col--;
			  else
			    this->setCursorMode(cmBEFORE);
			this->setCursor(line, col - 1);
			lines.line[line]->calculateLength(XtDisplay(da));
			*inital = *final = line;
		}
		else if(line > 0) {
			// backing up overnewline and not at beginning
			// of document
			// add text of this line to previous

			// remember col that you joined at
			col = this->lastCol(line-1);
			*lines.line[line - 1] += *lines.line[line];
			lines.line[line-1]->calculateLength(XtDisplay(da));
			// go to spot of join
			this->setCursor(line-1,col);
			*inital = line-1;
			*final  = this->lastLine();
		}
	}
	break;
	default:
		halt("unknown selection type.\n");
	} 
}


void 
UniEditor::deleteAtSelection(int *initial, int *final)
{
  Widget da = this->getWidget();

  int fl; int fc; int ll; int lc;
	switch (this->getSelectionType()) {
	case stSELECTION:
		this->blammoSelection();
		this->getCursor(&fl, &fc);
		ll = this->lastLine();
		*initial = fl;
		*final = ll;
		break;
	case stNONE:
		XBell(XtDisplay(da), 0);
		*initial = -1;
		*final = -1;
		break;
	case stCURSOR: {
		unsigned start, stop;
		int line, col;
		cMode mode;
		this->getCursor(&line, &col, &mode);
		this->clearSelection();
		if(mode == cmAFTER)
		  col++;

		start = stop = 0;

		if(col < this->lastCol(line)) {
			// not running over newline
			// delete this guy and keep cursor in sampe position
			start = lines.line[line]->base2text(col);
			stop  = lines.line[line]->base2text(col + 1);
			stop += lines.line[line]->nonspacing(stop);
			// bogus when is this not true
			if(stop != start)
				lines.line[line]->remove(start, stop - start);
			*initial = *final = line;
			lines.line[line]->calculateLength(XtDisplay(da));
		}
		else if(line < this->lastLine()){ // not at end of doc
			// grab the text on next line and kill that line
			*lines.line[line] += *lines.line[line + 1];
			lines.line[line]->calculateLength(XtDisplay(da));
			*initial = line;
			*final  = this->lastLine();
		}
		if(mode == cmAFTER) 
		  col--;
		this->setCursor(line, col);
	}
	break;
	default:
		halt("unknown selection type.\n");
	}
}

