
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */
/* Written by Dan Heller.  Copyright 1991, O'Reilly && Associates.
 * This program is freely distributable without licensing fees and
 * is provided without guarantee or warrantee expressed or implied.
 * This program is -not- in the public domain.
 */

/* action_area.c -- demonstrate how CreateActionArea() can be used
 * in a real application.  Create what would otherwise be identified
 * as a PromptDialog, only this is of our own creation.  As such,
 * we provide a TextField widget for input.  When the user presses
 * Return, the Ok button is activated.
 */

#include <Xm/PushBG.h>
#include <Xm/PushB.h>
#include <Xm/LabelG.h>
#include <Xm/PanedW.h>
#include <Xm/Form.h>
#include <Xm/RowColumn.h>
#include <Xm/TextF.h>

#include "misc.h"

#define TIGHTNESS 20

Widget
CreateActionArea(
	Widget parent,
	ActionAreaItem *actions,
	int num_actions)
{
    Widget action_area, widget;
    int i;
	
    action_area = XtVaCreateWidget("action_area", xmFormWidgetClass, parent,
								   XmNfractionBase, TIGHTNESS*num_actions - 1,
								   XmNleftOffset,   10,
								   XmNrightOffset,  10,
								   NULL);
	
    for (i = 0; i < num_actions; i++) {
        widget = XtVaCreateManagedWidget(actions[i].label,
										 xmPushButtonWidgetClass, action_area,
										 XmNleftAttachment,       i? XmATTACH_POSITION : XmATTACH_FORM,
										 XmNleftPosition,         TIGHTNESS*i,
										 XmNtopAttachment,        XmATTACH_FORM,
										 XmNbottomAttachment,     XmATTACH_FORM,
										 XmNrightAttachment,
										 i != num_actions-1? XmATTACH_POSITION : XmATTACH_FORM,
										 XmNrightPosition,        TIGHTNESS*i + (TIGHTNESS-1),
										 XmNshowAsDefault,        i == 0,
										 XmNdefaultButtonShadowThickness, 1,
										 NULL);
        if (actions[i].callback)
            XtAddCallback(widget, XmNactivateCallback,
						  actions[i].callback, actions[i].data);
        if (i == 0) {
            /* Set the action_area's default button to the first widget
             * created (or, make the index a parameter to the function
             * or have it be part of the data structure). Also, set the
             * pane window constraint for max and min heights so this
             * particular pane in the PanedWindow is not resizable.
             */
            Dimension height, h;
            XtVaGetValues(action_area, XmNmarginHeight, &h, NULL);
            XtVaGetValues(widget, XmNheight, &height, NULL);
            height += 2 * h;
            XtVaSetValues(action_area,
						  XmNdefaultButton, widget,
						  XmNpaneMaximum,   height,
						  XmNpaneMinimum,   height,
						  NULL);
        }
    }
	
    XtManageChild(action_area);
	
    return action_area;
}
