
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"

#include "dafs.h"
#include "oscroll.h"
#include "misc/XArg.h"

#ifndef MAX
#define MAX(x,y) (((x) > (y)) ? (x) : (y))
#endif
#ifndef MIN
#define MIN(x,y) (((x) < (y)) ? (x) : (y))
#endif

/*
// This has functions for figuring out and setting the visible
// area of a scrolled window
// these are not illum specific and should be move to
// misc
*/

void
scrollToPositionAuto(Widget sw, Position x, Position y)
{
	// bogus double check identity of scrolled window
	XArg args;
	Widget work_window;
	Widget hsb;
	Widget vsb;
	Dimension height;
	Dimension width;
	Dimension clip_height;
	Dimension clip_width;
	int minimum;
	int value;
	int maximum;
	int slider_size;
	int page_increment;
	int increment;;
	
	args.reset();
	args.setArg(XmNworkWindow, &work_window);
	args.setArg(XmNhorizontalScrollBar, &hsb);
	args.setArg(XmNverticalScrollBar, &vsb);
	args.getValues(sw);
	
	if(work_window) {
		args.reset();
		args.setArg(XmNheight, &clip_height);
		args.setArg(XmNwidth,  &clip_width);
		args.getValues(XtParent(work_window));
		
		
		args.reset();
		args.setArg(XmNheight, &height);
		args.setArg(XmNwidth,  &width);
		args.getValues(work_window);
		
		// bogus check if hsb is managed
		args.reset();
		args.setArg(XmNminimum, &minimum);
		args.setArg(XmNmaximum, &maximum);
		args.setArg(XmNvalue, &value);
		args.setArg(XmNsliderSize, &slider_size);
		args.setArg(XmNincrement, &increment);
		args.setArg(XmNpageIncrement, &page_increment);
		args.getValues(hsb);
		
		Position x_prime;
		x_prime = MAX(0, x - clip_width/2);
		if(clip_width < width)
			value = (x_prime * (maximum - slider_size - minimum)) / (width - clip_width);
		else
			value = 0;
		
		value = MAX(value, minimum);
		value = MIN(value, maximum - slider_size);
		XmScrollBarSetValues(hsb, value, slider_size, increment, page_increment, TRUE);
		
		// bogus check if vsb is managed
		args.reset();
		args.setArg(XmNminimum, &minimum);
		args.setArg(XmNmaximum, &maximum);
		args.setArg(XmNvalue, &value);
		args.setArg(XmNsliderSize, &slider_size);
		args.setArg(XmNincrement, &increment);
		args.setArg(XmNpageIncrement, &page_increment);
		args.getValues(vsb);
		
		Position y_prime;
		y_prime = MAX(0, y - clip_height/2);
		if(clip_height < height)
			value = (y_prime * (maximum - slider_size - minimum)) / (height - clip_height);
		else 
			value = 0;
		
		value = MAX(value, minimum);
		value = MIN(value, maximum - slider_size);
		XmScrollBarSetValues(vsb, value, slider_size, increment, page_increment, TRUE);
	}
}


void
getScrollPositionCenterAuto(Widget sw, Position *x_ptr, Position *y_ptr)
{
	// bogus double check identity of scrolled window
	XArg args;
	Widget work_window;
	Widget hsb;
	Widget vsb;
	Dimension height;
	Dimension width;
	Dimension clip_height;
	Dimension clip_width;
	int minimum;
	int value;
	int maximum;
	int slider_size;
	Position x, y;
	
	args.reset();
	args.setArg(XmNworkWindow, &work_window);
	args.setArg(XmNhorizontalScrollBar, &hsb);
	args.setArg(XmNverticalScrollBar, &vsb);
	args.getValues(sw);
	
	if(work_window) {
		args.reset();
		args.setArg(XmNheight, &clip_height);
		args.setArg(XmNwidth,  &clip_width);
		args.getValues(XtParent(work_window));
		
		
		args.reset();
		args.setArg(XmNheight, &height);
		args.setArg(XmNwidth,  &width);
		args.getValues(work_window);
		
		if (XtIsManaged(hsb)) {
			args.reset();
			args.setArg(XmNminimum, &minimum);
			args.setArg(XmNmaximum, &maximum);
			args.setArg(XmNvalue, &value);
			args.setArg(XmNsliderSize, &slider_size);
			args.getValues(hsb);
			
			if(slider_size == maximum - minimum)
				x = clip_width/2;
			else {
				Position x_prime;
				x_prime = (value * (width - clip_width)) / (maximum - slider_size - minimum);
				x = x_prime + clip_width / 2;
				x = MIN(width, x);
			}
		} else 
			x = width / 2;
		
		if (XtIsManaged(vsb)) {
			args.reset();
			args.setArg(XmNminimum, &minimum);
			args.setArg(XmNmaximum, &maximum);
			args.setArg(XmNvalue, &value);
			args.setArg(XmNsliderSize, &slider_size);
			args.getValues(vsb);
			
			
			if(slider_size == maximum - minimum)
				y = clip_height/2;
			else {
				Position y_prime;
				y_prime = (value * (height - clip_height)) / (maximum - slider_size - minimum);
				y = y_prime + clip_height / 2;
				y = MIN((int) height, (int) y);
			}
		} else 
			y = height / 2;
	} else {
		x = 0; y = 0;
	}
	
	*x_ptr = x;
	*y_ptr = y;
}

iBox 
visibleRectangleAuto(Widget sw)
{
	int wi, he;
	iBox window_region;
	Widget ww;
	XArg args;

	args.reset();
	args.setArg(XmNclipWindow, &ww);
	args.getValues(sw);
	
	Position s_x;
	Position s_y;
	getWidgetSizeAuto(ww, &wi, &he);
	getScrollPositionCenterAuto(sw, &s_x, &s_y);
	
	window_region.x = s_x - wi/2;
	window_region.y = s_y - he/2;
	window_region.x = MAX(window_region.x, 0);
	window_region.y = MAX(window_region.y, 0);
	window_region.w = wi;
	window_region.h = he;
	
	return window_region;
}

/*
// I wrote a new version that shares more code 
*/

int
visiblePositionPAuto(Widget sw, Position input_x, Position input_y)
{
	XArg args;
	Widget work_window;
	
	args.reset();
	args.setArg(XmNworkWindow, &work_window);
	args.getValues(sw);
	
	iBox vr = visibleRectangleAuto(sw);
	
	Position ulx = vr.x;
	Position uly = vr.y;
	Position lrx = vr.x + vr.w;
	Position lry = vr.y + vr.h;
	
	return ( 
		input_x >= ulx && 
		input_x <= lrx && 
		input_y >= uly && 
		input_y <= lry);
}



void
getWidgetSizeAuto(Widget wt, int *w, int *h)
{
	Dimension width;
	Dimension height;
	XArg args;
	
	args.setArg(XmNwidth, &width);
	args.setArg(XmNheight, &height);
	XtGetValues(wt, args.getArgs(), args.getSize());
	
	*w = width;   
	*h = height;
}

void
getScrollPositionULAuto(Widget sw, Position *tlx, Position *tly)
{
	int wi, he;
	iBox window_region;
	Widget ww;
	XArg args;

	args.reset();
	args.setArg(XmNclipWindow, &ww);
	args.getValues(sw);
	
	Position s_x;
	Position s_y;
	getWidgetSizeAuto(ww, &wi, &he);
	getScrollPositionCenterAuto(sw, &s_x, &s_y);
	
	*tlx = (Position)MAX(s_x - wi/2,0);
	*tly = (Position)MAX(s_y - he/2,0);
	
	return;
}	
