/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"
#include "rafutil.h"
#include "ucs.h"
#include "c-lib.h"
#include "smemory.h"
#include "str-lib.h"

#include "dafstype.h"
#include "dafs.h"
#include "utf.h"

Dll iDAFSError APIENTRY i_Collapse(iEntityPtr en,iUCode *nonrec);

Dll iDAFSError APIENTRY i_DeleteSelection(iEntityPtr *bp, int4 *beg, iEntityPtr *ep, int4 *end);

Dll iDAFSError APIENTRY i_InsertChar(iEntityPtr *bp, int4 *beg, 
									 iEntityPtr *ep, int4 *end, 
									 iUCode key);

Dll iDAFSError APIENTRY
i_Edit(iEntityPtr *bp, int4 *beg, 
	   iEntityPtr *ep, int4 *end, 
	   iUCode key);

Dll void APIENTRY i_Object2Range(iEntityPtr e,iEntityPtr *bp,int4 *beg,iEntityPtr *ep,int4 *end);

Dll iEntityPtr APIENTRY i_Range2Object(iEntityPtr bp,int4 beg,iEntityPtr ep,int4 end);

Dll iDAFSError APIENTRY i_StartObjEdit(iEntityPtr ent,iEditRecordPtr *r);

Dll iDAFSError APIENTRY i_EndObjEdit(iEntityPtr ent,iEditRecordPtr r,int4 *deleteme);

Dll iDAFSError APIENTRY i_ObjEdit(iEntityPtr ent,iEditRecordPtr r,iUCode key,int4 *deleteme);

Dll iDAFSError APIENTRY i_Text2Dafs(char *text, iEntityPtr *dp);

Dll iDAFSError APIENTRY i_UCode2Dafs(iUCode *text,iEntityPtr *dp, int4 swap);

Dll iDAFSError APIENTRY i_AddChars(iEntityPtr entity, int2 position, const char *text);

Dll iDAFSError APIENTRY i_AddChar(iEntityPtr ent, int2 position, char letter);

Dll iDAFSError APIENTRY i_DelChar(iEntityPtr entity, int2 position);

Dll iDAFSError APIENTRY i_AddUnis(iEntityPtr entity, int2 position, const iUCode *text);

Dll iDAFSError APIENTRY i_AddUni(iEntityPtr ent, int2 position, iUCode letter);

Dll iDAFSError APIENTRY i_DelUni(iEntityPtr entity, int2 position);

static int4 
isParent(iEntityPtr par, iEntityPtr ch)
{
	iEntityPtr cur;
	
	for(cur=i_GetParentEntity(ch);cur;cur=i_GetParentEntity(cur))
		if (par == cur)
			return 1;
	return 0;
}

static void 
deleteChildren(iEntityPtr en)
{
	iEntityPtr cur;
	
	for(cur=i_GetChildEntity(en);cur;cur=i_GetChildEntity(en)){
		i_DisposeEntity(cur);
	}
}

Dll iDAFSError APIENTRY
i_Collapse(iEntityPtr en,iUCode *nonrec)
{
	iEntityPtr cur;
	int4 num,i;
	const char *s;
	iDAFSError e=DAFSOK;
	iCallBackProc save=0;
	iUCode nr,c;
	iCallBackProc gCallBack;
	void *gUserData;

	i_GetCallBack(&gCallBack, &gUserData);
	if(gCallBack){
		MatrCallBack(mWillUpdate,gUserData,en);
		save=gCallBack;
		gCallBack=0;
	}
	nr=(nonrec) ? *nonrec : '^';
	cur=i_FindEntityForward(en,0,DATAEN,0);
	while(cur && isParent(en,cur)){
		switch(i_GetEDataType(cur)){
		case iDPosSet:
		case iD1PosSet:
			if (e=i_AddUni(en,-1,(iUCode) ((c=i_GetGlyph(en,0)) ? c : nr)))
				return e;
			break;
		case iDShort:
			num=i_GetShort(cur,0);
			for(i=0;i<num;i++) {
				if(i_GetEntityType(cur)==iSpaceT)
					e=i_AddUni(en,-1,' ');
				else if(i_GetEntityType(cur)==iNewLineT)
					e=i_AddUni(en,-1,'\n');
				if (e)
					return e;
			}
			break;
		case iDString:
			{
				const char *p;
				for(s = p = i_GetText(cur); *p; p++)
					e=i_AddUni(en,-1,(short)*p);
			}
			if (e)
				return e;
			break;
		case iDUString:
			e=i_AddUnis(en,-1,(iUCode*)i_GetUText(cur));
			if (e)
				return e;
			break;
		case iDNo:
			break;
		default :
			return UnknownDAFSError;
			break;
		}
		cur=i_FindEntityForward(cur,0,DATAEN,0);
	}	
	if (save) {
		gCallBack=save;
		MatrCallBack(mUpdate,gUserData,en);
	}
	deleteChildren(en);
	return DAFSOK;
}

static iDAFSError 
MergeEntity(iEntityPtr l,iEntityPtr r,
			iEntityPtr *bp,int4 *beg,iEntityPtr *ep,int4 *end)
{
	int4 t,len;
	iDAFSError e;
	iEntityPtr next,cur,temp;
	const char *text;
	const iUCode *utext;
	iBox bl,br, box;
	
	t = i_GetEDataType(l);
	switch (t) {
	case iDString:
		text = i_GetText(l);
		if (r == *ep)
			*end += strlen(text);
		*beg = strlen(text);
		*bp = r;
		if (e = i_AddChars(r,0,(char*)text))
			return e;
		bl = i_GetEnclosingBox(l);
		br = i_GetEnclosingBox(r);
		if (bl.h && br.h){
			i_UnionBox(bl, br, &box);
			if (e=i_SetBox(r,box))
				return e;
		}
		i_DisposeEntity(l);
		break;
	case iDUString:
		utext = i_GetUText(l);
		if (r == *ep)
			*end += ucslen(utext);
		*beg = ucslen(utext);
		*bp = r;
		if (e = i_AddUnis(r,0,(iUCode*)utext))
			return e;
		bl = i_GetEnclosingBox(l);
		br = i_GetEnclosingBox(r);
		if (bl.h && br.h){
			i_UnionBox(bl, br, &box);
			if (e=i_SetBox(r,box))
				return e;
		}
		i_DisposeEntity(l);
		break;
	case iDShort:
		len = i_GetShort(l,0);
		if (r == *ep)
			*end += len;
		*beg = len;
		*bp = r;
		i_SetShort(r,len + i_GetShort(r,0), 0);
		bl = i_GetEnclosingBox(l);
		br = i_GetEnclosingBox(r);
		if (bl.h && br.h){
			i_UnionBox(bl, br, &box);
			if (e=i_SetBox(r,box))
				return e;
		}
		i_DisposeEntity(l);
		break;
	case iDNo:
		if (!i_GetChildEntity(r) || !i_GetChildEntity(l))
			break;
		for (cur = i_GetChildEntity(r);cur;cur = next) {
			next = i_GetNextEntity(cur);
			temp = i_GetLastEntity(l);
			bl = i_GetEnclosingBox(l);
			br = i_GetEnclosingBox(cur);
			if (bl.h && br.h){
				i_UnionBox(bl, br, &box);
				if (e=i_SetBox(l,box))
					return e;
			}
			i_MoveToBack(cur,l);

			if ((i_GetEntityType(temp) == i_GetEntityType(cur)) &&
				(i_GetEDataType(temp) == i_GetEDataType(cur)) &&
				(e = MergeEntity(temp,cur,bp,beg,ep,end)))
				return e;
		}
		bl = i_GetEnclosingBox(l);
		br = i_GetEnclosingBox(r);
		if (bl.h && br.h){
				i_UnionBox(bl, br, &box);
				if (e=i_SetBox(l,box))
					return e;
		}
		i_DisposeEntity(r);
		break;
	case iDPosSet:
	case iD1PosSet:
		break;
	}
	return DAFSOK;
}

static iDAFSError 
killachar(iEntityPtr *bp,int4 *beg,
		  iEntityPtr *ep,int4 *end,
		  int4 merge)
{
	int4 et,len;
	iEntityPtr par,np,prev,next;
	iDAFSError e;
	
	et = i_GetEDataType(*bp);
	if (et == iDString) {
		len = strlen(i_GetText(*bp));
		if (len > 1) {
			len--;
			i_DelChar(*bp,*beg);
			if (*bp == *ep)
				(*end)--;
			else if (len == *beg) {
				*beg = 0;
				*bp = i_FindEntityForward(*bp,0,DATAEN,0);
			}
			return DAFSOK;
		}
	} else if (et == iDUString) {
		len = ucslen(i_GetUText(*bp));
		if (len > 1) {
			len--;
			i_DelUni(*bp,*beg);
			if (*bp == *ep)
				(*end)--;
			else if (len == *beg) {
				*beg = 0;
				*bp = i_FindEntityForward(*bp,0,DATAEN,0);
			}
			return DAFSOK;
		}
	} else if (et == iDShort) {
		len = i_GetShort(*bp,0);
		if (len > 1) {
			i_SetShort(*bp,--len,0);
			if (*bp == *ep)
				(*end)--;
			else if (len == *beg) {
				*beg = 0;
				*bp = i_FindEntityForward(*bp,0,DATAEN,0);
			}
			return DAFSOK;
		}
	}

    /* it is OK to set *bp to an */
    /* entity with no data because kill a char will be */
    /* called until *bp == *ep */
	np = i_FindEntityForward(*bp,0,DATAEN,0); 
	for (;;) {				
		next = i_GetNextEntity(*bp);	
		prev = i_GetPrevEntity(*bp);
		if (next || prev) break;
		par = i_GetParentEntity(*bp);
		i_DisposeEntity(*bp);
		*bp = par;
	}
	i_DisposeEntity(*bp);
	*bp = np;
	*beg = 0;
	if (merge && prev && next &&
			i_GetEDataType(next) == i_GetEDataType(prev) &&
#if 1
		    i_GetEDataType(next) != iDNo &&
#endif
			(e = MergeEntity(prev,next,bp,beg,ep,end)))
		return e;
	return DAFSOK;
}

static iBox 
fixBounds(iEntityPtr en)
{
	iEntityPtr cur;
	iBox nb,b,tb;

	b = i_GetEnclosingBox(en);
	if (!(cur=i_GetChildEntity(en))) return b;
	nb = fixBounds(cur);
	cur = i_GetNextEntity(cur);
	for (;cur;cur = i_GetNextEntity(cur)) {
		tb = fixBounds(cur);
		i_UnionBox(tb,nb,&nb);
	}
	return nb;
}

static int4 
getPrntBox(iEntityPtr en,iBox *bp)
{
	en = i_GetParentEntity(en);
	if (!en) return 0;
	*bp = i_GetEnclosingBox(en);
	return 1;
}

static iDAFSError 
guessBox(iEntityPtr new_entity)
{
	iEntityPtr prev,next;
	iBox b,nb,pb,prb;
	short nlp,nln;
	iTypeIndex nt,t;
	iDAFSError e;
	
	nt = i_GetEntityType(new_entity);
	nlp = 0;
	for (prev = i_FindEntityBack(new_entity,0,DATAEN,0);prev;prev = i_FindEntityBack(prev,0,DATAEN,0)) {
		t = i_GetEntityType(prev);
		if (t == iNewLineT) nlp += (short)i_GetShort(prev,0);
		else break;
	}
	nln = 0;
	for (next = i_FindEntityForward(new_entity,0,DATAEN,0);next;next = i_FindEntityBack(next,0,DATAEN,0)) {
		t = i_GetEntityType(next);
		if (t == iNewLineT) nln += (short)i_GetShort(next,0);
		else break;
	}
	if (next) nb = i_GetEnclosingBox(next);
	else nb.w = nb.h = 0;
	if (prev) pb = i_GetEnclosingBox(prev);
	else pb.w = pb.h = 0;
	if (nb.w && nb.h && pb.w && pb.h) {
		if (!nln && !nlp) {
			b.y = MIN(pb.y,nb.y);
			b.h = MAX(pb.y+pb.h,nb.y+nb.h) - b.y;
			b.x = pb.x + pb.w;
			b.w = nb.x - b.x;
		} else if (!nln) {
			b = nb;
			if (getPrntBox(new_entity,&prb))
				b.x = prb.x;
			else
				b.x = nb.x - (b.h * 2) / 3;
			b.w = nb.x - b.x;
		} else if (!nlp) {
			b = pb;
			b.x = pb.x + pb.w;
			if (getPrntBox(new_entity,&prb))
				b.w = (prb.x + prb.w) - b.x;
			else
				b.w = (b.h * 2) / 3;
		} else {
			if (!getPrntBox(new_entity,&b))
				b = pb;
			b.h = (nb.y - pb.y) / (nln + nlp);
			b.y += b.h * nlp;
		}
	} else if (pb.w && pb.h) {
		if (nlp) {
			if (!getPrntBox(new_entity,&b))
				b = pb;
			else b.h = pb.h;
			b.y += (b.h + (b.h / 8)) * nlp;
		} else {
			b = pb;
			b.x = pb.x + pb.w;
			if (getPrntBox(new_entity,&prb))
				b.w = (prb.x + prb.w) - b.x;
			else
				b.w = (b.h * 2) / 3;
		}
	} else if (nb.w && nb.h) {
		if (nln) {
			if (!getPrntBox(new_entity,&b))
				b = nb;
			else b.h = nb.h;
			b.y -= (b.h + (b.h / 8)) * nln;
		} else {
			b = nb;
			if (getPrntBox(new_entity,&prb))
				b.x = prb.x;
			else
				b.x = nb.x - (b.h * 2) / 3;
			b.w = nb.x - b.x;
		}
	}
	if (b.w <= 0) b.w = 1;
	if (b.h <= 0) b.h = 1;
	if (e=i_SetBox(new_entity,b))
		return e;
	return DAFSOK;
}

static iDAFSError 
splitEntity(iEntityPtr bp,iEntityPtr p,iTypeIndex type,iEntityPtr *newp)
{
	iEntityPtr next,par,new,lastnew;
	iDAFSError e;
	
	while (!i_GetPrevEntity(bp) && bp != p)
		bp = i_GetParentEntity(bp);
	if (i_GetParentEntity(bp) == i_GetParentEntity(p)) {
		new = i_NewEntityBefore(type,bp);
		if (!new) return MallocFailError;
		if (newp) *newp = new;
		return DAFSOK;
	}
	lastnew = 0;
	for (par = bp;par != p;) {
		if (bp) par = i_GetParentEntity(bp);
		else par = i_GetParentEntity(par);
		new = i_NewEntityAfter(i_GetEntityType(par),p);
		if (!new) return MallocFailError;
		if (lastnew)
		    i_MoveToBack(lastnew,new);
		for (;bp;bp = next) {
			next = i_GetNextEntity(bp);
			i_MoveToBack(bp,new);

		}
		i_SetBox(new,fixBounds(new));
		i_SetBox(par,fixBounds(par));
		i_DisposeProp(par,"Flag");
		bp = i_GetNextEntity(par);
		lastnew = new;
	}
	new = i_NewEntityAfter(type,p);
	if (!new) return MallocFailError;
	i_MoveAfter(lastnew,new);
	if (e=guessBox(new))
		return e;
	if (newp) *newp = new;
	return DAFSOK;
}

static iDAFSError 
deleteSelection(iEntityPtr *bp, int4 *beg, iEntityPtr *ep, int4 *end,
	   int4 merge) 
{
	iDAFSError e;
	
	while (*bp != *ep || *beg != *end)
		if (e = killachar(bp,beg,ep,end,merge))
			return e;
	return DAFSOK;
}

Dll iDAFSError APIENTRY
i_DeleteSelection(iEntityPtr *bp, int4 *beg, iEntityPtr *ep, int4 *end) 
{
	return deleteSelection(bp,beg,ep,end,1);
}

static iDAFSError 
splitBoxes(iEntityPtr new_entity,iEntityPtr bp,int4 spot,short len)
{
	iBox b,bb;
	iDAFSError e;
	
	b = bb = i_GetEnclosingBox(bp);
	b.w = (int2)((bb.w * spot) / len);
	bb.x += b.w;
	bb.w -= b.w;
	if (e=i_SetBox(new_entity,b))
		return e;
	if (e=i_SetBox(bp,bb))
		return e;
	return DAFSOK;
}

static iDAFSError
myAddChar(iEntityPtr p, int4 len, iUCode key)
{
	/* ah ha --can't add to the ascii string a non-ascii value */
			/* but aren't there more guys like this? */
	if(key > 127){
		iUCode *us;
		us = i_AllocUText(p);
		i_SetUText(p, us);
		smfree(us);
		return i_AddUni(p, len, key);
	} else {
		return i_AddChar(p, len, (char)key);
	}
}

Dll iDAFSError APIENTRY
i_InsertChar(iEntityPtr *bp, int4 *beg, 
			 iEntityPtr *ep, int4 *end, 
			 iUCode key) 
{
	iDAFSError e;
	iEntityPtr p,new_entity;
	int4 pdt,dt;
	short len;
	char *temp;
	iUCode *tempu;
	short usave; 
	char csave;
	iTypeIndex t,pt;
	p = i_FindEntityBack(*bp,0,DATAEN,0);
	if (p) {
		pt = i_GetEntityType(p);
		pdt = i_GetEDataType(p);
	} else {
		pt = iNullT;
		pdt = iDNo;
	}
	t = i_GetEntityType(*bp);
	dt = i_GetEDataType(*bp);
	switch (key) {
	case '\b':
		break;
	case '\n':
	case '\r':
		if (t == iNewLineT) {
			i_SetShort(*bp,i_GetShort(*bp,0) + 1,0);
			(*beg)++;
			if (*bp == *ep) (*end)++;
		} else if (*beg) {
			if (t == iSpaceT) {
				new_entity = i_NewEntityBefore(t,*bp);
				if (!new_entity) return MallocFailError;
				len = (short)i_GetShort(*bp,0);
				i_SetShort(new_entity,*beg,0);
				i_SetShort(*bp,len - *beg,0);
				if (e=splitBoxes(new_entity,*bp,*beg,len))
					return e;
				*beg = 0;
				if (*bp == *ep) *end = 0;
				if (e = splitEntity(*bp,new_entity,iNewLineT,&new_entity))
					return e;
				i_SetShort(new_entity,1,0);
			} else if (dt == iDString) {
				new_entity = i_NewEntityBefore(t,*bp);
				if (!new_entity) return MallocFailError;
				len = strlen(i_GetText(*bp));
				temp = smalloc(len + 1);
				strcpy(temp,i_GetText(*bp));
				csave = temp[*beg];
				temp[*beg] = 0;
				i_SetText(new_entity,temp);
				temp[*beg] = csave;
				i_SetText(*bp,&temp[*beg]);
				smfree(temp);
				if (e=splitBoxes(new_entity,*bp,*beg,len))
					return e;
				*beg = 0;
				if (*bp == *ep) *end = 0;
				p = i_FindEntityBack(*bp,iLineT,ALL,0);
				if (!p) {
					new_entity = i_NewEntityBefore(iNewLineT,*bp);
					if (!new_entity) return MallocFailError;
				} else
					if (e = splitEntity(*bp,p,iNewLineT,&new_entity))
						return e;
				i_SetShort(new_entity,1,0);
			} else if (dt == iDUString) {
				new_entity = i_NewEntityBefore(t,*bp);
				if (!new_entity) return MallocFailError;
				len = ucslen(i_GetUText(*bp));
				tempu = (iUCode*)smalloc((len + 1) * sizeof(iUCode));
				ucscpy(tempu,i_GetUText(*bp));
				usave = tempu[*beg];
				tempu[*beg] = 0;
				i_SetUText(new_entity,tempu);
				tempu[*beg] = usave;
				i_SetUText(*bp,&tempu[*beg]);
				smfree(tempu);
				if (e=splitBoxes(new_entity,*bp,*beg,len))
					return e;
				*beg = 0;
				if (*bp == *ep) *end = 0;
				p = i_FindEntityBack(*bp,iLineT,ALL,0);
				if (!p) {
					new_entity = i_NewEntityBefore(iNewLineT,*bp);
					if (!new_entity) return MallocFailError;
				} else
					if (e = splitEntity(*bp,p,iNewLineT,&new_entity))
						return e;
				i_SetShort(new_entity,1,0);
			} else {
				p = i_FindEntityBack(*bp,iLineT,ALL,0);
				if (!p) {
					new_entity = i_NewEntityBefore(iNewLineT,*bp);
					if (!new_entity) return MallocFailError;
				} else
					if (e = splitEntity(*bp,p,iNewLineT,&new_entity))
						return e;
				i_SetShort(new_entity,1,0);
			}
		} else if (pt == iNewLineT) {
			i_SetShort(p,i_GetShort(p,0) + 1,0);
		} else {
			if (i_GetEntityType(*bp) == iLineT) p = 0;
			else p = i_FindEntityBack(*bp,iLineT,ALL,0);
			if (!p) {
				new_entity = i_NewEntityBefore(iNewLineT,*bp);
				if (!new_entity) return MallocFailError;
			} else
				if (e = splitEntity(*bp,p,iNewLineT,&new_entity))
					return e;
			i_SetShort(new_entity,1,0);
		}
		break;
	case ' ':
		if (t == iSpaceT) {
			i_SetShort(*bp,i_GetShort(*bp,0) + 1,0);
			(*beg)++;
			if (*bp == *ep) (*end)++;
		} else if (*beg) {
			if (t == iNewLineT) {
				new_entity = i_NewEntityBefore(iNewLineT,*bp);
				if (!new_entity) return MallocFailError;
				i_SetShort(new_entity,*beg,0);
				i_SetShort(*bp,i_GetShort(*bp,0) - *beg,0);
				*beg = 0;
				if (*bp == *ep) *end = 0;
				new_entity = i_NewEntityBefore(iLineT,*bp);
				if (!new_entity) return MallocFailError;
				if (e=i_AddUni(new_entity,0,key))
					return e;
			} else if (dt == iDString) {
				if (e=myAddChar(*bp,*beg,key))
					return e;
				(*beg)++;
				if (*bp == *ep) (*end)++;
			} else if (dt == iDUString) {
				if (e=i_AddUni(*bp,*beg,key))
					return e;
				(*beg)++;
				if (*bp == *ep) (*end)++;
			}
		} else if (pt == iSpaceT) {
			i_SetShort(p,i_GetShort(p,0) + 1,0);
		} else if (dt == iDString) {
			if (e=myAddChar(*bp,0,key))
				return e;
			(*beg)++;
			if (*bp == *ep) (*end)++;
		} else if (dt == iDUString) {
			if (e=i_AddUni(*bp,0,key))
				return e;
			(*beg)++;
			if (*bp == *ep) (*end)++;
		} else if (pdt == iDString) {
			if (e=myAddChar(p,strlen(i_GetText(p)),key))
				return e;
		} else if (pdt == iDUString) {
			if (e=i_AddUni(p,ucslen(i_GetUText(p)),key))
				return e;
		} else if (t == iNewLineT) {
			new_entity = i_NewEntityBefore(iLineT,*bp);
			if (!new_entity) return MallocFailError;
			if (e=i_AddUni(new_entity,0,key))
				return e;
		} else {
			p = i_FindEntityBack(*bp,iLineT,ALL,0);
			if (e = splitEntity(*bp,p,iLineT,&new_entity))
				return e;
			if (e=i_AddUni(new_entity,0,key))
				return e;
		}
		break;
	default:
		if (*beg) {
			if (t == iSpaceT) {
				new_entity = i_NewEntityBefore(iSpaceT,*bp);
				if (!new_entity) return MallocFailError;
				len = (short)i_GetShort(*bp,0);
				i_SetShort(new_entity,*beg,0);
				i_SetShort(*bp,len - *beg,0);
				if (e=splitBoxes(new_entity,*bp,*beg,len))
					return e;
				*beg = 0;
				if (*bp == *ep) *end = 0;
				p = i_FindEntityBack(*bp,iLineT,ALL,0);
				if (e = splitEntity(*bp,p,iLineT,&new_entity))
					return e;
				if (e=i_AddUni(new_entity,0,key))
					return e;
			} else if (t == iNewLineT) {
				new_entity = i_NewEntityBefore(iNewLineT,*bp);
				if (!new_entity) return MallocFailError;
				i_SetShort(new_entity,*beg,0);
				i_SetShort(*bp,i_GetShort(*bp,0) - *beg,0);
				*beg = 0;
				if (*bp == *ep) *end = 0;
				new_entity = i_NewEntityBefore(iLineT,*bp);
				if (!new_entity) return MallocFailError;
				if (e=i_AddUni(new_entity,0,key))
					return e;
				if (e=guessBox(new_entity))
					return e;
			} else if (dt == iDString) {
				if (e=myAddChar(*bp,*beg,key))
					return e;
				(*beg)++;
				if (*bp == *ep) (*end)++;
			} else if (dt == iDUString) {
				if (e=i_AddUni(*bp,*beg,key))
					return e;
				(*beg)++;
				if (*bp == *ep) (*end)++;
			}
		} else if (dt == iDString) {
			if (e=myAddChar(*bp,0,key))
				return e;
			(*beg)++;
			if (*bp == *ep) (*end)++;
		} else if (dt == iDUString) {
			if (e=i_AddUni(*bp,0,key))
				return e;
			(*beg)++;
			if (*bp == *ep) (*end)++;
		} else if (pdt == iDString) {
			if (e=myAddChar(p,strlen(i_GetText(p)),key))
				return e;
		} else if (pdt == iDUString) {
			if (e=i_AddUni(p,ucslen(i_GetUText(p)),key))
				return e;
		} else if (t == iNewLineT) {
			new_entity = i_NewEntityBefore(iLineT,*bp);
			if (!new_entity) return MallocFailError;
			if (e=i_AddUni(new_entity,0,key))
				return e;
			if (e=guessBox(new_entity))
				return e;
		} else {
			p = i_FindEntityBack(*bp,iLineT,ALL,0);
			if (e = splitEntity(*bp,p,iLineT,&new_entity))
				return e;
			if (e=i_AddUni(new_entity,0,key))
				return e;
			if (e=guessBox(new_entity))
				return e;
		}
		break;
	}
	return DAFSOK;
}

Dll iDAFSError APIENTRY
i_Edit(iEntityPtr *bp, int4 *beg, 
	   iEntityPtr *ep, int4 *end, 
	   iUCode key) 
{
	iDAFSError e;
	iEntityPtr p;
	
	if (key == '\b' && *bp == *ep && *beg == *end) {
		if (*beg) (*beg)--;
		else {
			p = i_FindEntityBack(*bp,0,DATAEN,0);
			if (!p) return DAFSOK;
			*bp = p;
			switch (i_GetEDataType(*bp)) {
			case iDString:
				*beg = strlen(i_GetText(*bp))-1;
				break;
			case iDUString:
				*beg = ucslen(i_GetUText(*bp))-1;
				break;
			case iDShort:
				*beg = i_GetShort(*bp,0)-1;
				break;
			case iDPosSet:
			case iD1PosSet:
				*beg = 0;
				break;
			default: break;
			}
		}
	}
	if (e = deleteSelection(bp,beg,ep,end,key == '\b'))
		return e;
	return i_InsertChar(bp,beg,ep,end,key);
}

static int4 isDataEnSub(iEntityPtr target)
{
	return ((i_GetEDataType(target) != iDNo) && 
	        (i_GetEDataType(target) != iDConf) &&  
	        (i_GetEDataType(target) != iDUserData) &&
	        ((i_GetEDataType(target) != iDShort) || 
	         (i_GetEntityType(target) == iSpaceT) ||
	         (i_GetEntityType(target) == iNewLineT)));
}


int4 
_iAncestorIsData(iEntityPtr en)
{
	iEntityPtr parent = en ? i_GetParentEntity(en) : 0;
	int4 value;
	
	value =  (parent ? 
			  isDataEnSub(parent) || _iAncestorIsData(parent) : 
			  FALSE);
	
	return value;
}

int4 _iIsDataEn(iEntityPtr target)
{
	return isDataEnSub(target) && !_iAncestorIsData(target);
}

Dll void APIENTRY
i_Object2Range(iEntityPtr e,iEntityPtr *bp,int4 *beg,iEntityPtr *ep,int4 *end)
{
	iEntityPtr t;
	
	*beg=0;
	*end=0;	
	if (_iIsDataEn(e)) {
		*bp=e;
		*ep=i_FindEntityForward(e,0,DATAEN,0);
	}else{
		*bp=i_FindEntityForward(e,0,DATAEN,0);
		*ep=e;
		while(t=i_GetLastEntity(*ep))
			*ep=t;
		*ep=i_FindEntityForward(*ep,0,DATAEN,0);
	} 
	return;
}

Dll iEntityPtr APIENTRY
i_Range2Object(iEntityPtr bp,int4 beg,iEntityPtr ep,int4 end)
{
	iEntityPtr a,b;
	iBox bx;
	
	if (bp==ep) {
		bx=i_GetEnclosingBox(bp);
		if (bx.w)
			return bp;
		if(beg==end&&end==0){
			bp=i_FindEntityBack(ep,0,DATAEN,0);
			if(!bp)
				bp=ep;
			else
				ep=bp;
		}
	}
	if(end==0)
		ep=i_FindEntityBack(ep,0,DATAEN,0);
	
	for(a=bp;a;a=i_GetParentEntity(a))
		for(b=ep;b;b=i_GetParentEntity(b))
			if(a==b){
				bx=i_GetEnclosingBox(a);
				if(bx.w!=0)
					return a;
			}
	return 0;
}

Dll iDAFSError APIENTRY
i_StartObjEdit(iEntityPtr ent,iEditRecordPtr *r)
{
	*r=(iEditRecordPtr)smalloc(sizeof(int4));
	if (!*r) return MallocFailError;
	**(int4**)r = 1;
	return DAFSOK;
}

Dll iDAFSError APIENTRY
i_EndObjEdit(iEntityPtr ent,iEditRecordPtr r,int4 *deleteme)
{
	*deleteme = (*(int4*)r > 1) && !ucslen(i_GetUText(ent));
	smfree(r);
	return DAFSOK;
}

Dll iDAFSError APIENTRY
i_ObjEdit(iEntityPtr ent,iEditRecordPtr r,iUCode key,int4 *deleteme)
{
	iDAFSError e;
	iUCode buf[2];
	
	if (i_GetChildEntity(ent) && (e = i_Collapse(ent,0)))
		return e;
	*deleteme = 0;
	switch(key){
	case 20:	/*^u*/
		buf[0] = 0;
		if (e = i_SetUText(ent,buf))
			return e;
		break;
		
	case '\b':
		if (*(int4*)r == 1) {
			*deleteme = 1;
			return DAFSOK;
		}
		i_DelUni(ent,-1);
		break;
		
	default:
		if (*(int4*)r > 1) {
			if (e=i_AddUni(ent,-1,key))
				return e;
		} else {
			buf[0] = key;
			buf[1] = 0;
			if (e = i_SetUText(ent,buf))
				return e;
		}
		break;
	}
	*(int4*)r = 2;
	return DAFSOK;
	
}



/*
// bogus -- does not read in UTF
*/
Dll iDAFSError APIENTRY
i_Text2Dafs(char *text, iEntityPtr *dp)
{
	iEntityPtr doc,pl,pr;
	int4 beg,end;
	iDAFSError e;
	iEntityPtr dummy_nl;
	
	*dp = 0;
	doc = i_NewEntity(iDocumentT,0);
	if (!doc) return MallocFailError;
	dummy_nl = pl = pr = i_NewEntity(iNewLineT,doc);
	if (!pr) {
		i_DisposeEntity(doc);
		return MallocFailError;
	}
	i_SetShort(pl,1,0);
	beg = end = 0;
	while (*text) {
		/* bogus check here if it *text is a character or
		   a UTF sequence
		   */
		iUCode ubuf[4];
		int4 incr;
		incr = THOmbtouc(ubuf, text, 4);
		if (e = i_InsertChar(&pl,&beg,&pr,&end,ubuf[0])) {
			i_DisposeEntity(doc);
			return e;
		}
		text += incr;
	}
	*dp = doc;
	
	i_DisposeEntity(dummy_nl);

	return DAFSOK;
}

Dll iDAFSError APIENTRY
i_UCode2Dafs(iUCode *text,iEntityPtr *dp, int4 swap)
{
	iEntityPtr doc,pl,pr;
	int4 beg,end;
	iDAFSError e;
	iEntityPtr dummy_nl;
	
	*dp = 0;
	doc=i_NewEntity(iDocumentT,0);
	if (!doc) return MallocFailError;
	dummy_nl = pl = pr = i_NewEntity(iNewLineT,doc);
	if (!pr) {
		i_DisposeEntity(doc);
		return MallocFailError;
	}
	i_SetShort(pl,1,0);
	beg = end = 0;
	while (*text) {
		if (swap)
			*text = i_SwapS(*text);
		if (e = i_InsertChar(&pl,&beg,&pr,&end,*text)) {
			i_DisposeEntity(doc);
			return e;
		}
		text++;
	}
	*dp = doc;

	i_DisposeEntity(dummy_nl);

	return DAFSOK;
}



Dll iDAFSError APIENTRY
i_AddChars(iEntityPtr entity, int2 position, const char *text)
{
	char *p,*q, *temp;
	int2 newsize,os,ns,i;
	iCallBackProc gCallBack;
	void *gUserData;
	
	if (!entity)
		return NotEnoughData;
	i_GetCallBack(&gCallBack, &gUserData);
	if (gCallBack)
		MatrCallBack(mWillUpdate,gUserData,entity);
	if (i_GetEDataType(entity) != iDString)
		i_SetText(entity, text);
	else {
		const char * oldtext = i_GetText(entity);
		char *newtext = (char *) smalloc((strlen(text) + strlen(oldtext) + 1) * sizeof(char));
		if(!newtext)
			return MallocFailError;

		if(position < 0) position = strlen(oldtext);
		strncpy(newtext, oldtext, position);
		newtext[position] = '\0';
		strcat(newtext, text);
		strcat(newtext, oldtext+position);

		i_SetText(entity, newtext);
		smfree(newtext);
	}
	if (gCallBack)
		MatrCallBack(mUpdate,gUserData,entity);
	return DAFSOK;
}

Dll iDAFSError APIENTRY
i_AddChar(iEntityPtr ent, int2 position, char letter)
{
	char buf[2];
	buf[1] = 0;
	buf[0] = letter;
	return i_AddChars(ent,position,buf);
}

Dll iDAFSError APIENTRY
i_DelChar(iEntityPtr entity, int2 position)   
{
	char *p;
	iCallBackProc gCallBack;
	void *gUserData;

	if (!entity)
		return NotEnoughData;
	i_GetCallBack(&gCallBack, &gUserData);
	if (gCallBack)
		MatrCallBack(mWillUpdate,gUserData,entity);
	if (i_GetEDataType(entity) != iDString)
		i_FreeContents(entity);
	else {
		const char *oldtext = i_GetText(entity);
		char *text = (char *) smalloc((strlen(oldtext)+1)*sizeof(char));		
		strcpy(text, oldtext);
		if (position < 0)
			position = strlen(text)-1;
		for (p = (text+position); *p = p[1]; p++);
		i_SetText(entity, text);
		smfree(text);
	}
	if (gCallBack)
		MatrCallBack(mUpdate,gUserData,entity);
	return DAFSOK;
}

Dll iDAFSError APIENTRY
i_AddUnis(iEntityPtr entity, int2 position, const iUCode *text)  
{
	iUCode *p,*q, *temp;
	int2 ns,newsize,os,i;
	iCallBackProc gCallBack;
	void *gUserData;
	
	if (!entity)
		return NotEnoughData;
	i_GetCallBack(&gCallBack, &gUserData);	
	if (gCallBack)
		MatrCallBack(mWillUpdate,gUserData,entity);
	if (i_GetEDataType(entity) != iDUString)
		i_SetUText(entity, text);
	else {
		const iUCode *oldtext = i_GetUText(entity);
		iUCode *newtext = (iUCode *) smalloc((ucslen(text) + ucslen(oldtext) + 1) * sizeof(iUCode));
		if(!newtext)
			return MallocFailError;


		if(position < 0) position = ucslen(oldtext);
		ucsncpy(newtext, oldtext, position);
		newtext[position] = 0;
		ucscat(newtext, text);
		ucscat(newtext, oldtext+position);

		i_SetUText(entity, newtext);
		smfree(newtext);
	}
	if (gCallBack)
		MatrCallBack(mUpdate,gUserData,entity);
	return DAFSOK;
}

Dll iDAFSError APIENTRY
i_AddUni(iEntityPtr ent, int2 position, iUCode letter)  
{
	iUCode buf[2];
	
	buf[1] = 0;
	buf[0] = letter;
	return i_AddUnis(ent,position,buf);
}

Dll iDAFSError APIENTRY
i_DelUni(iEntityPtr entity, int2 position)   
{
	iUCode *p;
	iCallBackProc gCallBack;
	void *gUserData;

	if (!entity)
		return NotEnoughData;
	i_GetCallBack(&gCallBack, &gUserData);
	if (gCallBack)
		MatrCallBack(mWillUpdate,gUserData,entity);
	if (i_GetEDataType(entity) != iDUString)
		i_FreeContents(entity);
	else {
		const iUCode *oldtext = i_GetUText(entity);
		iUCode *unitext = (iUCode *) smalloc((ucslen(oldtext)+1)*sizeof(iUCode));
		ucscpy(unitext, oldtext);
		if (position < 0)
			position = ucslen(unitext)-1;
		for (p = (unitext + position); *p = p[1]; p++);
		i_SetUText(entity, unitext);
		smfree(unitext);
	}
	if (gCallBack)
		MatrCallBack(mUpdate,gUserData,entity);
	return DAFSOK;
}

