
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */
/*
// This file implements all the members of the TopList ADT
*/

#include "util/port.h"

#include "util/c-lib.h"
#include "util/util.h"
#include "util/defines.h"

#include "toplist.h"

static void
topSort(TopList *that)
{
	int i; 
	int j;
	VoidPtr temp;
	
	for(i = 0; i < that->num_elements; i++){
		for(j = i + 1; j < that->num_elements; j++){
			if(that->comparison_function(that->array[i], that->array[j]) < 0){
				temp = that->array[i];
				that->array[i] = that->array[j];
				that->array[j] = temp;
			}
		}
	}
}

void
topCtor(TopList * that, 
		ComparisonFunction input_comparison_function,
		int size)
{
	that->comparison_function = input_comparison_function;
	that->max_elements = size + 1;
	that->num_elements = 0;
	that->array = (VoidPtr *) smalloc(that->max_elements * sizeof(VoidPtr));
}

void
topDtor(TopList *that) 
{ 
	smfree(that->array);
}

VoidPtr
topFront(TopList *that)
{
	if(topEmpty(that))
		halt("TopList::topFront: accessed empty queue.\n");
	
	return that->array[0];
}

void
topEnqueue(TopList *that, VoidPtr new_item)
{
	if(that->num_elements < that->max_elements)
		that->num_elements++;
	that->array[that->num_elements - 1] = new_item;
	topSort(that);
}

VoidPtr
topDequeue(TopList *that)
{
	int i;
	
	if(topEmpty(that))
		halt("TopList::topDequeue: accessed empty queue.\n");      
	else {
		VoidPtr ret_val = topFront(that);
		
		for(i = 0; i < that->num_elements; i++)
			that->array[i] = that->array[i + 1];
		that->num_elements--;
		
		return ret_val;
	}
	
	/* NOT REACHED */
	return 0;
}

void
topRemoveFront(TopList *that)
{
	int i;
	
	if(topEmpty(that)) 
		halt("TopList::topRemoveFront: accessed empty queue.\n");
	else 
		for(i = 0; i < that->num_elements; i++)
			that->array[i] = that->array[i + 1];
}


int
topEmpty(TopList *that)
{ 
	return that->num_elements == 0;
}

int
topFull(TopList *that)
{ 
	return FALSE;
}

int
topLength(TopList *that)
{ 
	return that->num_elements;
}

void
topClear(TopList *that)
{ 
	that->num_elements = 0;
}


