
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */
#include <assert.h>
#include "aString.h"

const char *EmptyString = "";
//********************************************
//!!!something is messed up in the headers - this should be known
// S.Lewis
extern "C" int strncmp(const char *s1, const char *s2, size_t size);


#include "utf.h"
#include "smemory.h"
#include "uString.h"

void
aString::putUnicode(uString & us) const
{
	iUCode *buffer;
	int l;
	int incr;
	int index;
	int uc_index;
	const char *s;
	l = this->length() + 1;
	buffer = (iUCode *) smalloc(l * 3 * sizeof(iUCode));
	s = this->string();
	for (index = 0, uc_index = 0; index < l; uc_index++) {
		incr = THOmbtouc(buffer + uc_index, (char *) s + index, 4);
		if (incr == -1) {
			buffer[0] = '\0';
			break;
		}
		index += incr;
	}
	
	us = buffer;
	smfree(buffer);
}

int aString::utfP() const
{
	int i;
	int result;
	for (i = 0; i < len && ((result = data[i] & 0x80) == 0); i++);
	
	return result;
}

void aString::getUtf(const uString &us)
{
	us.putUtf(*this);
}

int aString::uLength() const
{
	uString us;
	int ul;
	return this->length();
	
	this->putUnicode(us);
	ul = us.length();
	
	return ul;
}

aString::aString(const uString & us)
	: allocated(0),
	  data(NULL),
	  len(0)
{
	this->getUtf(us);
}






aString::~aString(void)
{
	if (data)
		smfree(data);
	//I have trouble with delete[]
}

aString::aString(const aString &c)
	: allocated(0),
	  data(NULL),
	  len(0)
{
	this->extend(c.length());
	len = c.length();
	strcpy(data, c.data);
}

aString::aString(const char *c)
	: allocated(0),
	  data(NULL),
	  len(0)
{
	if (!c)
		c = EmptyString;
	
	len = strlen(c);
	this->extend(len);
	strcpy(data, c);
}

aString&
aString::operator = (const char *c) 
{
	if (!c)
		c = EmptyString;
	
	unsigned l = strlen(c);
	this->extend(l);
	len = l;
	strcpy(data, c);
	
	return (*this);
}

aString& aString::operator = (const aString &c) 
{
	
	if (this == &c)
		return *this;
	this->extend(c.length());
	len = c.length();
	strcpy(data, c.data);
	
	return (*this);
}

int
aString::operator != (const aString &c) const
{
	return strcmp(c.data, data);
}

int aString::operator != (const char *c) const
{
	if (!c)
		c = EmptyString;
	
	return (strcmp(c, data));
}


int aString::operator == (const aString & c) const
{
	return !strcmp(c.data, data);
}

int aString::operator == (const char *c) const
{
	if (!c)
		c = EmptyString;
	return (!strcmp(c, data));
}


int aString::operator <= (const char *c) const
{
	if (!c)
		c = EmptyString;
	return (strcmp(data, c) <= 0);
}

int aString::operator < (const char *c) const
{
	if (!c)
		c = EmptyString;
	
	return (strcmp(data, c) < 0);
}

int aString::operator > (const char *c) const
{
	if (!c)
		c = EmptyString;
	return (strcmp(data, c) > 0);
}

int aString::operator >= (const char *c) const
{
	if (!c)
		c = EmptyString;
	return (strcmp(data, c) >= 0);
}

int aString::operator <= (const aString &c) const
{
	return (strcmp(data, c.data) <= 0);
}

int aString::operator < (const aString &c) const
{
	return (strcmp(data, c.data) < 0);
}

int aString::operator > (const aString &c) const
{
	return (strcmp(data, c.data) > 0);
}

int aString::operator >= (const aString &c) const
{
	return (strcmp(data, c.data) >= 0);
}

aString aString::operator + (const aString &c) const
{
	aString a;
	a = *this;
	a += c;
	return a;
}


aString aString::operator + (const char *c) const
{
	aString a;
	a = *this;
	a += c;
	return a;
}


aString &aString::operator += (const aString &c) {
	unsigned newlen = len + c.length();
	if (newlen == len)
		return (*this);
	//+=null string
	this->extend(newlen);
	if (&c != this) {
		strcpy(&data[len], c.data);
	} else {
		//special handling for +=self
		strncpy(&data[len], data, len);
		data[newlen] = 0;
	}
	len = newlen;
	return (*this);
	
}


aString &aString::operator += (const char *c) {
	if (!c)
		return (*this);
	
	unsigned newlen = len + strlen(c);
	if (newlen == len)
		return (*this);
	//+=null string
	if (newlen > allocated)
		this->extend(newlen);
	
	strcpy(&data[len], c);
	len = newlen;
	return (*this);
}



void aString::clear(void)
{
	len = 0;
	this->extend(1);
	data[0] = '\0';
}

#ifndef max
#define max(x,y) ((x>y) ? x : y)
#endif

/*
// There were some bugs related to not asking for enough memory
// from extend -- confusiion about how to count the null
// therefore -- I always add at least 1 to requested length
// and report allocated as 1 less than actual
//
*/
void
aString::extend(unsigned newsize)
{
	// allocated stores the space for characters less null
	if (allocated >= newsize && allocated != 0)
		return;
	//big enough
	
	newsize = NextPower2(newsize + 1);
	newsize = max(newsize, 4);
	if (data) {
		data = (char *) smrealloc(data, newsize);
	} else {
		data = (char *) smalloc(newsize);
		data[0] = '\0';
	}
	
	allocated = newsize - 1;
}



//*************************************************************
//-Name Operators
//*************************************************************
int operator == (const char *c, const aString & s) {
	if (!c)
		c = EmptyString;
	return (strcmp(c, s.data));
}

int operator < (const char *c, const aString &s) {
	if (!c)
		c = EmptyString;
	return (strcmp(c, s.data) < 0);
}

int operator > (const char *c, const aString &s) {
	if (!c)
		c = EmptyString;
	return (strcmp(c, s.data) > 0);
}

int operator >= (const char *c, const aString &s) {
	if (!c)
		c = EmptyString;
	return (strcmp(c, s.data) >= 0);
}

int operator <= (const char *c, const aString &s) {
	if (!c)
		c = EmptyString;
	return (strcmp(c, s.data) <= 0);
	return 0;
}

int operator != (const char *c, const aString &s) {
	if (!c)
		c = EmptyString;
	return (!strcmp(c, s.data));
}


aString operator + (const char *c, const aString &s) {
	aString r = c;
	r += s;
	return (r);
}


//{function ******************************************
//-Name NextPower2
// -Function return smallest power of 2 > x
// *max out at 0x10000000
// *min = 4
// -Return - that number
// -Arguments
// -x input value
//} ***************************************************
unsigned long NextPower2(unsigned long x) {
	const unsigned long MaxPower = 0x10000000;
	unsigned long p = 4;
	
	//make sure we do not wrap
	if (x >= MaxPower)
		return (MaxPower);
	
	while (p < x)  p <<= 1;
	
	return (p);
}



/* ---------------------------------
* dup_string
*  returns a copy of the input string
*  allocated
*  will work even if in is NULL
---------------------------------- */
char *dup_string(const char *in) {
	char *out;
	int len;
	if (!in)
		return (dup_string(""));
	
	len = strlen(in);
	out = (char *) smalloc(len + 1 * sizeof(char));
	strcpy(out, in);
	out[len] = 0;
	return (out);
}

const char *aString::string(void) const  { if(data) return(data); return(EmptyString);  }
int   aString::length(void) const { return(len);} 
